#include <hb.h>

/*
 * File: body.c
 *
 * References:
 *  Users' Guide for the Harwell-Boeing Sparse Matrix Collection (Release I),
 *     Duff, Grimes, and Lewis, publisher?, Oct. 1992
 *
 * Purpose: To read the body of harwell-boeing formatted sparse matrices
 *          given that the header has already been read in.
 *
 * Contents: ReadHBBody ProcInts ProcDoubles ReadHBFile DumpHBMatrixColumnRange
 * 
 * Started: about 2/1/94
 *
 * Modified: 2/12/94 - comment cleanup
 *
 */

/*
 * Function: ProcInts
 *
 * Purpose: Process lines of the file (stored in buffer) using
 *          the format held in format 
 *
 * Parameters: format = the format to be used
 *             buffer = the input lines
 *             array = the processed integers
 *             lines = the number of lines expected
 *             bufsize = size in bytes of the buffer
 *
 * Return Value: the number of bytes processed
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609, 
 *             tloos@cs.indiana.edu
 *
 * Started: 2/6/94
 *
 * Modified: 2/7/94 return bytes processed and remove left parameter (tjl)
 *           2/9/94 convert line to a char * to reflect changes in GetBuffLine
 *           10/28/94 -- fix overflow in last line bug 
 * 
 */
int ProcInts(Format *format, char *buffer, int *array, int *lines, int bufsize)
{
    char *line;
    int nl_seen;
    int i;
    int bytecount;
    int ints_per_line = format->count;
    int cur_ints_per_line; /* 10/28/94 */
    int int_size = format->width;
    int cur_byte;
    int total = 0;

    /* while there's work */
    while (*lines > 0)
    {
	/* get a line from the buffer */
	bytecount = GetBuffLine(buffer, &line, MAX_LINE_SIZE, &nl_seen);

	/* 10/28/94 -- get actual number of ints per line */
	cur_ints_per_line = bytecount/int_size;

	/* if no newline seen, need to re-read, so break */
	if ((!nl_seen) || ((total+bytecount) > bufsize))
	{
	    break;
	}
	/* else */

	cur_byte = 0;

	/* for each integer in the current line, format it and go on */
/*	for (i = 0; i < ints_per_line; i++) */
	/* changed on 10/28/94 */
	for (i = 0; i < cur_ints_per_line; i++) 
	{
	    *array = FormatInt( &(line[cur_byte]), int_size );
	    cur_byte += int_size;
	    array++; /* move onto next element */
	}
	/* update total */
	total += bytecount;

	/* decrement number of lines   */
	(*lines)--;
	buffer += bytecount;
    }
    return( total );
}
/*
 * Function: ProcDoubles
 *
 * Purpose: Process one line of the file (stored in buffer) using
 *          the format held in format 
 *
 * Parameters: format = the format to be used
 *             buffer = the input line
 *             array = the processed doubles
 *             lines = the number of lines expected
 *             left = a buffer for unprocessed chars
 *             max = a pointer to storage for the max value of the section
 *             min = a pointer to storage for the min value of the section
 *             absmax = a pointer to storage for the min abs. value of the 
 *                      section
 *             absmin = a pointer to storage for the min abs. value of the 
 *                      section
 *             bufsize = size in bytes of the buffer
 *
 * Return Value: the number of doubles processed
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: 2/6/94
 *
 * Modified: 2/7/94 return bytes processed and remove left parameter (tjl)
 *           2/9/94 update for changed GetBuffLine
 * 
 */
int ProcDoubles(Format *format, char *buffer, double *array, int *lines,
		double *max, double *min, double *absmax, double *absmin,
		int bufsize)
{
    char *line, c;
    double absval;
    int nl_seen;
    int i;
    int bytecount;
    int doubles_per_line = format->count;
    int double_size = format->width;
    int cur_byte;
    int total = 0;
    int checkmax;
    double FormatFloat();

    if (max == (double *) NULL)
	checkmax = 0;
    else
	checkmax = 1;

    /* while there's work */
    while (*lines > 0)
    {
	/* get a line from the buffer */
	bytecount = GetBuffLine(buffer, &line, MAX_LINE_SIZE, &nl_seen);

	/* if no newline seen, need to re-read, so break */
	if ((!nl_seen) || ((total+bytecount) > bufsize))
	{
	    break;
	}
	/* else */

	/* for each integer in the current line, format it and go on */
	cur_byte = 0;
	for (i = 0; i < doubles_per_line; i++)
	{
	    /* if we need to scan float before conversion, do it */
	    if (( format->format == 'D') || (format->format == 'Q'))
	    {
		*array = FormatFloat( &(line[cur_byte]), double_size );
	    }
	    else
	    /* roll our own conversion, since it'll be faster */
	    {
		/* make line a string */
		c = line[cur_byte+double_size];
		line[cur_byte+double_size] = '\0';
		
		/* atof the string */
		*array = atof( &(line[cur_byte]) );

		/* restore prev. contents */
		line[cur_byte+double_size] = c;
	    }
	    /* check for max, min, etc. if necessary */
	    if (checkmax)
	    {
		if (*array > *max)
		    *max = *array;
		if (*array < *min)
		    *min = *array;
		absval = fabs(*array);
		if (absval > *absmax)
		    *absmax = absval;
		if (absval < *absmin)
		    *absmin = absval;
	    }
		    
	    cur_byte += double_size;
	    array++; /* move onto next element */
	}
	/* update total */
	total += bytecount;
	/* decrement number of lines needed to do */
	(*lines)--;
	buffer += bytecount;
    }
    return( total );
}

/*
 * Function: ReadHBBody()
 * 
 * Parameters: fd = file descriptor for matrix
 *             hb = a pointer to an HB formatted matrix
 *             use_c = use C language conventions for matrix indices
 *                     (0 == leave it alone, !0 == use C)
 *             
 * Purpose: to read in the body of a Harwell Boeing formatted matrix 
 *          and determine some more stats that we store in the header.
 *          
 * Algorithm:  
 *              kill any elemental matrix, since unsupported
 *              malloc out the col_inds, row_inds, and values
 *                 arrays
 *              malloc out RHS stuff if necessary
 *              section = col
 *              while still can read in a big buffer
 *                 do per section processing
 *                    process integers for ROW and POINTER sections
 *                    process doubles for VALUE and RHS... sections
 *                 handle, if needed, end of input buffer problems
 *              end of while
 *
 * Return Values:  OK - if the file body reads in OK
 *                !OK - if an error occurs (see the file hb.h for internal
 *                      error codes)
 *
 * Calls: CProcInts ProcInts ProcDoubles Get1DArray
 *
 * Notes: RHS and end of buffer code hasn't been tested for one reason
 *        or another -- EOB stuff was looked at in an earlier version,
 *        but not tested here.
 *
 *        Unassembled (Elemental) matrices not supported yet.
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: 2/4/94
 *
 * Modified:  2/ 6/94 -- kept on working
 *            2/ 9/94 -- added use_c flag
 *            2/12/94 -- cleaned up comments
 *            2/28/94 -- hacked in ignoring of sparse RHSs 
 *            3/22/94 -- support elemental matrix reading
 *            3/23/94 -- figured out RHS stuff (I hope)
 *            5/16/94 -- fixed bugs on over 8MB section of code 
 */

int ReadHBBody( int fd, HBMatrix *hb, int use_c )
{
    char *bigbuff;
    char okbuff[BIG_BUF_SIZE+100];
    char debugmsg[100];
    int continuation = 0;
    int SparseRHS = 0; /* are sparse RHS's in the code */
    Format *format[MAX_SECTIONS]; /* format pointer for each section */
    int count[MAX_SECTIONS]; /* line count for each section of the file */
    int *int_ptr[2]; /* pointers to the int arrays for ROW and POINTER sect. */
    double *double_ptr[MAX_SECTIONS]; /* pointers to double arrays */

    char names[MAX_SECTIONS][30];
    int input_count;
    int curcount,origcount[MAX_SECTIONS];
    int leftover;
    int section;
    int curbyte;
    int items_done;
    int i;
    int RHS_subsections;
    int number_of_values; /* number of values in values array to read */

    /* read in a big buffer */
    
    /*
     * init arrays 
     * these arrays will be used in the main processing (per section)
     * loop
     *
     * one may wonder why I didn't set things up this way initially.
     *  to be honest, the only reason why I can think of was to make
     *  the header code more human readable vs. machine readable.
     *
     *  well, let's fix that here.
     */

    TRACE("ReadHBBody entered\n");

    format[POINTER_SECTION] = &(hb->head.PointerFormat);
    format[ROW_SECTION] = &(hb->head.RowIndexFormat);
    format[VALUE_SECTION] = &(hb->head.ValueFormat);
    format[RHS_SECTION] = &(hb->head.RHSFormat);
    format[GUESS_SECTION] = &(hb->head.RHSFormat);
    format[SOLUTION_SECTION] = &(hb->head.RHSFormat);

    count[POINTER_SECTION] = hb->head.PointerLines;
    count[ROW_SECTION] = hb->head.IndexLines;
    count[VALUE_SECTION] = hb->head.ValueLines;
    
    /* calculate stuff for RHS sub-sections */
    RHS_subsections = 1;
    count[RHS_SECTION] = hb->head.RHSLines;
    count[GUESS_SECTION] = hb->head.RHSLines;
    count[SOLUTION_SECTION] = hb->head.RHSLines;

    strcpy(names[POINTER_SECTION], "column indices");
    strcpy(names[ROW_SECTION], "row indices");
    strcpy(names[VALUE_SECTION], "values");
    strcpy(names[RHS_SECTION], "RHS values");
    strcpy(names[GUESS_SECTION], "guess values");
    strcpy(names[SOLUTION_SECTION], "solution values");
    /* save backup copies of all the counts */
    for (i = 0; i < MAX_SECTIONS; i++)
	origcount[i] = count[i];

    if (hb->head.RHSLines > 0)
    {
	if (hb->head.GuessSupplied == 'G')
	{
	    RHS_subsections++;
	}
	else
	{
	    count[GUESS_SECTION] = 0;
	}

	if (hb->head.ExactSupplied == 'X')
	{
	    RHS_subsections++;
	}
	else
	{
	    count[SOLUTION_SECTION] = 0;
	}
	count[RHS_SECTION] /= RHS_subsections;
	count[GUESS_SECTION] /= RHS_subsections;
	count[SOLUTION_SECTION] /= RHS_subsections;
    }
    sprintf(debugmsg,"RHS_subections = %d, RHS count = %d, Guess count = %d, Soln count = %d\n",RHS_subsections,count[RHS_SECTION],count[GUESS_SECTION],count[SOLUTION_SECTION]);
    DEBUG2(debugmsg);
    /* reset number_of_values depending on the assembly flag */
    if (hb->head.Assembly == IsElemental)
    {
	number_of_values = hb->head.NumElemMatEntries + 10;
	sprintf(debugmsg,"Elemental Matrix: number_of_values = %d\n",number_of_values);
    }
    else /* assembled (std.) matrix */
    {
	number_of_values = hb->head.NumRowIndex + 10;
	sprintf(debugmsg,"Assembled Matrix: number_of_values = %d\n",number_of_values);
    }
    DEBUG2(debugmsg);

    /* get yer memory here */

    sprintf(debugmsg,"ReadHBBody: memory allocation 1: numcols = %d\n", hb->head.NumCols);  
    DEBUG2(debugmsg);

    /* add on some "engineering safety" */
    if ((hb->col_inds = (int *) Get1DArray( (hb->head.NumCols + 20), sizeof(int),"col_inds", hb)) == NULL)
    {
	error("Couldn't allocate col_inds",MEMERR);
	/* hopefully won't get here */
	exit(MEMERR);
    }

    sprintf(debugmsg,"ReadHBBody: memory allocation 2: NumRows = %d\n",hb->head.NumRowIndex);  
    DEBUG2(debugmsg);
    
    if ((hb->row_inds = (int *) Get1DArray( (hb->head.NumRowIndex + 20), 
				  sizeof(int),"row_inds", hb)) == NULL)
    {
	error("Couldn't allocate row_inds",MEMERR);
	/* hopefully won't get here */
	exit(MEMERR);
    }

    sprintf(debugmsg,"ReadHBBody: memory allocation 3, number_of_values = %d\n",number_of_values); 
    DEBUG2(debugmsg);

   /*
    *  hb->values = (double *)Get1DArray( (hb->head.NumRowIndex + 10), 
    *			       sizeof(double), "values", hb);
    */
    if ((hb->values = (double *) Get1DArray( number_of_values, sizeof(double),
					"values", hb)) == NULL)
    {
	error("Couldn't allocate values",MEMERR);
	/* hopefully won't get here */
	exit(MEMERR);
    }
    /* 
     * for now, only do full RHS's 
     */
    if (hb->head.RHSLines > 0)
    {
	if (hb->head.RHSVectorFormat == IsFull)
	{
	    DEBUG2("Mallocing for an RHS\n");
/* fprintf(stderr,"RHS mallocing %d elements\n",(hb->head.NumRHS * hb->head.NumRows)+10);*/
	    if ((hb->RHS = (double *) Get1DArray( (hb->head.NumRHS * hb->head.NumRows) + 10, sizeof (double), "RHS", hb)) == (double *) NULL)
	    {
		error("Couldn't allocate RHS",MEMERR);
		/* hopefully won't get here */
		exit(MEMERR);
	    }
	}
	else
	{
	    /* HBCleanup(hb); tjl 4/9/94*/
	    count[RHS_SECTION]=0; /* ignore them -- 4/9/94 */
	    SparseRHS = 1;
	    fprintf(stderr,
		"Sparse matrix RHSs unsupported now -- will be ignored\n");
	    hb->head.RHSLines = 0;
	    /* error("Sparse matrix RHSs unsupported now",UNSUPERR); */
	}	

	if ((hb->head.GuessSupplied == 'G') && (SparseRHS == 0))
	{
	    DEBUG2("Mallocing for a guess\n");
/* fprintf(stderr,"Guess mallocing %d elements\n",hb->head.NumRHS * hb->head.NumRows); */
	    if ((hb->guess = (double *) Get1DArray( (hb->head.NumRHS * hb->head.NumRows) + 10, sizeof (double), "guess", hb)) == NULL)
	    {
		error("Couldn't allocate guess",MEMERR);
		/* hopefully won't get here */
		exit(MEMERR);
	    }
	}

	if ((hb->head.ExactSupplied == 'X') && (SparseRHS == 0))
	{
	    DEBUG2("Mallocing for a solution\n");
/* fprintf(stderr,"solution mallocing %d elements\n",hb->head.NumRHS * hb->head.NumRows);*/
	    if ((hb->exact = (double *) Get1DArray( (hb->head.NumRHS * hb->head.NumRows) + 10, sizeof (double), "guess", hb)) == NULL)
	    {
		error("Couldn't allocate exact",MEMERR);
		/* hopefully won't get here */
		exit(MEMERR);
	    }
	}

    }
    sprintf(debugmsg,"ReadHBBody: done with memory allocation\n");
    DEBUG2(debugmsg);

    /* main processing loop */
    /* init's for loop */

    items_done = 0;
    bigbuff = okbuff;
    leftover = 0;
    section = POINTER_SECTION;

    /* need to set up pointer arrays AFTER the memory is allocated */
    int_ptr[POINTER_SECTION] = hb->col_inds;
    int_ptr[ROW_SECTION] = hb->row_inds;

    double_ptr[VALUE_SECTION] = hb->values;
    double_ptr[RHS_SECTION] = hb->RHS;
    double_ptr[GUESS_SECTION] = hb->guess;
    double_ptr[SOLUTION_SECTION] = hb->exact;


    /* while there's data  */
    while ((input_count=read(fd,&(bigbuff[leftover]),BIG_BUF_SIZE)) > 0)
    {
	/* 5/16/94 */
	input_count += leftover;

	/* go thru the sections */
	leftover = 0;
	curbyte = 0;

	/* while there's data in the buffer */
	while ((input_count > curbyte) && (section < MAX_SECTIONS))
	{
	    sprintf(debugmsg,"ReadHBBody: working on section %d, curbyte = %d\n",section,curbyte);
	    DEBUG3(debugmsg);
/*
 *	    printf("section = %d, input_count = %d, leftover = %d, items_done = %d, curbyte = %d\n",section,input_count,leftover,items_done,curbyte);
 *	    printf("buffer size = %d\n",BIG_BUF_SIZE);
 */
	    /* do per section processing */
	    if ((section == POINTER_SECTION) || (section == ROW_SECTION))
	    {
		/* 6/29/94 -- change last arg from input_count to 
		    input_count - curbyte for section continuation */
		if (use_c)
		    curbyte += CProcInts( format[section], bigbuff+curbyte,
					  &((int_ptr[section])[items_done]), 
					  &(count[section]), 
					  input_count-curbyte);
		else
		    curbyte += ProcInts( format[section], bigbuff+curbyte,
					 &((int_ptr[section])[items_done]), 
					 &(count[section]), 
					 input_count-curbyte);
	    }
	    else  /* otherwise they're doubles */
	    {
		if (section == VALUE_SECTION) /* do max/min calculations */
		{
		    curbyte += ProcDoubles( format[section], bigbuff+curbyte,
					    &((double_ptr[section])[items_done]), 
					    &(count[section]), 
					    &(hb->head.MaxValue),
					    &(hb->head.MinValue),
					    &(hb->head.AbsMaxValue),
					    &(hb->head.AbsMinValue),
					    input_count-curbyte);
		}
		else /* RHS, GUESS, or EXACT sections */
		{
		    curbyte += ProcDoubles( format[section], bigbuff+curbyte,
					    &((double_ptr[section])[items_done]), 
					    &(count[section]), 
					    (double *) NULL,
					    (double *) NULL,
					    (double *) NULL,
					    (double *) NULL,
					    input_count-curbyte);
		}
	    }
	    /* if there's left over data, keep track of it */
	    if (count[section] != 0)
	    {
		leftover = input_count - curbyte;
/*
 *		fprintf(stderr,
* "input_count = %d, curbyte = %d, leftover = %d, origcount = %d, count[%d] = %d\n",input_count, curbyte, leftover,origcount[section], section, count[section]);
*/
		/* hmmm -- 5/16/94 */
 		for (i = 0; i < leftover; i++)
		{
		    bigbuff[i] = bigbuff[curbyte+i];
		}
		curbyte = input_count;
		items_done = (origcount[section] - count[section]) * format[section]->count;
		fprintf(stderr,
			"Read in %d %s. Still have to read about %d more.\n",items_done, names[section], count[section]*format[section]->count);
		sprintf(debugmsg,
			"not done with sect. %s (%d): leftover = %d, items_done = %d\n",names[section],section,leftover,items_done);
		DEBUG1(debugmsg);
		continuation = 1;
	    }
	    else /* section finished */
	    {
		/* reset items_done 5/17/94 */
		items_done = 0;
		/* else update section number */
		while ((count[section] == 0)  && (section <= MAX_SECTIONS))
		{
		    section++;
		}
	    }
	}
    }
    /* if used to debugging messages, say done reading the file */
    if (continuation == 1)
	fprintf(stderr,"Done reading the file!\n");
    TRACE("leaving DumpHBBody\n");
    return(OK);
}

/*
 * Function: DumpHBMatrixColumnRange
 *
 * Purpose: as silly as the name sounds, that's its purpose.  to dump
 *          a range of columns in an HB Matrix, probably for debugging
 *
 *
 * Parameters: lo_col = the minimum column in the range to be dumped
 *             hi_col = the maximum column in the range
 *             hb = a pointer to the HBMatrix to be dumped
 *             use_c = a flag to indiciate if C (use_c != 0) array or
 *                     FORTRAN (use_c == 0) array conventions are to be used
 *
 * Return Values: none
 * 
 * Calls: printf
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started:   about 2/10/94
 *
 * Modified:  2/12/94 -- add comments
 *
 */
void DumpHBMatrixColumnRange( int lo_col, int hi_col, HBMatrix *hb, int use_c )
{
    int col, rowidx, min_row_index, max_row_index, rownum;
    int indx;
    double value;
    char errmsg[100];

    /* adjust lo and hi columns -- if out of order, swap */
    if (lo_col > hi_col)
    {
	/* swap elements */
	min_row_index = lo_col; /* use it just as a temp */
	lo_col = hi_col;
	hi_col = min_row_index;
    }

    /* see if low is too low and adjust if necessary */
    if ((lo_col < 0) && (use_c))
    {
	
	lo_col = 0;
	printf("resetting the minimum column number to %d\n",lo_col);
    }
    else if ((lo_col < 1) && (!use_c))
    {
	lo_col = 1;
	printf("resetting the minimum column number to %d\n",lo_col);
    }

    /* see if high is too high and adjust if necessary */
    if (hi_col > hb->head.NumCols)
    {
	if (!use_c)
	    hi_col = hb->head.NumCols+1;
	else
	    hi_col = hb->head.NumCols;
	    
	printf("resetting the maximum column number to %d\n",hi_col);
    }
    else if ((hi_col == hb->head.NumCols) && (use_c))
    {
	hi_col--;
	printf("resetting the maximum column number to %d\n",hi_col);
    }	

    /* state the range of columns to be dumped */
    printf("A dump of columns %d to %d:\n\n",lo_col, hi_col);
    for (col = lo_col; col <= hi_col; col++)
    {
	/* figure out indices based on language */
	if (!use_c)
	{
	    indx = col-1;
	    min_row_index = hb->col_inds[indx]-1;
	    max_row_index = hb->col_inds[indx+1]-1;
	}
	else
	{
	    indx = col;
	    min_row_index = hb->col_inds[indx];
	    max_row_index = hb->col_inds[indx+1];
	}

	/* acutally do the dump */
	for (rowidx = min_row_index; rowidx < max_row_index; rowidx++)
	{
	    if (!use_c)
		rownum = hb->row_inds[rowidx]+1;
	    else
		rownum = hb->row_inds[rowidx];

	    value = hb->values[rowidx];
	    printf("array[%d][%d] = %e\n",rownum, col, value);
	    
	}
    }
}
    


/*
 * Function: ReadHBFile
 *
 * Purpose: To read in a Harwell-Boeing (HB) format sparse matrix 
 *          It's merely a driver function
 *
 * Usage:  int rc; 
 *         HBMatrix hb;
 *         rc = ReadHBMatrix("matrix", &hb, 1 );
 *
 * Parameters: fn = name of file containing a HB matrix
 *             mat = the HB matrix to be read in
 *             use_c = 1 if the c array conventions are to be used 
 *                     0 if the FORTRAN array conventions are to be used
 *
 * Return Values: OK = file read in
 *                !OK == some error, error codes in hb.h
 * 
 * Calls: open ReadHBHeader ReadHBBody close
 *
 * Programmer: Tom Loos, LH310, 855-3609, tloos@cs.indiana.edu
 *
 * Started:   2/4/94
 *
 * Modified:  2/10/94 add use_c flag
 *            2/12/94 clean up comments
 *             9/8/94 if fn == "-", read from stdin
 *
 */
int ReadHBFile(char *fn, HBMatrix *mat, int use_c)
{
    int fd;
    char buffer[1000];
    char debugmsg[100];
    int count, err;

    TRACE("entering ReadHBFile\n");
    /* try to open file read only */
    /* 9/8/94 -- add stdin support if fn == "-" */
    if (strcmp(fn,STDIN_ARG) == 0)
    {
	fd = STDIN_FD;
    }
    else if ((fd = open(fn, O_RDONLY )) < 0)
    {
	sprintf(buffer,"Couldn't open %s\n",fn);
	error(buffer, IOERR);
    }

    DEBUG1("ReadHBFile: opened file\n");
    /* any init's needed before reading in the file are done here */
    HBInit(mat);
    DEBUG1("ReadHBFile: mat init ok\n"); 

    /* if header read in OK, then read in the body */
    if ((err=ReadHBHeader(fd, &(mat->head), &count )) == OK)
    {
	DEBUG1("ReadHBFile: header ok\n"); 
	/* read in the body */
	err = ReadHBBody( fd, mat, use_c ); 
	sprintf(debugmsg,"ReadHBFile: read body, err = %d\n",err);
	DEBUG1(debugmsg);
    }

    /* close file and return status */
    close(fd);
    TRACE("leaving ReadHBFile\n");
    return(err);
}

#ifdef HANDLE_PTNS
#ifndef AEOK
#define AEOK 0
#define AEBAD (-1)
#define AETRUE 1
#define AEFALSE 0
#endif
/*
 * Function: AEisSpace
 *
 * Purpose: see if a string is white space
 *
 * Parameters: str: the string of interest
 *
 * Return Values: AETRUE if it's all whitespace, AEFALSE if not
 * 
 * Calls: strlen iswhite
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: 3/25/95
 *
 * Modified:   
 *
 */

int AEisSpace( char *str )
{
    int i;

    /* get the string length */
    /* lth = strlen(str); */

    /* if the length is bad, no space is considered to be space (zen, eh?) */

    /* look thru all chars to see if we find non-space, if so get outta town */
    i = 0;
    while (str[i] != '\0')
    {
	if (!isspace(str[i]))
	    return(AEFALSE);
	i++;
    }
    /* if here, all white now */
    return(AETRUE);
}
/*
 * Function: ReadPtnFile()
 *
 * Purpose: to read in the partition vector
 *
 * Parameters: ptnfd -- file descriptor to the partition file
 *             ptn_vect -- the partition vector
 *             *num_ptns -- calculated number of partitions
 *             num_rows -- number of input rows/cols (assume a square input)
 *
 * Return Values: AEOK -- if ASW, AEBAD otherwise
 * 
 * Calls: read() atoi() open()
 *
 * Notes: Partition file is of the form:
 *
 *  first row index of ptn1
 *  first row index of ptn2
 *  ...
 *  first row index of ptnN
 *  first row index beyond column
 * 
 * OR 
 *   a Chaco output file (a bunch of 0's, one per line, then a bunch
 *   of 1's, etc.)
 *
 * NOTE: PTN_VECT MUST HAVE AT LEAST 256 ELEMENTS.
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: 9/18/94
 *
 * Modified: 3/24/95 -- changed from ptnfd as parm to filename (fn) to be 
 *              consistent with other AET I/O routines.
 *           4/17/95 -- remove sorted restriction
 *           4/18/95 -- renamed to ReadPtnFileWork and added count only
 *           4/28/95 -- split ReadPtnFileWork into 2 parts: 
 *                        GetPtnCount and GetPtnVect
 */
int ReadPtnFile(char *fn, int *ptn_vect, int *num_ptns)
{
    int chaco_input;
    if (GetPtnCount(fn, num_ptns, &chaco_input) == AEOK)
	return(GetPtnVect(fn, ptn_vect, *num_ptns, chaco_input));
    return(AEBAD);
}
int GetPtnCount(char *fn, int *num_ptns, int *chaco_input)
{
    char buffer[1000];

    int cur_ptn, cur_input_val;
    char *buff_ptr;
    int i;
    int linecount, old_test_val, test_val,max_test_val; /* 4/17 */
    int first_val;
    int lci; /* lci == local chaco input */
    FILE *ptnfp;

    /* printf("in ReadPtnFile\n"); */
    buff_ptr = buffer;
    cur_ptn = -1; /* init to a bad value */
    if ((ptnfp = fopen(fn, "r")) == (FILE *) NULL)
    {
	fprintf(stderr,"ReadPtnFile: failed to open %s\n",fn);
	return(AEBAD);
    }


    /* heuristic for figuring out what kind of input:
     *  count the lines and see what the last number is:
     *  if the last number > the number of lines, definitely not chaco
     *  if the last number < number of lines or unsorted, definitely chaco
     */
    old_test_val = -1;
    max_test_val = old_test_val;
    linecount = 0;
    /* assume not chaco input */
    lci = 0;
    while ((buff_ptr = fgets(buffer,1000,ptnfp)) != (char *) NULL)
    {
	/* skip white space */
	if (AEisSpace(buffer) == AETRUE)
	    continue;

	linecount++;
	test_val = atoi(buffer);

	/* 3/26/96 -- from 9/11/95 HMio.C: save first val */
	if (linecount == 1) /* if first line, save first val special */
	{
	    first_val = test_val;
	}

	if (test_val < 0)
	{
	    printf("bad partition input: line %d, value = %d\n",
		   linecount,test_val);
	    *num_ptns = 0;
	    return(AEBAD);
	}
	/* if not sorted in ascending order, chaco */
	if ((lci == 0) && (test_val < old_test_val))
	    lci = 1;
	/* printf("ci = %d, test_val = %d\n",lci, test_val); */
	/* keep the max and prev. values */
	if (max_test_val < test_val)
	    max_test_val = test_val;
	old_test_val = test_val;
    }

    /* rewind */
    fclose(ptnfp);
    /* if we still think it's not chaco, check line count heuristic */
    if (lci == 0)
    {
	if (linecount > max_test_val)
	    lci = 1;
    }
    /* 3/22/96 -- bought back new bstart from hmio.C stuff in HMP */
    if (lci)
    {
      /* printf("first_val = %d, linecount = %d\n",first_val,linecount);   */
      if (first_val == (linecount - 2)) /* new bstart */
      {
	*num_ptns = first_val;
	/*
	 * printf("%s is a new bstart file with %d partitions\n",fn,
	 * 	 first_val);
	 */
	lci = 2; /* say new bstart */
      }
      else
      {
	*num_ptns = (max_test_val+1);
	/*
	 * printf("%s is a chaco input file with %d partitions\n",fn,
	 *      max_test_val+1);  
	 */
      }
    }
    else
    {
	/*
	 * printf("%s is a partition vector file with %d partitions\n",fn,
	 *  linecount-1);  
	 */
	*num_ptns = linecount-1;
    }
    /* save the chaco input thang */
    /* printf("lci = %d\n",lci);*/
    *chaco_input = lci;
    return(AEOK);
}

int GetPtnVect(char *fn, int *ptn_vect, int num_ptns, int chaco_input)
{
    char buffer[1001];
    int cur_ptn, cur_input_val;
    char *buff_ptr;
    int i;
    int linecount, old_test_val, test_val,max_test_val; /* 4/17 */
    FILE *ptnfp;

    /* printf("in GetPtnVect: num = %d, ci = %d\n",num_ptns,chaco_input);  */
    buff_ptr = buffer;
    cur_ptn = -1; /* init to a bad value */
    if ((ptnfp = fopen(fn, "r")) == (FILE *) NULL)
    {
	fprintf(stderr,"ReadPtnFile: failed to open %s\n",fn);
	return(AEBAD);
    }

    /* zap array */
    for (i = 0; i <= num_ptns; i++)
	ptn_vect[i] = 0;
    
    /*
     * second pass is easy now.  know what kinda file it is and that all
     *  values are good.
     */
    /* while there's data */
    if ((ptnfp = fopen(fn, "r")) == (FILE *) NULL)
    {
	fprintf(stderr,"ReadPtnFile: failed to open %s\n",fn);
	return(AEBAD);
    }
    linecount = 0;
    
    while ((buff_ptr = fgets(buffer,1000,ptnfp)) != (char *) NULL)
    {
	/* skip white space */
	 if (AEisSpace(buffer) == AETRUE)
	     continue;

	cur_input_val = atoi(buffer);
	/* 3/22/96: handle new block start (chaco_input == 2) stuff */
	 if ((chaco_input == 2) && (linecount == 0))
	 {
	     /* this should skip num_ptns value ==> use old bstart 
		code after first line */
	     chaco_input = 0; /* not chaco anymore */

	     /* note that as a side effect, ptn_vect will be indexed starting
		with 0.  To make it indexed by 1, uncomment the next line */
	     /* linecount++; */ /* index by 1 if executed  */
	     continue; /* and skip this iteration */
	 }

	if (chaco_input)
	{
	    ptn_vect[(cur_input_val+1)]++;
	}
	else
	{
	    ptn_vect[linecount] = cur_input_val;
	}
	/* do the linecount afterwards so ptn 0 is assigned */
	linecount++;
    } 

    fclose(ptnfp);
    /* final clean up for chaco */
    if (chaco_input)
    {
	/* get cumulative totals in the ptn vect */
	for (i = 1; i <= num_ptns; i++)
	{
	    ptn_vect[i] = ptn_vect[i-1]+ptn_vect[i];
	}
    }
    return(AEOK);
}

#endif
