#include <hb.h>

/*
 * File: compress.c
 *
 * References:
 *  Users' Guide for the Harwell-Boeing Sparse Matrix Collection (Release I),
 *     Duff, Grimes, and Lewis, publisher?, Oct. 1992
 *
 *  Matlab External Interface Guide, The Math Works, Natick, Mass, Aug. 1992
 * 
 *  Volume I: Xlib Programming Manual, Adrian Nye, O'Reilly and Associates,
 *    7th printing, 12/91
 *
 * Includes: CompressMat CompressSparse CompressDense CompressSparseWork
 *           CompressDenseWork
 *
 * Started: 2/13/94
 *
 * Modified: 3/18/94 -- rewrite of CDM and add TRACE/DEBUG
 *           3/20/94 -- add zoom support by handling submats
 *           6/12/95 -- changed bitmap epsilon test to look at BITMAP_EPSILON
 *
 */

/*
 * Function: CalculateScreenSize
 *
 * Purpose: to calculate the screen size based on the option chosen
 *          by conformal and the aspect ratios of the input and
 *          output matrices
 *
 * Assumptions: 1. all row and column values are > 0
 *              2. out_cols/out_rows already contain the screen size values
 *
 * Algorithm: if conforming to screen
 *                do nothing
 *            if conforming to matrix
 *                let in_aspect_ratio = (double) in_cols / (double) in_rows
 *                let out_aspect_ratio = (double) out_cols / (double) out_rows
 *                if (aspect ratios are equal)
 *                   do nothing
 *                if (in aspect ratio > out aspect ratio)
 *                then
 *                     since we don't have enough cols in the output matrix
 *                     conform as best we can:
 *                        out_rows = (int) max((1.0/in_aspect_ratio*out_cols),
 *                                             1.0))
 *                else
 *                        out_cols = (int) max(in_aspect_ratio * out_rows,
 *                                             1.0)
 *                endif
 *            endif
 * 
 * Parameters: in_rows, in_cols - the row and column sizes, respectively,
 *                                of the input matrix
 *             out_rows, out_cols - the row and column sizes, respectively,
 *                                  of the output matrix
 *             conformal - the conformality requested 
 *
 * Return Values:  none
 * 
 * Calls: none
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: 2/13/94
 *
 * Modified:  9/23/94 -- assure changed value is no bigger than init value
 *
 */
void CalculateScreenSize( int in_rows, int in_cols,
			  int *out_rows, int *out_cols, 
			  int conformal)
{
    double in_aspect_ratio;
    double out_aspect_ratio;
    int orig_out_rows, orig_out_cols;
    char debugmsg[100];

    /* orig_out_rows = *out_rows;*/
    /* orig_out_cols = *out_cols; */
    in_aspect_ratio = (double) in_cols / (double) in_rows;

    out_aspect_ratio = (double) *out_cols / (double) *out_rows;

    switch (conformal)
    {
    case CONFORMAL_TO_SCREEN:
	/* nothing to do */
	break;
    case CONFORMAL_TO_MATRIX:
	
	/* update out_rows and out_cols according to aspect_ratios */

	/* if ratios not equal */
	if (fabs(in_aspect_ratio - out_aspect_ratio) > EPSILON)
	{
	    /* if input has "too many" columns, adjust rows */
	    if (in_aspect_ratio > out_aspect_ratio)
	    {
		*out_rows = (int)mymax((1.0/in_aspect_ratio*(*out_cols)), 1.0);
		/* *out_rows = mymin((*out_rows),orig_out_rows); */
	    }
	    else /* input has too many rows, adjust columns */
	    {
		*out_cols = (int)mymax(in_aspect_ratio * (*out_rows), 1.0);
		/* *out_cols = mymin((*out_cols),orig_out_cols); */
	    }
	}
	break;
    default: /* do nothing */
	break;
    } /* end of switch */

    sprintf(debugmsg,"trying to pour a %d x %d matrix into a %d x %d pot\n",
	   in_rows, in_cols, *out_rows, *out_cols);
    DEBUG3(debugmsg);
}

/*
 * Function: CompressMat
 *
 * Purpose: to compress a matrix into a dense matrix of values suitable
 *          for eventual scaling and displaying
 *
 * Algorithm: 1. malloc out dense matrix
 *            2. if ok, call correct routine depending on if matrix
 *               is sparse or dense
 * 
 * Parameters: M - the matrix to be compressed
 *             comp - the output matrix
 *             c - a compression request
 *
 * Return Values:  OK - all went well
 *                !OK - guess it didn't
 * 
 * Calls: CompressSparseMat CompressDenseMat error GetDenseMatrix
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: 2/13/94
 *
 * Modified: 2/15/94
 *           3/19/94 -- added selected_rows/cols for submat support.
 *
 */

int CompressMat( DaMatrix *M, DenseMatrix **comp, CompressionRequest *c )
{
    int row_major; /* need row major arrays below */
    int nrows, ncols, in_nrows, in_ncols, one_d, rv;
    int selected_rows, selected_cols;
    char debugstr[100];
    double matrix_initval = 0.0; /* init matrix to 0's */

    TRACE("Entered CompressMat\n");

    /* if input matrix isn't HB or Dense, not supported */
    if ((M->type != HB) && (M->type != Dense))
    {
	error("Only sparse and dense matrices supported",UNSUPERR);
	/* shouldn't get here */
	return(BAD);
    }

    DEBUG2("In CompressMat 1\n");
    /* calculate screen storage */
    /* first, check input values */
    nrows = c->max_y;
    ncols = c->max_x;
    if ((nrows <= 0) || (ncols <= 0))
    {
	error("bad screen size input to CompressMat",ARGERR);
	/* shouldn't get here */
	return(BAD);
    }

    DEBUG2("In CompressMat 2\n");
    /* check nrows and cols from input matrix */
    if (M->type == HB)
    {
	in_nrows = M->hb_NumRows;
	in_ncols = M->hb_NumCols;
    }
    else /* it's dense */
    {
	in_nrows = M->d_NumRows;
	in_ncols = M->d_NumCols;
    }
    if ((in_nrows <= 0) || (in_ncols <= 0))
    {
	error("bad input matrix size to CompressMat",ARGERR);
	/* shouldn't get here */
	return(BAD);
    }
    /* +1 added 4/13/94 */
    selected_rows = IDXRANGE(c->input_endrow, c->input_startrow);
    selected_cols = IDXRANGE(c->input_endcol, c->input_startcol);
    sprintf(debugstr,"CompressMat: sel_row = %d, sel_col = %d\n",
	    selected_rows,selected_cols);
    DEBUG2(debugstr);
    sprintf(debugstr,"startrow = %d, endrow = %d, startcol = %d, endcol= %d\n",
	    c->input_startrow, c->input_endrow, c->input_startcol, c->input_endcol);
    DEBUG2(debugstr);

    if ((selected_rows <= 0) || (selected_rows > in_nrows))
    {
	error("CompressMat: too many or not enough rows selected", ARGERR);
	return(BAD);
    }
    if ((selected_cols <= 0) || (selected_cols > in_ncols))
    {
	error("CompressMat: too many or not enough columns selected", ARGERR);
	return(BAD);
    }
    /* figure out screen size to allocate */
    
    sprintf(debugstr,"CompressMat pre-CSS: trying to compress %d x %d matrix into  %d x %d\n",
	   selected_rows, selected_cols, nrows, ncols);
    DEBUG1(debugstr);

    CalculateScreenSize( selected_rows, selected_cols, &nrows, &ncols, c->conformal);

    sprintf(debugstr,"after CSS: trying to pour a %d x %d matrix into a %d x %d pot\n",
	   selected_rows, selected_cols, nrows, ncols);
/* fprintf(stderr,"%s",debugstr); */
    DEBUG1(debugstr);
    
    /* set initial input_min/max values */
    c->input_max = c->input_absmax = -HUGE_VAL;
    c->input_min = c->input_absmin = HUGE_VAL;

    /* try to get dense matrix for output */
    one_d = 1; /* use one_d storage */
    row_major = 1; /* and use row major organization */
    if ((*comp = GetDenseMatrix( nrows, ncols, row_major, one_d, matrix_initval )) ==
	(DenseMatrix *) NULL)
    {
	error("Unable to get dense matrix",MEMERR);
	/* shouldn't get here */
	return(BAD);
    }
    
    DEBUG1("and awaaaaaay we go! (CompressMat calling compression routine)\n");

    /* call correct routine */
    c->bitmap = 1; /* set default value 4/20/94 */
    if (M->type == HB)
    {
	rv = CompressSparseMat(M->data.hb, *comp, c);
    }
    else
    {
	/* not sparse, so do the dense thing */
	sprintf(debugstr,"calling CDM, input one_d = %d\n", M->data.d->one_d);
	DEBUG1(debugstr);

	rv = CompressDenseMat(M->data.d, *comp, c );
    }

    /* if all went well, find min and max values */
    if (rv == OK)
    {
	/* if a pattern matrix, set min = 0 and max = 1 */
	if ((M->type == HB) && (M->data.hb->head.NumericalData == IsPattern))
	{
	    c->mat_min = c->mat_absmin = 0;
	    c->mat_max = c->mat_absmax = 1;
	}
	else
	{
	    GetMinMax(*comp, c);
	}
    }
    return(rv);
}


/*
 * Function: CompressSparseMat
 *
 * Purpose: to compress a sparse matrix into a dense matrix of values suitable
 *          for eventual scaling and displaying
 *
 * Parameters: hb - the sparse matrix to be compressed
 *             comp - the output matrix
 *             c - a compression request
 *
 * Return Values:  OK - all went well
  *                !OK - guess it didn't
  * 
  * Calls: sprintf error ...
  *
  * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
  *             tloos@cs.indiana.edu
  *
  * Started: 2/15/94
  *
  * Modified: 2/16 -- hopefully fix use_c bug
  *           2/28 -- fix min/max calculation
  *                -- added flip_rows
  *                -- added col major input (randy_order)
  *           3/18 -- deleted flip_rows and changed randy_order to sp_col_major
  *           3/19 -- added support for compression of an input submatrix
  *           4/13 -- check for MaxCompMatIdx
  *           4/18 -- rewrite using centered pixels.
  *           4/20 -- check for input bitmap
  *           5/14 -- huge rewrite for speed -- MUCH DUPLICATED CODE!!!
  *           6/1  -- fix last column/row problems
  *           6/3  -- calculate input max/min values
  *           10/13 -- fix bug found with SGI matlab that made non-bitmap matrices look
  *                    like bitmaps (init othervalue to 0)
  *
  */
 int CompressSparseMat( HBMatrix *hb, DenseMatrix *comp, CompressionRequest *c)
 {
     double value, othervalue, absvalue;
     char errmsg[100];
     double RowWeight, ColWeight, ColWeightFrac;
     double leftwt, rightwt, rowwt[2], colwt[2], topwt, botwt, parmval;
     double ftop, fbottom, fleft, fright;

     double input_min, input_max, input_absmin, input_absmax;
     int input_min_row, input_min_col, input_max_row, input_max_col;
     int input_absmin_row, input_absmin_col, input_absmax_row, 
	 input_absmax_col;

     int newmax, newmin;
     int InStartCol, InEndCol, InStartRow, InEndRow;
     int irow, icol, orow, ocol, orowidx, ocolidx,  idx;
     int left, right, top, bottom, rownum, colnum;
     int InNumCols, InNumRows, OutNumCols, OutNumRows;

     int minrowidx, maxrowidx,  mincolidx, maxcolidx;
     int other_val_init = 0;
     int index_offset, two_rows, two_cols, notlastcol, false_bottom;
     CompressionStrategyParameters parms;

     /* check for good matrix sizes (i.e. number of rows and cols all > 0) */

     TRACE("entered CompressSparseMat (CSM)\n");
     /* changed 3/19 */

     /* InNumCols = hb->head.NumCols; */
     /* InNumRows = hb->head.NumRows; */

     InStartCol = c->input_startcol;
     InEndCol = c->input_endcol;
     InStartRow = c->input_startrow;
     InEndRow = c->input_endrow;

     InNumCols = IDXRANGE(InEndCol,InStartCol);
     InNumRows = IDXRANGE(InEndRow,InStartRow);
     OutNumCols = comp->NumCols;
     OutNumRows = comp->NumRows;

     /* 10/13 -- init othervalue! */
     othervalue = 0.0;

     if ((InNumRows <= 0) || (InNumCols <= 0) || 
	 (InNumCols > hb->head.NumCols) || (InNumRows > hb->head.NumRows))
     {
	 sprintf(errmsg,
		 "CompressSparseMat:input row count (%d) or col count (%d) wrong",
		 hb->head.NumRows, hb->head.NumCols);
	 error(errmsg, ARGERR);
	 /* shouldn't get here */
	 return(BAD);
     }

     if ((OutNumRows <= 0) || (OutNumCols <= 0))
     {
	 sprintf(errmsg,
		 "CompressSparseMat: output row (%d) or col count (%d) < 0",
		 comp->NumRows, comp->NumCols);
	 error(errmsg, ARGERR);
	 /* shouldn't get here */
	 return(BAD);
     }
     /*
      * calculate weights:
      * A weight is the number of output rows/cols that correspond
      *  to an input row or column.
      *
      * For compression, expect weights to be  < 1
      * For expansion, expect weights to be > 1
      */
     if (c->sp_col_major)
     {
	 RowWeight = (double) OutNumRows / (double) InNumCols;
	 ColWeight = (double) OutNumCols / (double) InNumRows;
     }
     else
     {
	 RowWeight = (double) OutNumRows / (double) InNumRows;
	 ColWeight = (double) OutNumCols / (double) InNumCols;
     }
     ColWeightFrac = ColWeight - (int) ColWeight;
     sprintf(errmsg,"CompressSparseMatrix: OutNumRows = %d, OutNumCols = %d ",OutNumRows,OutNumCols);
     DEBUG2(errmsg);
     sprintf(errmsg,"CompressSparseMatrix: InNumRows = %d, InNumCols = %d\n",InNumRows,InNumCols);
     DEBUG2(errmsg);

     if (c->use_c)
	 index_offset = 0;
     else /* FORTRAN -- yuck */
	 index_offset = 1;

     /* This looks backwards, but this is the intent:
	assign the initial input minimum value to the global input_max
	value and set the input max value to the global input_min value
	for both actual and abs. values */
     
     /* 
      * commented out on 9/1/94
      *
      * input_min = hb->head.MaxValue;
      * input_max = hb->head.MinValue;
      * input_absmin = hb->head.AbsMaxValue;
      * input_absmax = hb->head.AbsMinValue;
      */

     /* 9/1/94 replace min/max switches with MAX/MIN_SCALING_VALUE */

     input_min = input_absmin =  MAX_SCALING_VALUE;
     input_max = input_absmax =  MIN_SCALING_VALUE;
     input_min_col = input_max_col = input_absmin_col = input_absmax_col = 
	 InStartCol;
     input_min_row = input_max_row = input_absmin_row = input_absmax_row = 
	 InStartRow;

     /* fprintf(stderr,"input_absmin = %e\n",input_absmin);*/

/*      
 * fprintf(stderr,"CSM:  input_min = %e, input_max = %e, input_absmin = %e, input_absmax = %e\n",input_min, input_max, input_absmin, input_absmax);
 * fprintf(stderr,"CSM (TEST): input globmin = %e, input globmax = %e, input globabsmin = %e, input globabmax = %e\n",
 *	    hb->head.MinValue, hb->head.MaxValue,
 *	    hb->head.AbsMinValue, hb->head.AbsMaxValue);
 */

     if (!c->sp_col_major)
     {
	 if (RowWeight < 1.0) /* if compressing rows */
	 {
	     if (ColWeight < 1.0) /* if compressing columns too */
	     {
		 for (icol = InStartCol; icol <= InEndCol; icol++)
		 {
		     /* get current row indices */
		     /* if (c->use_c)
		      *{
		      *	minrowidx = hb->col_inds[icol];
		      *	maxrowidx = hb->col_inds[icol+1];
		      * }
		      *else
		      * {
		      *	minrowidx = hb->col_inds[icol]-1;
		      *	maxrowidx = hb->col_inds[icol+1]-1;
		      * }
		      */
		     minrowidx = hb->col_inds[icol] - index_offset;
		     maxrowidx = hb->col_inds[icol+1] - index_offset;
		     /* calculate left and right most output columns 
			affected by this input column  */
		     fleft = (icol-InStartCol) * ColWeight;
		     fright = fleft + ColWeight;
		     left = (int) fleft;
		     right = (int) fright;
		     two_cols = (left != right) && (right < OutNumCols);
		     colwt[1] = two_cols * (fright - right);
		     colwt[0] = ColWeight - colwt[1];
 /*
  * fprintf(stderr,"!sp_col_major with compression icol = %d, minr = %d, maxr = %d\n",icol,minrowidx, maxrowidx);
  * fprintf(stderr,"   2 cols = %d, colwt0 = %f, colwt1 = %f\n",two_cols,colwt[0],colwt[1]);
  */
		     /* for the useful row numbers under the current column */
		     for (irow = minrowidx; irow < maxrowidx; irow++)
		     {
			 /* if (c->use_c)
			  *  rownum = hb->row_inds[irow];
			  * else
			  *   rownum = hb->row_inds[irow]-1;
			  */
			 rownum = hb->row_inds[irow] - index_offset;

			 /* added 3/19/94 -- support for submats */
			 if ((rownum < InStartRow) || (rownum > InEndRow))
			     continue;
			 /* end 3/19/94 */

			 /* get value */
			 /* 4/9/94 -- support for pattern matrices */
			 if (hb->head.NumericalData == IsPattern)
			     value = DEFAULT_PATTERN_VALUE;
			 else
			 {
			     value = hb->values[irow];
			     /* bitmap checks 4/20 */
			     /* 4/28 -- fixed bug since myabs is a MACRO */
			     if ((c->bitmap) && (myabs(value) > BITMAP_EPSILON) && 
				 (myabs((value-othervalue))>BITMAP_EPSILON))
			     {
				 if (other_val_init)
				     c->bitmap = 0;
				 else
				 {
				     other_val_init = 1;
				     othervalue = value;
				 }
			     }
			 } /* end of else not pattern */
			 /* do max/min calculations of the input */
			 absvalue = (value < 0.0) * -value + 
			            (value >=0.0) * value;
			 newmax = (value > input_max);
			 newmin = (value < input_min);
			 input_max = newmax * value + (1-newmax)*input_max;
			 input_min = newmin * value + (1-newmin)*input_min;

			 input_max_col = newmax * icol +
			     (1-newmax) * input_max_col;
			 input_max_row = newmax * rownum +
			     (1-newmax) * input_max_row;

			 input_min_col = newmin * icol +
			     (1-newmin) * input_min_col;
			 input_min_row = newmin * rownum +
			     (1-newmin) * input_min_row;
			     
/*			 fprintf(stderr,"CSM: newmax1 = %d, newmin1 = %d\n",
				 newmax,newmin);
*/

			 newmax = (absvalue > input_absmax);
			 /* 9/14  change abs newmin to skip value = 0.0 */
			 newmin = ((absvalue < input_absmin) && 
				   (absvalue != 0.0));
			 input_absmax = newmax * absvalue + 
			     (1-newmax)* input_absmax;
			 input_absmin = newmin * absvalue + 
			     (1-newmin)*input_absmin;

			 input_absmax_col = newmax * icol +
			     (1-newmax) * input_absmax_col;
			 input_absmax_row = newmax * rownum +
			     (1-newmax) * input_absmax_row;

			 input_absmin_col = newmin * icol +
			     (1-newmin) * input_absmin_col;
			 input_absmin_row = newmin * rownum +
			     (1-newmin) * input_absmin_row;
 
/* if (newmin)
 * {
 *			 fprintf(stderr,"CSM: newmax = %d, newmin = %d\n",
 *				 newmax,newmin);
 *   fprintf(stderr,"CSM: value = %f, absvalue = %f, input_min = %f, input_max = %f, input_absmin = %e, input_absmax = %f\n",value,absvalue,input_min, input_max, input_absmin, input_absmax);
 * }
 */
			 /* since RW < 1, one value covers at most only 2
			    output rows */
			 /* ftop = certain output row -- fbottom only if input
			    straddles 2 pixels */
			 ftop = (rownum-InStartRow) * RowWeight;
			 fbottom = ftop+RowWeight;

			 /* need array indices for fleft and fright */
			 top = (int) ftop;
			 bottom = (int) fbottom;

			 /* two_rows = 1 if there are 2 rows (bottom != top) */
			 /*              and if the bottom row < OutNumRows */
			 /*          = 0 if there's only 1 (bottom == top) */
			 two_rows = ((bottom != top) && (bottom < OutNumRows));

			 /* get weights for the 2 output rows */
			 /* wt[1] = weight for second pixel:
			  *       = 0 if integer row numbers are the same
			  *       = fp(fbottom) if different 
			  * use two_rows to figure out if there are 2 rows;
			  *  fp(fbottom) = fbottom - bottom
			  *  so wt[1] is:
			  *   two_rows * (fbottom-bottom)
			  *
			  * wt[0] = RowWeight - wt[1];
			  */
			 rowwt[1] = two_rows *  (fbottom-bottom);
			 rowwt[0] = RowWeight - rowwt[1];
 /* fprintf(stderr,"     2 rows = %d, top = %d, bottom = %d, rowwt = %f, rowwt1 = %f\n",two_rows,top,bottom,rowwt[0],rowwt[1]); */

			 for (orowidx = 0; orowidx <= two_rows; orowidx++)
			 {
			     /* calculate actual row number */
			     /* orow = top + orowidx */
			     idx = (top + orowidx) * OutNumCols + left;

			     /* parmval is the part of the value that's
				invariant below */
			     parmval = value * rowwt[orowidx];
			     for (ocolidx = 0; ocolidx <= two_cols; ocolidx++)
			     {
				 /* update the desired output entry */
				 /* BTW: ocol = ocolidx + left */
				 parms.old = comp->data.d1[idx];
				 parms.newp = parmval * colwt[ocolidx];
				 comp->data.d1[idx] = (c->strat)(&parms);
				 idx++; /* for next go round */
			     } /* end of for ocolidx */
			 } /* end of for orowidx */
		     } /* end of for irow */
		 } /* end of for icol */
	     } /* of if ColWeight < 1.0 */
	     else /* ColWeight > 1.0 */
	     {
		 for (icol = InStartCol; icol <= InEndCol; icol++)
		 {

		     /* get current row indices */
		     /* if (c->use_c)
		      *{
		      *	minrowidx = hb->col_inds[icol];
		      *	maxrowidx = hb->col_inds[icol+1];
		      * }
		      *else
		      * {
		      *	minrowidx = hb->col_inds[icol]-1;
		      *	maxrowidx = hb->col_inds[icol+1]-1;
		      * }
		      */
		     minrowidx = hb->col_inds[icol] - index_offset;
		     maxrowidx = hb->col_inds[icol+1] - index_offset;
		     /* calculate left and right most output columns 
			affected by this input column  */
		     fleft = (icol-InStartCol) * ColWeight;
		     fright = fleft + ColWeight;
		     left = (int) fleft;
		     right = (int) fright;
		     notlastcol = (right < OutNumCols);
		     /* know that the cols are different so no need to check */
		     leftwt = (1.0 - (fleft - left));
		     rightwt = (fright-right);

		     /* for the useful row numbers under the current column */
		     for (irow = minrowidx; irow < maxrowidx; irow++)
		     {
			 /* if (c->use_c)
			  *  rownum = hb->row_inds[irow];
			  * else
			  *   rownum = hb->row_inds[irow]-1;
			  */
			 rownum = hb->row_inds[irow] - index_offset;

			 /* added 3/19/94 -- support for submats */
			 if ((rownum < InStartRow) || (rownum > InEndRow))
			     continue;
			 /* end 3/19/94 */

			 /* get value */
			 /* 4/9/94 -- support for pattern matrices */
			 if (hb->head.NumericalData == IsPattern)
			     value = DEFAULT_PATTERN_VALUE;
			 else
			 {
			     value = hb->values[irow];
			     /* bitmap checks 4/20 */
			     /* 4/28 -- fixed bug since myabs is a MACRO */
			     if ((c->bitmap) && (myabs(value) > BITMAP_EPSILON) && 
				 (myabs((value-othervalue))>BITMAP_EPSILON))
			     {
				 if (other_val_init)
				     c->bitmap = 0;
				 else
				 {
				     other_val_init = 1;
				     othervalue = value;
				 }
			     }
			 } /* end of else not pattern */

			 /* do max/min calculations of the input */
			 absvalue = (value < 0.0) * -value + 
			            (value >=0.0) * value;
			 newmax = (value > input_max);
			 newmin = (value < input_min);
			 input_max = newmax * value + (1-newmax)*input_max;
			 input_min = newmin * value + (1-newmin)*input_min;

			 input_max_col = newmax * icol + 
			     (1-newmax) * input_max_col;
			 input_max_row = newmax * rownum + 
			     (1-newmax) * input_max_row;

			 input_min_col = newmin * icol + 
			     (1-newmin) * input_min_col;
			 input_min_row = newmin * rownum + 
			     (1-newmin) * input_min_row;

			 newmax = (absvalue > input_absmax);
			 /* 9/14  change abs newmin to skip value = 0.0 */
			 newmin = ((absvalue < input_absmin) && 
				   (absvalue != 0.0));
			 input_absmax = newmax * absvalue + 
			     (1-newmax)* input_absmax;
			 input_absmin = newmin * absvalue + 
			     (1-newmin)*input_absmin;

			 input_absmax_col = newmax * icol +
			     (1-newmax) * input_absmax_col;
			 input_absmax_row = newmax * rownum +
			     (1-newmax) * input_absmax_row;

			 input_absmin_col = newmin * icol +
			     (1-newmin) * input_absmin_col;
			 input_absmin_row = newmin * rownum +
			     (1-newmin) * input_absmin_row;
/*
			 fprintf(stderr,"CSM: newmax = %d, newmin = %d\n",
				 newmax,newmin);
    fprintf(stderr,"CSM: value = %f, absvalue = %f, input_min = %f, input_max = %f, input_absmin = %f, input_absmax = %f\n",value,absvalue,input_min, input_max, input_absmin, input_absmax);
*/

			 /* since RW < 1, one input value covers at most only 2
			    output rows */

			 /* ftop = certain output row -- fbottom only if input
			    straddles 2 pixels */
			 ftop = (rownum-InStartRow) * RowWeight;
			 fbottom = ftop+RowWeight;

			 /* need array indices for fleft and fright */
			 top = (int) ftop;
			 bottom = (int) fbottom;

			 /* two_rows = 1 if there are 2 rows (bottom != top) */
			 /*          = 0 if there's only 1 (bottom == top) */
			 two_rows = ((bottom != top) && (bottom != OutNumRows));


			 /* get weights for the 2 output rows */
			 /* wt[1] = weight for second pixel:
			  *       = 0 if integer row numbers are the same
			  *       = fp(fbottom) if different 
			  * use two_rows to figure out if there are 2 rows;
			  * fp(fbottom) = fbottom - bottom
			  *  so wt2 is:
			  *   two_rows  * (fbottom-bottom)
			  *
			  * wt[0] = RowWeight - wt[1];
			  */
			 rowwt[1] = two_rows * (fbottom-bottom);
			 rowwt[0] = RowWeight - rowwt[1];

			 for (orowidx = 0; orowidx <= two_rows; orowidx++)
			 {
			     /* calculate actual row number */
			     orow = orowidx + top;
			     /* parmval is the part of the value that's
				invariant below */
			     parmval = value * rowwt[orowidx];

			     /* first on top row */
			     /* left corner */
			     idx = orow * OutNumCols + left;
			     parms.old = comp->data.d1[idx];
			     parms.newp = parmval * leftwt;
			     comp->data.d1[idx] = (c->strat)(&parms);

			     /* then across the top row */
			     for (ocol = left+1; ocol < right; ocol++)
			     {
				 idx++;
				 parms.old = comp->data.d1[idx];
				 parms.newp = parmval;
				 comp->data.d1[idx] = (c->strat)(&parms);
			     }

			     /* now UR corner */
			     idx+= notlastcol; /* can do this since ColWeight > 1, so
				       left != right */
			     parms.old = comp->data.d1[idx];
			     parms.newp = parmval * rightwt;
			     comp->data.d1[idx] = (c->strat)(&parms);
			 } /* end of for orowidx */
		     } /* end of for irow */
		 } /* end of for icol */
	     } /* of if colweight > 1.0 */
	 } /* of if RowWeight < 1.0 */
	 else /* RowWeight >= 1.0 */ 
	 {
	     for (icol = InStartCol; icol <= InEndCol; icol++)
	     {

		 /* get current row indices */
		 minrowidx = hb->col_inds[icol] - index_offset;
		 maxrowidx = hb->col_inds[icol+1] - index_offset;

		 /* calculate left and right most output columns 
		    affected by this input column  */
		 fleft = (icol-InStartCol) * ColWeight;
		 fright = fleft + ColWeight;
		 left = (int) fleft;
		 right = (int) fright;
		 two_cols = (left != right) && (right < OutNumCols);
		 leftwt = two_cols * (1.0 - (fleft - left)) +
		      ((1-two_cols) * ColWeightFrac);
		 rightwt = (two_cols * (fright-right));

		 /* for the useful row numbers under the current column */
		 for (irow = minrowidx; irow < maxrowidx; irow++)
		 {
		     /* if (c->use_c)
		      *  rownum = hb->row_inds[irow];
		      * else
		      *   rownum = hb->row_inds[irow]-1;
		      */
		     rownum = hb->row_inds[irow] - index_offset;

		     /* added 3/19/94 -- support for submats */
		     if ((rownum < InStartRow) || (rownum > InEndRow))
			     continue;
		     /* end 3/19/94 */

		     /* get value */
		     /* 4/9/94 -- support for pattern matrices */
		     if (hb->head.NumericalData == IsPattern)
			 value = DEFAULT_PATTERN_VALUE;
		     else
		     {
			 value = hb->values[irow];
			 /* bitmap checks 4/20 */
			 /* 4/28 -- fixed bug since myabs is a MACRO */
			 if ((c->bitmap) && (myabs(value) > BITMAP_EPSILON) && 
			     (myabs((value-othervalue))>BITMAP_EPSILON))
			 {
			     if (other_val_init)
				 c->bitmap = 0;
			     else
			     {
				 other_val_init = 1;
				 othervalue = value;
			     }
			 }
		     } /* end of else not pattern */
		     /* do max/min calculations of the input */
		     absvalue = (value < 0.0) * -value + 
			 (value >=0.0) * value;
		     newmax = (value > input_max);
		     newmin = (value < input_min);
		     input_max = newmax * value + (1-newmax)*input_max;
		     input_min = newmin * value + (1-newmin)*input_min;
/* fprintf(stderr,"newmax = %d, value = %f, (1-newmax) = %d, input_max = %f\n",
	newmax,value, (1-newmax),input_max);
*/		     
		     input_max_col = newmax * icol + 
			 (1-newmax) * input_max_col;
		     input_max_row = newmax * rownum + 
			 (1-newmax) * input_max_row;

		     input_min_col = newmin * icol + 
			 (1-newmin) * input_min_col;
		     input_min_row = newmin * rownum + 
			 (1-newmin) * input_min_row;
		     newmax = (absvalue > input_absmax);
		     /* 9/14  change abs newmin to skip value = 0.0 */
		     newmin = ((absvalue < input_absmin) && 
			       (absvalue != 0.0));
		     input_absmax = newmax * absvalue + 
			 (1-newmax)* input_absmax;
		     input_absmin = newmin * absvalue + 
			 (1-newmin)*input_absmin;

		     input_absmax_col = newmax * icol +
			 (1-newmax) * input_absmax_col;
		     input_absmax_row = newmax * rownum +
			 (1-newmax) * input_absmax_row;

		     input_absmin_col = newmin * icol +
			 (1-newmin) * input_absmin_col;
		     input_absmin_row = newmin * rownum +
			 (1-newmin) * input_absmin_row;
/*
 fprintf(stderr,"CSM: newmax = %d, newmin = %d\n", newmax,newmin);
 fprintf(stderr,"CSM: value = %f, absvalue = %f, input_min = %f, input_max = %f, input_absmin = %f, input_absmax = %f\n",value,absvalue,input_min, input_max, input_absmin, input_absmax);
 fprintf(stderr,"CSM: min row, col = (%d, %d), max row, col = (%d, %d)\n",input_min_row, input_min_col, input_max_row, input_max_col);
 fprintf(stderr,"CSM: absmin row, col = (%d, %d), absmax row, col = (%d, %d)\n",input_absmin_row, input_absmin_col, input_absmax_row, input_absmax_col);
*/

		     /* calculate top and bottom row and their weights */
		     ftop = (rownum-InStartRow) * RowWeight;
		     fbottom = ftop + RowWeight;
		     top = (int) ftop;
		     bottom = (int) fbottom;
		     false_bottom = bottom - (bottom == OutNumRows);
		     topwt = 1.0 - (ftop - top);
		     botwt = fbottom - bottom;

		     /* do top row */
		     parmval = value * topwt;

		     /* first UL (upper left)  corner */
		     idx = top * OutNumCols + left;
		     parms.old = comp->data.d1[idx];
		     parms.newp = parmval * leftwt;
		     comp->data.d1[idx] = (c->strat)(&parms);

		     /* then across the top row */
		     for (ocol = left+1; ocol < right; ocol++)
		     {
			 idx++;
			 parms.old = comp->data.d1[idx];
			 parms.newp = parmval;
			 comp->data.d1[idx] = (c->strat)(&parms);
		     }

		     /* now UR corner */
		     idx += two_cols; /* this should work */
		     parms.old = comp->data.d1[idx];
		     parms.newp = value * topwt * rightwt;
		     comp->data.d1[idx] = (c->strat)(&parms);

		     /* now on bottom row */
		     parmval = value * botwt;

		     /* left corner */
		     idx = false_bottom * OutNumCols + left;
		     parms.old = comp->data.d1[idx];
		     parms.newp = parmval * leftwt;
		     comp->data.d1[idx] = (c->strat)(&parms);

		     /* then across the bottom row */
		     for (ocol = left+1; ocol < right; ocol++)
		     {
			 idx++;
			 parms.old = comp->data.d1[idx];
			 parms.newp = parmval;
			 comp->data.d1[idx] = (c->strat)(&parms);
		     }

		     /* now LR corner */
		     idx += two_cols;
		     parms.old = comp->data.d1[idx];
		     parms.newp = parmval * rightwt;
		     comp->data.d1[idx] = (c->strat)(&parms);

		     /* left and right columns */
		     for (orow = top+1; orow < bottom; orow++)
		     {
			 idx = orow * OutNumCols + left;
			 parms.old = comp->data.d1[idx];
			 parms.newp = value * leftwt;
			 comp->data.d1[idx] = (c->strat)(&parms);

			 idx = orow * OutNumCols + right;
			 parms.old = comp->data.d1[idx];
			 parms.newp = value * rightwt;
			 comp->data.d1[idx] = (c->strat)(&parms);
		     }

		     /* finally the center of the area covered */
		     for (orow = top+1; orow < bottom; orow++)
		     {
			 idx = orow * OutNumCols + left;
			 for (ocol = left+1; ocol < right; ocol++)
			 {
			     idx++;
			     parms.old = comp->data.d1[idx];
			     parms.newp = value;
			     comp->data.d1[idx] = (c->strat)(&parms);
			 } /* of for ocol */
		     } /* of for orow */
		 } /* end of for irow */
	     } /* end of for icol */
	 } /* end of if RowWeight > 1 */
     } /* end of if !sp_col_major */
     else /* sp_col_major ON THE INPUT! */
     {
	 if (RowWeight < 1.0)
	 {
	     if (ColWeight < 1.0)
	     {
		 for (irow = InStartCol; irow <= InEndCol; irow++)
		 {

 /* fprintf(stderr,"icol = %d\n",icol); */

		     /* get current row indices */
		     /* if (c->use_c)
		      *{
		      *	minrowidx = hb->col_inds[icol];
		      *	maxrowidx = hb->col_inds[icol+1];
		      * }
		      *else
		      * {
		      *	minrowidx = hb->col_inds[icol]-1;
		      *	maxrowidx = hb->col_inds[icol+1]-1;
		      * }
		      */
		     mincolidx = hb->col_inds[icol] - index_offset;
		     maxcolidx = hb->col_inds[icol+1] - index_offset;

		     /* calculate left and right most output columns 
			affected by this input column  */
		     ftop = (irow-InStartCol) * RowWeight;
		     fbottom = ftop + RowWeight;
		     top = (int) ftop;
		     bottom = (int) fbottom;
		     /* two_rows = (bottom != top);*/
		     two_rows = ((bottom != top) && (bottom != OutNumRows));
		     rowwt[1] = two_rows *  (fbottom-bottom);
		     rowwt[0] = RowWeight - rowwt[1];

		     /* for the useful row numbers under the current column */
		     for (icol = mincolidx; icol < maxcolidx; icol++)
		     {
			 colnum = hb->row_inds[icol] - index_offset;

			 /* added 3/19/94 -- support for submats */
			 if ((colnum < InStartRow) || (colnum > InEndRow))
			     continue;
			 /* end 3/19/94 */

			 /* get value */
			 /* 4/9/94 -- support for pattern matrices */
			 if (hb->head.NumericalData == IsPattern)
			     value = DEFAULT_PATTERN_VALUE;
			 else
			 {
			     value = hb->values[icol];
			     /* bitmap checks 4/20 */
			     /* 4/28 -- fixed bug since myabs is a MACRO */
			     if ((c->bitmap) && (myabs(value) > BITMAP_EPSILON) && 
				 (myabs((value-othervalue))>BITMAP_EPSILON))
			     {
				 if (other_val_init)
				     c->bitmap = 0;
				 else
				 {
				     other_val_init = 1;
				     othervalue = value;
				 }
			     }
			 } /* end of else not pattern */
			 /* since CW > 1, one input value covers more than 2
			    output cols */

			 /* do max/min calculations of the input */
			 absvalue = (value < 0.0) * -value + 
			     (value >=0.0) * value;
			 newmax = (value > input_max);
			 newmin = (value < input_min);
			 input_max = newmax * value + (1-newmax)*input_max;
			 input_min = newmin * value + (1-newmin)*input_min;
/* fprintf(stderr,"newmax = %d, value = %f, (1-newmax) = %d, input_max = %f\n",
	newmax,value, (1-newmax),input_max);
*/		     
			 input_max_col = newmax * icol + 
			     (1-newmax) * input_max_col;
			 input_max_row = newmax * rownum + 
			     (1-newmax) * input_max_row;

			 input_min_col = newmin * icol + 
			     (1-newmin) * input_min_col;
			 input_min_row = newmin * rownum + 
			     (1-newmin) * input_min_row;

			 newmax = (absvalue > input_absmax);
		         /* 9/14  change abs newmin to skip value = 0.0 */
		         newmin = ((absvalue < input_absmin) && 
			           (absvalue != 0.0));
			 input_absmax = newmax * absvalue + 
			     (1-newmax)* input_absmax;
			 input_absmin = newmin * absvalue + 
			     (1-newmin)*input_absmin;

		         input_absmax_col = newmax * icol +
			     (1-newmax) * input_absmax_col;
		         input_absmax_row = newmax * rownum +
			     (1-newmax) * input_absmax_row;

		         input_absmin_col = newmin * icol +
			     (1-newmin) * input_absmin_col;
		         input_absmin_row = newmin * rownum +
			     (1-newmin) * input_absmin_row;
		     
/*
			 fprintf(stderr,"CSM: newmax = %d, newmin = %d\n",
				 newmax,newmin);
    fprintf(stderr,"CSM: value = %f, absvalue = %f, input_min = %f, input_max = %f, input_absmin = %f, input_absmax = %f\n",value,absvalue,input_min, input_max, input_absmin, input_absmax);
*/

			 fleft = (colnum-InStartRow) * ColWeight;
			 fright = fleft+ColWeight;

			 /* need array indices for fleft and fright */
			 left = (int) fleft;
			 right = (int) fright;
			 two_cols = (left != right) && (right < OutNumCols);
			 colwt[1] = two_cols * (fright - right);
			 colwt[0] = ColWeight - colwt[1];

			 for (orowidx = 0; orowidx <= two_rows; orowidx++)
			 {
			     /* calculate actual row number */
			     /* orow = top + orowidx */
			     idx = (top + orowidx) * OutNumCols + left;

			     /* parmval is the part of the value that's
				invariant below */
			     parmval = value * rowwt[orowidx];
			     for (ocolidx = 0; ocolidx <= two_cols; ocolidx++)
			     {
				 /* update the desired output entry */
				 /* BTW: ocol = ocolidx + left */
				 parms.old = comp->data.d1[idx];
				 parms.newp = parmval * colwt[ocolidx];
				 comp->data.d1[idx] = (c->strat)(&parms);
				 idx++; /* for next go round */
			     } /* end of for ocolidx */
			 } /* end of for orowidx */
		     } /* end of for irow */
		 } /* end of for icol */
	     } /* of if ColWeight < 1.0 */
	     else /* (ColWeight > 1.0) */
	     {
		 for (irow = InStartCol; irow <= InEndCol; irow++)
		 {

		     /* get current row indices */
		     /* if (c->use_c)
		      *{
		      *	minrowidx = hb->col_inds[icol];
		      *	maxrowidx = hb->col_inds[icol+1];
		      * }
		      *else
		      * {
		      *	minrowidx = hb->col_inds[icol]-1;
		      *	maxrowidx = hb->col_inds[icol+1]-1;
		      * }
		      */
		     mincolidx = hb->col_inds[icol] - index_offset;
		     maxcolidx = hb->col_inds[icol+1] - index_offset;

		     /* calculate left and right most output columns 
			affected by this input column  */
		     ftop = (irow-InStartCol) * RowWeight;
		     fbottom = ftop + RowWeight;
		     top = (int) ftop;
		     bottom = (int) fbottom;
		     /* two_rows = (bottom != top); */
		     two_rows = ((bottom != top) && (bottom != OutNumRows));

		     rowwt[1] = two_rows * (fbottom-bottom);
		     rowwt[0] = RowWeight - rowwt[1];

		     /* for the useful row numbers under the current column */
		     for (icol = mincolidx; icol < maxcolidx; icol++)
		     {
			 /* if (c->use_c)
			  *  rownum = hb->row_inds[irow];
			  * else
			  *   rownum = hb->row_inds[irow]-1;
			  */
			 colnum = hb->row_inds[icol] - index_offset;

			 /* added 3/19/94 -- support for submats */
			 if ((colnum < InStartRow) || (colnum > InEndRow))
			     continue;
			 /* end 3/19/94 */

			 /* get value */
			 /* 4/9/94 -- support for pattern matrices */
			 if (hb->head.NumericalData == IsPattern)
			     value = DEFAULT_PATTERN_VALUE;
			 else
			 {
			     value = hb->values[icol];
			     /* bitmap checks 4/20 */
			     /* 4/28 -- fixed bug since myabs is a MACRO */
			     if ((c->bitmap) && (myabs(value) > BITMAP_EPSILON) && 
				 (myabs((value-othervalue))>BITMAP_EPSILON))
			     {
				 if (other_val_init)
				     c->bitmap = 0;
				 else
				 {
				     other_val_init = 1;
				     othervalue = value;
				 }
			     }
			 } /* end of else not pattern */
			 /* since CW > 1, one input value covers more than 2
			    output cols */

			 /* do max/min calculations of the input */
			 absvalue = (value < 0.0) * -value + 
			     (value >=0.0) * value;

			 newmax = (value > input_max);
			 newmin = (value < input_min);

			 input_max = newmax * value + (1-newmax)*input_max;
			 input_min = newmin * value + (1-newmin)*input_min;

			 input_max_col = newmax * icol + 
			     (1-newmax) * input_max_col;
			 input_max_row = newmax * rownum + 
			     (1-newmax) * input_max_row;

			 input_min_col = newmin * icol + 
			     (1-newmin) * input_min_col;
			 input_min_row = newmin * rownum + 
			     (1-newmin) * input_min_row;

/* fprintf(stderr,"newmax = %d, value = %f, (1-newmax) = %d, input_max = %f\n",
	newmax,value, (1-newmax),input_max);
*/		     
			 newmax = (absvalue > input_absmax);
		         /* 9/14  change abs newmin to skip value = 0.0 */
		         newmin = ((absvalue < input_absmin) && 
			           (absvalue != 0.0));
			 input_absmax = newmax * absvalue + 
			     (1-newmax)* input_absmax;
			 input_absmin = newmin * absvalue + 
			     (1-newmin)*input_absmin;
		     
		         input_absmax_col = newmax * icol +
			     (1-newmax) * input_absmax_col;
		         input_absmax_row = newmax * rownum +
			     (1-newmax) * input_absmax_row;

		         input_absmin_col = newmin * icol +
			     (1-newmin) * input_absmin_col;
		         input_absmin_row = newmin * rownum +
			     (1-newmin) * input_absmin_row;
/*
			 fprintf(stderr,"CSM: newmax = %d, newmin = %d\n",
				 newmax,newmin);
    fprintf(stderr,"CSM: value = %f, absvalue = %f, input_min = %f, input_max = %f, input_absmin = %f, input_absmax = %f\n",value,absvalue,input_min, input_max, input_absmin, input_absmax);
*/
			 fleft = (colnum-InStartRow) * ColWeight;
			 fright = fleft+ColWeight;

			 /* need array indices for fleft and fright */
			 left = (int) fleft;
			 right = (int) fright;
			 notlastcol = (right != OutNumCols);
			 leftwt = (1.0 - (fleft - left));
			 rightwt = (fright-right);

			 for (orowidx = 0; orowidx <= two_rows; orowidx++)
			 {
			     /* calculate actual row number */
			     orow = orowidx + top;
			     /* parmval is the part of the value that's
				invariant below */
			     parmval = value * rowwt[orowidx];

			     /* left corner */
			     idx = orow * OutNumCols + left;
			     parms.old = comp->data.d1[idx];
			     parms.newp = parmval * leftwt;
			     comp->data.d1[idx] = (c->strat)(&parms);

			     /* then across the row */
			     for (ocol = left+1; ocol < right; ocol++)
			     {
				 idx++;
				 parms.old = comp->data.d1[idx];
				 parms.newp = parmval;
				 comp->data.d1[idx] = (c->strat)(&parms);
			     }

			     /* right corner */
			     idx+= notlastcol; /* 6/1/94 update only if not
						  on last col */

			     parms.old = comp->data.d1[idx];
			     parms.newp = parmval * rightwt;
			     comp->data.d1[idx] = (c->strat)(&parms);
			 } /* end of for orowidx */
		     } /* end of for irow */
		 } /* end of for icol */
	     } /* of else colweight > 1.0 */
	 } /* of if RowWeight < 1.0 */
	 else /* RowWeight >= 1.0 */ 
	 {
	     for (irow = InStartCol; irow <= InEndCol; irow++)
	     {

		 /* get current row indices */
		 /* if (c->use_c)
		  *{
		  *	minrowidx = hb->col_inds[icol];
		  *	maxrowidx = hb->col_inds[icol+1];
		  * }
		  *else
		  * {
		  *	minrowidx = hb->col_inds[icol]-1;
		  *	maxrowidx = hb->col_inds[icol+1]-1;
		  * }
		  */
		 mincolidx = hb->col_inds[icol] - index_offset;
		 maxcolidx = hb->col_inds[icol+1] - index_offset;

		 /* calculate left and right most output rows
		    affected by this input column  */
		 ftop = (irow-InStartCol) * RowWeight;
		 fbottom = ftop + RowWeight;
		 top = (int) ftop;
		 bottom = (int) fbottom;
		 false_bottom = bottom - (bottom == OutNumRows);
		 topwt = 1.0 - (ftop - top);
		 botwt = fbottom - bottom;

		 /* for the useful row numbers under the current column */
		 for (icol = mincolidx; icol < maxcolidx; icol++)
		 {
		     /* if (c->use_c)
		      *  rownum = hb->row_inds[irow];
		      * else
		      *   rownum = hb->row_inds[irow]-1;
		      */
		     colnum = hb->row_inds[icol] - index_offset;

		     /* added 3/19/94 -- support for submats */
		     if ((colnum < InStartRow) || (colnum > InEndRow))
			 continue;
		     /* end 3/19/94 */

		     /* get value */
		     /* 4/9/94 -- support for pattern matrices */
		     if (hb->head.NumericalData == IsPattern)
			 value = DEFAULT_PATTERN_VALUE;
		     else
		     {
			 value = hb->values[icol];
			 /* bitmap checks 4/20 */
			 /* 4/28 -- fixed bug since myabs is a MACRO */
			 if ((c->bitmap) && (myabs(value) > BITMAP_EPSILON) && 
			     (myabs((value-othervalue))>BITMAP_EPSILON))
			 {
			     if (other_val_init)
				 c->bitmap = 0;
			     else
			     {
				 other_val_init = 1;
				 othervalue = value;
			     }
			 }
		     } /* end of else not pattern */
		     /* since CW > 1, one input value covers more than 2
			output cols */

		     /* do max/min calculations of the input */
		     absvalue = (value < 0.0) * -value + 
			 (value >=0.0) * value;
		     newmax = (value > input_max);
		     newmin = (value < input_min);
		     input_max = newmax * value + (1-newmax)*input_max;
		     input_min = newmin * value + (1-newmin)*input_min;
/* fprintf(stderr,"newmax = %d, value = %f, (1-newmax) = %d, input_max = %f\n",
	newmax,value, (1-newmax),input_max);
*/		     
		     input_max_col = newmax * icol + 
			 (1-newmax) * input_max_col;
		     input_max_row = newmax * rownum + 
			 (1-newmax) * input_max_row;

		     input_min_col = newmin * icol + 
			 (1-newmin) * input_min_col;
		     input_min_row = newmin * rownum + 
			 (1-newmin) * input_min_row;

		     newmax = (absvalue > input_absmax);
		     /* 9/14  change abs newmin to skip value = 0.0 */
		     newmin = ((absvalue < input_absmin) && 
			       (absvalue != 0.0));

		     input_absmax = newmax * absvalue + 
			 (1-newmax)* input_absmax;
		     input_absmin = newmin * absvalue + 
			 (1-newmin)*input_absmin;

		     input_absmax_col = newmax * icol +
			 (1-newmax) * input_absmax_col;
		     input_absmax_row = newmax * rownum +
			 (1-newmax) * input_absmax_row;

		     input_absmin_col = newmin * icol +
			 (1-newmin) * input_absmin_col;
		     input_absmin_row = newmin * rownum +
			 (1-newmin) * input_absmin_row;
		     
/*
			 fprintf(stderr,"CSM: newmax = %d, newmin = %d\n",
				 newmax,newmin);
    fprintf(stderr,"CSM: value = %f, absvalue = %f, input_min = %f, input_max = %f, input_absmin = %f, input_absmax = %f\n",value,absvalue,input_min, input_max, input_absmin, input_absmax);
*/

		     fleft = (colnum-InStartRow) * ColWeight;
		     fright = fleft+ColWeight;

		     /* need array indices for fleft and fright */
		     left = (int) fleft;
		     right = (int) fright;
		     two_cols = (left != right) && (right < OutNumCols);
		     leftwt = two_cols * (1.0 - (fleft - left)) +
			 ((1-two_cols) * ColWeightFrac);
		     rightwt = (two_cols * (fright-right));

		     /* do top row */
		     parmval = value * topwt;
		     /* first UL (upper left)  corner */
		     idx = top * OutNumCols + left;
		     parms.old = comp->data.d1[idx];
		     parms.newp = parmval * leftwt;
		     comp->data.d1[idx] = (c->strat)(&parms);

		     /* then across the top row */
		     for (ocol = left+1; ocol < right; ocol++)
		     {
			 idx++;
			 parms.old = comp->data.d1[idx];
			 parms.newp = parmval;
			 comp->data.d1[idx] = (c->strat)(&parms);
		     }

		     /* now UR corner */
		     idx += two_cols; /* this should work */
		     parms.old = comp->data.d1[idx];
		     parms.newp = parmval * rightwt;
		     comp->data.d1[idx] = (c->strat)(&parms);

		     /* now on bottom row */
		     parmval = value * botwt;

		     /* bottom left corner */
		     idx = false_bottom * OutNumCols + left;
		     parms.old = comp->data.d1[idx];
		     parms.newp = parmval * leftwt;
		     comp->data.d1[idx] = (c->strat)(&parms);

		     /* then across the bottom row */
		     for (ocol = left+1; ocol < right; ocol++)
		     {
			 idx++;
			 parms.old = comp->data.d1[idx];
			 parms.newp = parmval;
			 comp->data.d1[idx] = (c->strat)(&parms);
		     }

		     /* now LR corner */
		     idx += two_cols;
		     parms.old = comp->data.d1[idx];
		     parms.newp = parmval * rightwt;
		     comp->data.d1[idx] = (c->strat)(&parms);

		     /* left and right columns */
		     for (orow = top+1; orow < bottom; orow++)
		     {
			 idx = orow * OutNumCols + left;
			 parms.old = comp->data.d1[idx];
			 parms.newp = value * leftwt;
			 comp->data.d1[idx] = (c->strat)(&parms);

			 idx = orow * OutNumCols + right;
			 parms.old = comp->data.d1[idx];
			 parms.newp = value * rightwt;
			 comp->data.d1[idx] = (c->strat)(&parms);
		     }

		     /* finally the center of the area covered */
		     for (orow = top+1; orow < bottom; orow++)
		     {
			 idx = orow * OutNumCols + left;
			 for (ocol = left+1; ocol < right; ocol++)
			 {
			     idx++;
			     parms.old = comp->data.d1[idx];
			     parms.newp = value;
			     comp->data.d1[idx] = (c->strat)(&parms);
			 } /* of for ocol */
		     } /* of for orow */
		 } /* end of for irow */
	     } /* end of for icol */
	 } /* end of if RowWeight > 1 */
     } /* end of else sp_col_major */
     /* clean up at end of loop */
     sprintf(errmsg,"compress sparse matrix thinks bitmap = %d\n",c->bitmap);
     DEBUG2(errmsg);

     /* return calculated values */
     c->input_min = input_min;
     c->input_max = input_max;
     c->input_absmin = input_absmin;
     c->input_absmax = input_absmax;

     c->input_min_row = input_min_row;
     c->input_min_col = input_min_col;

     c->input_max_row = input_max_row;
     c->input_max_col = input_max_col;

     c->input_absmin_row = input_absmin_row;
     c->input_absmin_col = input_absmin_col;

     c->input_absmax_row = input_absmax_row;
     c->input_absmax_col = input_absmax_col;
/*
 * fprintf(stderr,"FINAL CSM: min row, col = (%d, %d), max row, col = (%d, %d)\n",input_min_row, input_min_col, input_max_row, input_max_col);
 * fprintf(stderr,"FINAL CSM: absmin row, col = (%d, %d), absmax row, col = (%d, %d)\n",input_absmin_row, input_absmin_col, input_absmax_row, input_absmax_col);
 */
     sprintf(errmsg,"input_min = %e, input_max = %e, input_absmin = %e, input_absmax = %e\n",
	     c->input_min,c->input_max,c->input_absmin,c->input_absmax);
     DEBUG2(errmsg);
/* fprintf(stderr,errmsg); */

     TRACE("leaving CompressSparseMat (CSM)\n");
     return(OK);
 }

 /*
  * Function: CompressDenseMat
  *
  * Purpose: to compress a dense matrix into a dense matrix of values suitable
  *          for eventual scaling and displaying
  * 
  * Parameters: d - the sparse matrix to be compressed
  *             comp - the output matrix
  *             c - a compression request
  *
  * Return Values:  OK - all went well
  *                !OK - guess it didn't
  * 
  * Calls: sprintf error
  *
  * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
  *             tloos@cs.indiana.edu
  *
  * Started: 2/15/94
  *
  * Modified: 2/16/94 -- fixed addressing bug in column major case!
  *           3/18/94 -- removed duplicated code for both col and row major
  *                      cases: replaced with for major.. for minor... loops
  *           3/20/94 -- added submat support 
  *           5/11/94 -- try to speed things up again
  *           6/1/94  -- fix last row/col problems
  *           10/13/94 -- fix bitmap problem (see CSM)
  *
  */
 int CompressDenseMat( DenseMatrix *d, DenseMatrix *comp, CompressionRequest *c)
 {
     double value, othervalue, absvalue;
     char errmsg[100], majorstr[5];
     double RowWeight, ColWeight, ColWeightFrac;
     double leftwt, rightwt, topwt, botwt, parmval;
     double basecw, curcw, CWplus1, CWplus0;
     int irow, icol;
     int *major, *minor, idx, curcol, offset, ocol, orow;
     int orowidx, ocolidx, two_rows, two_cols;
     int orownum1, orownum2, left, right, top, bottom, notlastcol, false_bottom;
     double orow1, orow2, outrowwt1, outrowwt2, ftop, fbottom, fleft, fright;
     double rowwt[2], colwt[2];
     int  majormax, minormax, majormin, minormin;
     int input_min_row, input_max_row, input_min_col, input_max_col;
     int input_absmin_row, input_absmax_row, input_absmin_col, input_absmax_col;
     int InStartRow, InStartCol, InEndRow, InEndCol;
     int InNumCols, InNumRows, OutNumCols, OutNumRows, other_val_init = 0 ;
     double input_min, input_max, input_absmin, input_absmax;
     int newmax, newmin;
     CompressionStrategyParameters parms;

     /* check for good matrix sizes (i.e. number of rows and cols all > 0) */

     TRACE("entered CompressDenseMat (CDM)\n");

     /* get some shorthand names */
     InStartRow = c->input_startrow;
     InEndRow = c->input_endrow;
     InStartCol = c->input_startcol;
     InEndCol = c->input_endcol;

     /* InNumCols = d->NumCols; */
     /* InNumRows = d->NumRows; */

     /* changed from simple subtraction on 4/14/94 */
     InNumRows = IDXRANGE(InEndRow, InStartRow);
     InNumCols = IDXRANGE(InEndCol, InStartCol);

     OutNumCols = comp->NumCols;
     OutNumRows = comp->NumRows;

     /* 10/13/94 -- fix bitmap problem */
     othervalue = 0.0;

     if ((InNumRows <= 0) || (InNumCols <= 0) || (InNumRows > d->NumRows) || 
	 (InNumCols > d->NumCols))
     {
	 sprintf(errmsg,
		 "CompressDenseMat:input row count (%d) or col count (%d) < 0",
		 InNumRows, InNumCols);
	 error(errmsg, ARGERR);
	 /* shouldn't get here */
	 return(BAD);
     }

     if ((OutNumRows <= 0) || (OutNumCols <= 0))
     {
	 sprintf(errmsg,
		 "CompressDenseMat: output row (%d) or col count (%d) < 0",
		 OutNumRows, OutNumCols);
	 error(errmsg, ARGERR);
	 /* shouldn't get here */
	 return(BAD);
     }

     DEBUG2("CompressDenseMatrix before weight calculation\n");
     /*
      * calculate weights:
      * A weight is the number of output rows/cols that correspond
      *  to an input row or column.
      *
      * For compression, expect weights to be  < 1
      * For expansion, expect weights to be > 1
      */
     RowWeight = (double) OutNumRows / (double) InNumRows;
     ColWeight = (double) OutNumCols / (double) InNumCols;
     ColWeightFrac = ColWeight - (int) ColWeight;
     sprintf(errmsg,"CompressDenseMatrix 3: one_d = %d, RowWeight = %f, ColWeight = %f\n",d->one_d,RowWeight, ColWeight);
     DEBUG2(errmsg);

     /* 
      * init max and min
      */

     /* assignments to allow input min and max to be set */
     icol = InStartCol;
     irow = InStartRow;

     /* two different loops, depending on row_major of input */
     if (d->row_major)
     {
	 major = &irow;
	 minor = &icol;
	 offset = d->NumCols;
     }
     else 
     {
	 major = &icol;
	 minor = &irow;
	 offset = d->NumRows;
     }
     /* store the input min/max */
     /*
      * input_min = c->input_min;
      * input_max = c->input_max;
      * input_absmin = c->input_absmin;
      * input_absmax = c->input_absmax;
      */

     input_min = input_absmin =  MAX_SCALING_VALUE;
     input_max = input_absmax =  MIN_SCALING_VALUE;
     /* and their positions */
     input_min_col = input_max_col = input_absmin_col = input_absmax_col = 
	 InStartCol;
     input_min_row = input_max_row = input_absmin_row = input_absmax_row = 
	 InStartRow;

     /* move some of the bitmap and min/max stuff out of the loop */
     if (d->one_d)
     {
	 othervalue = d->data.d1[0];
	 /* input_min = input_max = d->data.d1[(*major) * offset + (*minor)]; */
	 absvalue = myabs(input_min);
	 /* input_absmin = input_absmax = absvalue; */
     }
     else
     {
	 othervalue = d->data.d2[0][0];
	 /* input_min = input_max = d->data.d2[(*major)][(*minor)]; */
	 absvalue = myabs(input_min);
	 /* input_absmin = input_absmax = absvalue; */
     }
     other_val_init = ((myabs(othervalue)) > BITMAP_EPSILON);

     /* fprintf(stderr,"input_absmin = %e\n",input_absmin); */
     if (RowWeight < 1.0)    /* if compressing rows */
     {
	 if (ColWeight < 1.0) /* and if compresssing columns */
	 {
	     for (icol = InStartCol; icol <= InEndCol; icol++)
	     {
		 /* calculate left and right most output columns affected by 
		    this input column  */
		 fleft = (icol-InStartCol) * ColWeight;
		 fright = fleft + ColWeight;
		 left = (int) fleft;
		 right = (int) fright;
		 two_cols = (left != right) && (right < OutNumCols);
		 colwt[1] = two_cols * (fright - right);
		 colwt[0] = ColWeight - colwt[1];


		 for (irow = InStartRow; irow <= InEndRow; irow++)
		 {
		     if (d->one_d)
			 value = d->data.d1[(*major)*offset+(*minor)];
		     else
			 value = d->data.d2[(*major)][(*minor)];

		     /* do max/min calculations of the input */		
		     absvalue = (value < 0.0) * -value + 
			 (value >=0.0) * value;
		     newmax = (value > input_max);
		     newmin = (value < input_min);
		     input_max = newmax * value + (1-newmax)*input_max;
		     input_min = newmin * value + (1-newmin)*input_min;

		     input_max_col = newmax * icol +
			 (1-newmax) * input_max_col;
		     input_max_row = newmax * irow +
			 (1-newmax) * input_max_row;

		     input_min_col = newmin * icol +
			 (1-newmin) * input_min_col;
		     input_min_row = newmin * irow +
			 (1-newmin) * input_min_row;
		     
/*			 fprintf(stderr,"CDM: newmax1 = %d, newmin1 = %d\n",
 *			 newmax,newmin);
 */

		     newmax = (absvalue > input_absmax);
		     /* 9/14 -- change newmin for dense to skip value = 0.0 */
		     newmin = ((absvalue < input_absmin) && (absvalue != 0.0));
		     input_absmax = newmax * absvalue + 
			 (1-newmax)* input_absmax;
		     input_absmin = newmin * absvalue + 
			 (1-newmin)*input_absmin;
/*
 * if (newmin)
 *    fprintf(stderr,"input_absmin = %e\n",input_absmin);
 */

		     input_absmax_col = newmax * icol +
			 (1-newmax) * input_absmax_col;
		     input_absmax_row = newmax * irow +
			 (1-newmax) * input_absmax_row;

		     input_absmin_col = newmin * icol +
			 (1-newmin) * input_absmin_col;
		     input_absmin_row = newmin * irow +
			 (1-newmin) * input_absmin_row;
		     
/*
			fprintf(stderr,"CDM: newmax = %d, newmin = %d\n",
				 newmax,newmin);
    fprintf(stderr,"CDM: value = %f, absvalue = %f, input_min = %f, input_max = %f, input_absmin = %f, input_absmax = %f\n",value,absvalue,input_min, input_max, input_absmin, input_absmax);
*/

		     /* check for a bitmap */
		     if ((c->bitmap) && (myabs(value) > BITMAP_EPSILON) && 
			 (myabs((value-othervalue))>BITMAP_EPSILON))
		     {
			 if (other_val_init)
			     c->bitmap = 0;
			 else
			 {
			     other_val_init = 1;
			     othervalue = value;
			 }
		     }

		     /* since RW < 1, one value covers at most only 2 output 
			rows */
		     /* ftop = certain output row -- fbottom only if input
			straddles 2 pixels */
		     ftop = (irow-InStartRow) * RowWeight;
		     fbottom = ftop+RowWeight;


		     /* need array indices for fleft and fright */
		     top = (int) ftop;
		     bottom = (int) fbottom;
		     /* two_rows = 1 if there are 2 rows (bottom != top) */
		     /*          = 0 if there's only 1 (bottom == top) */
		     /* two_rows = (bottom != top); */
		     two_rows = ((bottom != top) && (bottom != OutNumRows));

		     /* get weights for the 2 output rows */
		     /* wt[1] = weight for second pixel:
		      *       = 0 if integer row numbers are the same
		      *	     = fp(fbottom) if different 
		      * use two_rows to determine if they differ
		      * so wt2 is:
		      *    two_rows *  (fbottom-bottom)
		      *
		      * wt[0] = RowWeight - wt[1];
		      */
		     rowwt[1] = two_rows * (fbottom-bottom);
		     rowwt[0] = RowWeight - rowwt[1];

		     for (orowidx = 0; orowidx <= two_rows; orowidx++)
		     {
			 /* calculate actual row number */
			 /* orow = top + orowidx */
			 idx = (top + orowidx) * OutNumCols + left;

			 /* parmval is the part of the value that's
			    invariant below */
			 parmval = value * rowwt[orowidx];
			 for (ocolidx = 0; ocolidx <= two_cols; ocolidx++)
			 {
			     /* update the desired output entry */
			     /* BTW: ocol = ocolidx + left */
			     parms.old = comp->data.d1[idx];
			     parms.newp = parmval * colwt[ocolidx];
			     comp->data.d1[idx] = (c->strat)(&parms);
			     idx++; /* for next go round */
			 } /* end of for ocolidx */
		     } /* end of for orowidx */
		 } /* end of for irow */
	     } /* end of for icol */
	 } /* of if colweight < 1.0 */
	 else /* ColWeight > 1.0 */ 
	 {
	     for (icol = InStartCol; icol <= InEndCol; icol++)
	     {
		 /* calculate left and right most output columns affected by 
		    this input column  */
		 fleft = (icol-InStartCol) * ColWeight;
		 fright = fleft + ColWeight;
		 left = (int) fleft;
		 right = (int) fright;
		 leftwt = (1.0 - (fleft - left));
		 notlastcol = (right < OutNumCols);
		 rightwt = (fright-right);

		 for (irow = InStartRow; irow <= InEndRow; irow++)
		 {
		     if (d->one_d)
			 value = d->data.d1[(*major)*offset+(*minor)];
		     else
			 value = d->data.d2[(*major)][(*minor)];

		     /* do max/min calculations of the input */
		     absvalue = (value < 0.0) * -value + 
			 (value >=0.0) * value;
		     newmax = (value > input_max);
		     newmin = (value < input_min);
		     input_max = newmax * value + (1-newmax)*input_max;
		     input_min = newmin * value + (1-newmin)*input_min;

		     input_max_col = newmax * icol +
			 (1-newmax) * input_max_col;
		     input_max_row = newmax * irow +
			 (1-newmax) * input_max_row;

		     input_min_col = newmin * icol +
			 (1-newmin) * input_min_col;
		     input_min_row = newmin * irow +
			 (1-newmin) * input_min_row;
		     
		     newmax = (absvalue > input_absmax);
		     /* 9/14 -- change newmin for dense to skip value = 0.0 */
		     newmin = ((absvalue < input_absmin) && (absvalue != 0.0));
		     input_absmax = newmax * absvalue + 
			 (1-newmax)* input_absmax;
		     input_absmin = newmin * absvalue + 
			 (1-newmin)*input_absmin;

		     input_absmax_col = newmax * icol +
			 (1-newmax) * input_absmax_col;
		     input_absmax_row = newmax * irow +
			 (1-newmax) * input_absmax_row;

		     input_absmin_col = newmin * icol +
			 (1-newmin) * input_absmin_col;
		     input_absmin_row = newmin * irow +
			 (1-newmin) * input_absmin_row;

		     /* check for a bitmap */
		     if ((c->bitmap) && (myabs(value) > BITMAP_EPSILON) && 
			 (myabs((value-othervalue))>BITMAP_EPSILON))
		     {
			 if (other_val_init)
			     c->bitmap = 0;
			 else
			 {
			     other_val_init = 1;
			     othervalue = value;
			 }
		     }

		     /* since RW < 1, one value covers at most only 2 output 
			rows */
		     /* fleft = certain output row -- fright only if input
			straddles 2 pixels */
		     ftop = (irow-InStartRow) * RowWeight;
		     fbottom = ftop+RowWeight;

		     /* need array indices for ftop and fbottom */
		     top = (int) ftop;
		     bottom = (int) fbottom;
		     /* two_rows = 1 if there are 2 rows (bottom != top) */
		     /*          = 0 if there's only 1 (bottom == top) */
		     /* two_rows = (bottom != top);*/
		     two_rows = ((bottom != top) && (bottom != OutNumRows));

		     /* get weights for the 2 output rows */
		     /* wt[1] = weight for second pixel:
		      *       = 0 if integer row numbers are the same
		      *	     = fp(fright) if different 
		      *  use two_rows to determine if on 2 diff. rows
		      *  fp(fbottom) = fbottom - bottom
		      *  so wt[1] is:
		      *    (two_rows) * (fbottom-bottom)
		      *
		      * wt1 = RowWeight - wt2;
		      */

		     rowwt[1] = two_rows * (fbottom-bottom);
		     rowwt[0] = RowWeight - rowwt[1];

		     for (orowidx = 0; orowidx <= two_rows; orowidx++)
		     {
			 /* calculate actual row number */
			 orow = orowidx + top;
			 /* parmval is the part of the value that's
			    invariant below */
			 parmval = value * rowwt[orowidx];

			 /* first on top row */
			 /* left corner */
			 idx = orow * OutNumCols + left;
			 parms.old = comp->data.d1[idx];
			 parms.newp = parmval * leftwt;
			 comp->data.d1[idx] = (c->strat)(&parms);

			 /* then across the top row */
			 for (ocol = left+1; ocol < right; ocol++)
			 {
			     idx++;
			     parms.old = comp->data.d1[idx];
			     parms.newp = parmval;
			     comp->data.d1[idx] = (c->strat)(&parms);
			 }

			 /* now UR corner */
			 idx+= notlastcol; /* only update if not on last column */

			 parms.old = comp->data.d1[idx];
			 parms.newp = parmval * rightwt;
			 comp->data.d1[idx] = (c->strat)(&parms);
		     } /* of for orowidx */
		 } /* end of for irow */
	     }/* end of for icol */
	 } /* of else colweight > 1.0 */
     } /* end of if RowWeight < 1 */
     else /* RowWeight >= 1 */
     {

/*	 fprintf(stderr,"RW>=1: offset = %d, othervalue = %f\n",offset, othervalue);  */

	 for (icol = InStartCol; icol <= InEndCol; icol++)
	 {

	     /* calculate left and right most output columns affected by this
		input column  */
	     fleft = (icol-InStartCol) * ColWeight;
	     fright = fleft + ColWeight;
	     left = (int) fleft;
	     right = (int) fright;

	     /* two_cols is a flag that answers:
	      *  "Are the left and the right columns different and not on the
	      *   last column?"
	      *
	      *  The answer is:
	      *     Yes, if left != right.
	      *  If the left and right cols are different, then
	      *      left_weight = 1-fp(fleft)
	      *	     right_weight = fp(fright)
	      *  else
	      *      left_weight = fp(ColWeight);
	      *	     right_weight = 0;
	      *  endif
	      *  Since Yes == 1 (for boolean operators):
	      *
	      *  left_weight = two_cols*(1-fp(fleft)) + (1-two_cols)*ColWeight;
	      *  right_weight = two_cols * fp(fright) +
	      *                 (1-two_cols) * ColWeight;
	      */

	    two_cols = (left != right) && (right < OutNumCols);

	    leftwt = two_cols * (1.0 - (fleft - left)) + 
		    ((1-two_cols) * ColWeightFrac);
	    rightwt = (two_cols * (fright-right));


	    /* DEBUGGING ONLY */
/*	    if (right >= OutNumCols)
 *		fprintf(stderr,"ONE STEP BEYOND!: leftwt = %f, rightwt = %f, two_cols = %d, right = %d,  left = %d, ColWeight = %f, OutNumCols = %d\n",
 * 			leftwt,rightwt,two_cols,right,left,ColWeight,OutNumCols);
*/

	    /* for the input rows */
	    for (irow = InStartRow; irow <= InEndRow; irow++)
	    {
		/* get value */
		if (d->one_d)
		    value = d->data.d1[(*major)*offset+(*minor)];
		else
		    value = d->data.d2[(*major)][(*minor)];

		/* do max/min calculations of the input */
		absvalue = (value < 0.0) * -value + 
		    (value >=0.0) * value;
		newmax = (value > input_max);
		newmin = (value < input_min);
		input_max = newmax * value + (1-newmax)*input_max;
		input_min = newmin * value + (1-newmin)*input_min;

		input_max_col = newmax * icol + (1-newmax) * input_max_col;
		input_max_row = newmax * irow + (1-newmax) * input_max_row;

		input_min_col = newmin * icol + (1-newmin) * input_min_col;
		input_min_row = newmin * irow + (1-newmin) * input_min_row;
		
		newmax = (absvalue > input_absmax);
		newmin = ((absvalue < input_absmin) && (absvalue != 0.0));
		input_absmax = newmax * absvalue + 
		    (1-newmax)* input_absmax;
		input_absmin = newmin * absvalue + 
		    (1-newmin)*input_absmin;

		input_absmax_col = newmax * icol +
		    (1-newmax) * input_absmax_col;
		input_absmax_row = newmax * irow +
		    (1-newmax) * input_absmax_row;

		input_absmin_col = newmin * icol +
		    (1-newmin) * input_absmin_col;
		input_absmin_row = newmin * irow +
		    (1-newmin) * input_absmin_row;
		
		/* bitmap checks 4/20 */
		/* 4/28 -- fixed bug since myabs is a MACRO */
		if ((c->bitmap) && (myabs(value) > BITMAP_EPSILON) && 
		    (myabs((value-othervalue))>BITMAP_EPSILON))
		{
		    if (other_val_init)
			c->bitmap = 0;
		    else
		    {
			other_val_init = 1;
			othervalue = value;
		    }
		}
		
		/* calculate top and bottom row and their weights */
		ftop = (irow-InStartRow) * RowWeight;
		fbottom = ftop + RowWeight;
		top = (int) ftop;
		bottom = (int) fbottom;
		false_bottom = bottom - (bottom == OutNumRows);
		topwt = 1.0 - (ftop - top);
		botwt = fbottom - bottom;
/* DEBUGGING */
/*
 *		if (value > 0)
 *		{
 *		    fprintf(stderr,"fleft = %f, fright = %f, left = %d, right = %d\n",
 *			    fleft, fright, left, right);
 *		    fprintf(stderr,"   ftop = %f, fbottom = %f, top = %d, bottom = %d\n",
 *			    ftop, fbottom, top, bottom);
 *		    fprintf(stderr,"   topwt = %f, botwt = %f, leftwt = %f, rightwt = %f\n",
 *			   topwt, botwt, leftwt, rightwt);
 *		}
 *		else
 *		    fprintf(stderr,"[%d, %d]: top = %d, bottom = %d, left = %d, right = %d, two_cols = %d\n",irow,icol,top,bottom,left,right,two_cols);
 *
 */

		/* do top row */
		/* first UL (upper left)  corner */
		parmval = value * topwt;

		idx = top * OutNumCols + left;
		parms.old = comp->data.d1[idx];
		parms.newp = parmval * leftwt;
		comp->data.d1[idx] = (c->strat)(&parms);
		
		/* then across the top row */
		for (ocol = left+1; ocol < right; ocol++)
		{
		    idx++;
		    parms.old = comp->data.d1[idx];
		    parms.newp = parmval;
		    comp->data.d1[idx] = (c->strat)(&parms);
		}

		/* now UR corner */
                /* could only have one col, so can't do idx++ here */
		idx += two_cols;
		parms.old = comp->data.d1[idx];
		parms.newp = parmval * rightwt;
		comp->data.d1[idx] = (c->strat)(&parms);

		/* now for the bottom row */
		parmval = value * botwt;

		/* first LL (lower left)  corner */
		idx = false_bottom * OutNumCols + left;
		parms.old = comp->data.d1[idx];
		parms.newp = parmval * leftwt;
		comp->data.d1[idx] = (c->strat)(&parms);
		
		/* then across the bottom row */
		for (ocol = left+1; ocol < right; ocol++)
		{
		    idx++;
		    parms.old = comp->data.d1[idx];
		    parms.newp = parmval;
		    comp->data.d1[idx] = (c->strat)(&parms);
		}

		/* now LR corner */
                /* could only have one col, so can't do idx++ here */
		idx += two_cols;
		parms.old = comp->data.d1[idx];
		parms.newp = parmval * rightwt;
		comp->data.d1[idx] = (c->strat)(&parms);
		
		/* left and right columns */
		for (orow = top+1; orow < bottom; orow++)
		{
		    idx = orow * OutNumCols + left;
		    parms.old = comp->data.d1[idx];
		    parms.newp = value * leftwt;
		    comp->data.d1[idx] = (c->strat)(&parms);

		    idx = orow * OutNumCols + right;
		    parms.old = comp->data.d1[idx];
		    parms.newp = value * rightwt;
		    comp->data.d1[idx] = (c->strat)(&parms);
		}
		/* finally the center of the area covered */
		for (orow = top+1; orow < bottom; orow++)
		{
		    idx = orow * OutNumCols + left;
		    for (ocol = left+1; ocol < right; ocol++)
		    {
			idx++;
			parms.old = comp->data.d1[idx];
			parms.newp = value;
			comp->data.d1[idx] = (c->strat)(&parms);
		    } /* of for ocol */
		} /* of for orow */
	    } /* of for irow */
	} /* of for icol */
    } /* of else RowWeight > 1 */
    /* fprintf(stderr,"bitmap = %d\n",c->bitmap); */

    /* store the input min/max */
    c->input_min = input_min;
    c->input_max = input_max;
    /*  fprintf(stderr,"final input_absmin = %e\n",input_absmin); */
    c->input_absmin = input_absmin;
    c->input_absmax = input_absmax;
    c->input_min_row = input_min_row;
    c->input_min_col = input_min_col;
     
    c->input_max_row = input_max_row;
    c->input_max_col = input_max_col;
    
    c->input_absmin_row = input_absmin_row;
    c->input_absmin_col = input_absmin_col;

    c->input_absmax_row = input_absmax_row;
    c->input_absmax_col = input_absmax_col;

    TRACE("leaving CompressDenseMat (CDM)\n");
    return(OK);
}

/*
 * Function: GetMinMax
 *
 * Purpose: find the min, max, absolute min and absolute max of the input
 *          compressed (1D dense) matrix
 *
 * Parameters: comp -  pointer to the one-d dense matrix
 *             c - compression request to store max/min/absmax/absmin values
 *
 * Return Values: none
 * 
 * Calls: none
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started:   3/1/94
 *
 * Modified:  6/1/94 -- check for zero input/hack out zero absvals
 *
 */

void GetMinMax( DenseMatrix *comp, CompressionRequest *c )
{
    int i,nrows, ncols, total_elements;
    int zero_input = 0;
    double min, max, absmin, absmax, value, absvalue;
    char debugmsg[100];
    int min_idx, max_idx, amin_idx, amax_idx;


    /* based on the fact that comp is one-d */
    if (comp->one_d == 0)
    {
#ifdef STANDALONE
	error("GetMinMax: compressed matrix has to be one dimensional",BUGERR);
#else
	fprintf(stderr,"\nGetMinMax: compressed matrix has to be 1D\n");
#endif
	/* shouldn't get here */
	return;
    }

    /* get initial values */

    total_elements = comp->NumRows * comp->NumCols;
    value = comp->data.d1[0];
    /* fprintf(stderr,"GetMinMax: initial value = %f\n",value); */ 
    zero_input = (zero_input || (value == 0.0));
    min = max = value;
    absmin = absmax = myabs(value);
    min_idx = max_idx = amin_idx = amax_idx = 0;

    /* go thru the elements */
    for (i = 1; i < total_elements; i++)
    {
	value = comp->data.d1[i];
	zero_input = (zero_input || (value == 0.0));
	absvalue = myabs(value);

	if (value < min)
	{
	    min_idx = i;
	    min = value;
	}
	if (value > max)
	{
	    max_idx = i;
	    max = value;
	}

	if (absvalue == 0.0)
	    continue;

	absmin =  (absmin != 0.0) * absmin + (absmin == 0.0) * absvalue;

	if (absvalue < absmin)
	{
	    amin_idx = i;
	    absmin = absvalue;
	}
	if ( absvalue > absmax)
	{
	    amax_idx = i;
	    absmax = absvalue;
	}
    } /* of for i */
	
    c->mat_min = min;
    c->mat_max = max;
    c->mat_absmin = absmin;
    c->mat_absmax = absmax;
    c->zero_in_input = zero_input;

/*
 *  fprintf(stderr,"GetMinMax: amin (%f) = (%d,%d), amax = (%d,%d), min (%f) = (%d,%d), max = (%d,%d)\n",
 *	   c->mat_absmin,
 *	    amin_idx/comp->NumCols,amin_idx%comp->NumCols,
 *	    amax_idx/comp->NumCols,amax_idx%comp->NumCols,
 *	    c->mat_absmin,min_idx/comp->NumCols,min_idx%comp->NumCols,
 *	    max_idx/comp->NumCols,max_idx%comp->NumCols);
 */

    sprintf(debugmsg,"min = %9.1e, max = %9.1e, absmin = %9.1e, absmax = %9.1e, zero_in_input = %d\n",min,max,absmin,absmax,zero_input);
    DEBUG1(debugmsg);
}
