/*
 * File: hbdraw.h
 *
 * References:
 *  Users' Guide for the Harwell-Boeing Sparse Matrix Collection (Release I),
 *     Duff, Grimes, and Lewis, publisher?, Oct. 1992
 *
 * Volume I: Xlib Programming Manual, Adrian Nye, O'Reilly and Associates,
 *    7th printing, 12/91
 * 
 * Includes: definitions for drawing, scaling, and compressing routines
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: 2/13/94
 *
 * Modified: ...
 *           2/28/94 added flip_rows to CompresssionRequest
 *           ...
 *           3/18/94 removed flip_rows
 *           3/31/94 added the GETMENUBAR macro
 *
 */
#ifndef HBDRAW_H
#define HBDRAW_H

/* GETMENUBAR(w) gets the menu bar widget associated with a returned widget
   from a menu selection -- mainly used for GetZoomcount calls in callbacks.
 */
#define GETMENUBAR(w) XtParent(XtParent(XtParent(w)))

/*
 * Note: below the term "compression" can also mean expansion, 
 *  since the CompressMat routine will expand a small matrix into
 *  a big window, if need be.
 */

/*
 * WRT CompressMat(), the strategy implments the decision of 
 *  what to store into the matrix at that point.  
 *  and the effect of the code that will be executed is:
 *
 *  compresssion_strategy_parameters.old = compressed_matrix[row][column];
 *  compresssion_strategy_parameters.new = uncompressed_matrix[urow][ucol];
 *
 *  (urow == uncompressed_matrix's current row and 
 *   ucol == uncompressed_matrix's current col)
 *
 *  compressed_matrix[row][column] = 
 *       strategy( &compression_strategy_parameters )
 *	         
 * for example:
 *    if the output matrix should store non-zero counts, we could use:
 *    NZstrategy( double old, double new )
 *    {
 *          if (fabs(new) > EPSILON)
 *          {
 *               ++old;
 *          }
 *          return(old);
 *    }
 * 
 * CompressMat() will not perform any averaging on the elements in the 
 *    compressed matrix, but will defer that to ScaleMat()
 */

typedef struct _CSP
{
    double old; /* the current value in the compressed matrix */
    double newp; /* the value in the current cell of the uncompressed matrix */
} CompressionStrategyParameters;

/*  
 * ScaleMat() also uses a strategy with 6 parameters, to implement the
 *   following strategy:
 *
 * scale_strategy_parameters.value =  compressed_matrix[row][column];
 * scale_strategy_parameters.min =  min(compressed_matrix);
 * scale_strategy_parameters.max =  max(compressed_matrix);
 * scale_strategy_parameters.absmin = min(abs(compressed_matrix));
 * scale_strategy_parameters.absmax = max(abs(compressed_matrix));
 * scale_strategy_parameters.numcolors = total_number_of_colors_available;
 * 
 * output_color[row][column] = strategy( &scale_strategy_parameters );
 *
 *   An output color is an integer in the range:
 *        [0, total_number_of_colors_available]
 *
 * A simple strategy that implements a linear color scaling of the absolute 
 * values of matrix elements is:
 * 
 *   ShowAbsValuesStrategy( ScaleStrategyParameters *s)
 *   {
 *        double val = fabs(s->value);
 *        double range = s->absmax - s->absmin;
 *        // if there is no max, why not all be the max color
 *        if (range < EPSILON)
 *            return( s->numcolors );
 *        return( (val-s->absmin)/range * s->numcolors );
 *   }
 *
 */

typedef struct _SSP
{
    double val;    /* the current compressed matrix value */
    double min;    /* the minimum value in the compressed matrix */
    double max;    /* the maximum value in the compressed matrix */
    double lmin;    /* the log of the minimum value in the compressed matrix */
    double lmax;    /* the log of the maximum value in the compressed matrix */
    double absmin; /* the minimum value (in absolute terms) in the matrix */
    double absmax; /* the maximum value (in absolute terms) in the matrix */
    int numcolors; /* the total number of colors available */
    int bitmap; /* 1 if input was a bitmap */
    int log10; /* 1 if log should use log 10 scaling, 0 otherwise */
} ScaleStrategyParameters;

/* values for compression requests conformal values */
#define CONFORMAL_TO_SCREEN 0 /* try to scale to screen's shape */
#define CONFORMAL_TO_MATRIX 1 /* try to scal to matrix's shape */

typedef double (CompressionStrategy)( CompressionStrategyParameters *c );

typedef int (ScaleStrategy)( ScaleStrategyParameters *s );

/*
 * a compression request, which is a fancy name for a structure
 * to hold all the parameters to CompressMat() that aren't the input
 * or output matrix, so that it can be easily changed 
 *
 * max and min values are of the COMPRESSSED matrix 
 * 
 * a similar idea is used for ScaleMat()
 */
typedef struct _CompressionRequest
{
    int max_x; /* screen_x */
    int max_y; /* screen_y */
    int use_c; /* use c indexing or fortran? */
    /* next for to determine parameters of actual compression */
    int input_startrow;
    int input_startcol;
    int input_endrow;
    int input_endcol;

    double input_min; /* minimum value in source matrix */
    int input_min_col;
    int input_min_row;

    double input_max; /* maximum value in source matrix */
    int input_max_col;
    int input_max_row;

    double input_absmin; /* |min value| in source matrix */
    int input_absmin_col;
    int input_absmin_row;

    double input_absmax; /* |max value| in source matrix */
    int input_absmax_col;
    int input_absmax_row;

    /* col_major is overridden by the row_major field of a dense matrix */
    int sp_col_major; /* is sparse input column major? (randy_order :-) ) */
    double mat_min; /* the maximum value in the matrix (return value) */
    double mat_max; /* the maximum value in the matrix (return value) */
    double mat_absmin; /* the minimum value in |matrix| (return value) */
    double mat_absmax; /* the maximum value in |matrix| (return value) */
    int zero_in_input; /* is there a 0.0 in the compression matrix? (rv) */
    int bitmap; /* were there only 2 values in the input (return value) */
    int conformal; /* see above */
    CompressionStrategy *strat; /* what to do with each element */
} CompressionRequest;

typedef struct _ScaleRequest
{
    double min;    /* minimum value in the matrix */
    double max;    /* maximum value in the matrix */
    double absmin; /* absolute min. value in the matrix */
    double absmax; /* absolute max. value in the matrix */
    int zero_in_input; /* is there a 0.0 in the compression matrix? */
    int log10; /* use log10 or nat. log scaling for logarithmic scaling? */
    Display *display; /* display of pixmap */
    
    int bitmap; /* 1 if input was a bitmap */
    GC *gcs;       /* array of GCs with colors to fill up the pixmap with */
    int numcolors; /* the size of the gcs array */
    int startcolor; /* color starting value */
    int force_black; /* do we force black to be 0? */
    int what_is_black; /* the color for black, if forced */
    ScaleStrategy *strat; /* what to do with each element */
    int num_row_ptns; /* number of row partitions */
    int num_col_ptns; /* number of column partitions */
    int *row_ptn; /* the row partition vector */
    int *col_ptn; /* the col partition vector */
    int start_row; /* starting row of input */
    int end_row; /* ending row of input */
    int start_col; /* starting col of input */
    int end_col; /* ending col of input */
} ScaleRequest;

/* As usual, the function prototypes are at the bottom */

/* compress.c */
void CalculateScreenSize( int in_rows, int in_cols, int *out_rows, int *out_cols, int conformal);
int CompressMat( DaMatrix *M, DenseMatrix **comp, CompressionRequest *c );
int CompressSparseMat( HBMatrix *hb, DenseMatrix *comp, CompressionRequest *c );
int CompressDenseMat( DenseMatrix *d, DenseMatrix *comp, CompressionRequest *c );
void GetMinMax( DenseMatrix *comp, CompressionRequest *c );
/* compress_strat.c: some example compression strategies */
CompressionStrategy CSsum;    /* add up the values */
CompressionStrategy CSsquare; /* sum of squares */
CompressionStrategy CSone;   /* just return one for a bitmap matrix */
CompressionStrategy CSnzcount; /* count number of non-zeroes in the matrix */

/* comp_main.c */
/* main(int argc, char **argv); */
double CSSum();

#ifdef HANDLE_PTNS
int  CalcScreenPtn( int ptn[], int new_ptn[], int *num_ptns, int in_start, 
		    int in_end, int screen_dimension);
#else
int ScaleMat( DenseMatrix *comp, Pixmap p, ScaleRequest *s );

#endif
/* SUPPORTED SCALING STRATEGIES */
ScaleStrategy SClinear2; /* linear averaging using actual min/max/value w bitmap*/
ScaleStrategy SCabslinear; /* linear averaging using actual min/max/value */
ScaleStrategy SClog; /* logarithmic scaling */

/* UNSUPPORTED SCALING STRATEGIES */
/* ScaleStrategy SCsqrt; */   /* assuming sum of squares, this gives the 2 norm */
/* ScaleStrategy SClinear; */ /* linear averaging using actual min/max/value */
/* ScaleStrategy SCpattern; */ /* bitmap of values (plot numcolors if value is nz) */

/* from Ken's code and other stuff needed */
typedef struct _event {
#	define BUTTON_PRESS		0
#	define BUTTON_RELEASE		1
#	define MOUSE_MOVEMENT		2
	int type;
	int x, y;
	int button_number;
	int zoomcount;
} Event;


typedef struct _RGB {
	int Red;
	int Green;
	int Blue;
} RGB;

enum _RCASAction
{
    RCASread = 0, /* implies compress and scale */
    RCAScompress = 1, /* implies compress */
    RCASscale = 2, 
    RCASscalearray = 3,  /* scaling output to an array */
    RCASexpose = 4
};
typedef enum _RCASAction RCASAction;

#define USER_COLORS 240 /* number of user usable colors */
#define RESERVED_COLORS 8 /* hardcoded into GetColors() -- see below */

#define MAXIMUM_COLORS USER_COLORS+RESERVED_COLORS /* maxcolors used by ScaleMat() */

/* indices of the additional colors */
#define BLACK_INDEX  0
#define WHITE_INDEX  1
#define GRAY_INDEX  2
#define RED_INDEX  3
#define GREEN_INDEX  4
#define BLUE_INDEX  5
#define FG_INDEX 6 /* the foreground color */
#define BG_INDEX 7 /* the foreground color */

#define PTN_LINE_COLOR WHITE_INDEX

#define RCAS_FREE 0
#define RCAS_USED 1

enum _RCASNuke
{
    RCAS_NO_NUKES = 0,
    RCAS_NUKE_IT = 1
};

typedef enum _RCASNuke RCASNuke;

/* widget storage structure */
typedef struct _WidStruct
{
    Widget toplevel;
    Widget form;
    Widget menu;
    Widget row_col;
    Widget canvas;
    Widget frame;
    Widget msg_win;
    Widget cur_pos;
    Widget color_bar;
    Widget color_min;
    Widget color_max;
    Widget input_min;
    Widget input_min_pos;
    Widget input_max;
    Widget input_max_pos;
    Widget msg_label;
    Widget icons;
    Pixmap color_pix;
} WidStruct;

/* RCAS == Read Compress And Scale */
typedef struct _RCASParameters
{
    int used; /* is this RCAS currently in use? */
    RCASAction action; /* what to do */
    char *file;        /* input file */
    char *vectfile;    /* 9/16/94 -- added for partition handling */
    int NumRowPtns; /* size of row ptn vector */
    int *RowPtn;
    int NumColPtns;
    int *ColPtn;
    int UsedRowPtns;
    int UsedColPtns;
    int *NewColPtn;
    int *NewRowPtn;
    DaMatrix *input;   /* input matrix */
    DenseMatrix *comp; /* compression matrix */
    CompressionStrategy *cstrat; /* how to compress? */
    ScaleStrategy *sstrat;       /* how to scale? */
    int conformal; /* conform to matrix? or screen? */

    /* sub matrix support */
    int input_startrow; /* starting input row of interest */
    int input_startcol; /* starting input column of interest */
    int input_endrow; /* final input row of interest */
    int input_endcol; /* final input column of interest */

    /* returned parameters */
    double min; /* minimum value in the compression matrix */
    double max; /* maximum value in the compression matrix */
    /* absmin/max only needed for scale */
    double absmin; /* minimum value wrt absolute value in the comp. matrix */
    double absmax; /* maximum value wrt absolute value in the comp. matrix */
    int zero_in_input; /* was zero seen in input? */
    int log10; /* use log base 10 scaling? */

    /* data about the input */
    double input_min; /* minimum value in source matrix */
    int input_min_col;
    int input_min_row;

    double input_max; /* maximum value in source matrix */
    int input_max_col;
    int input_max_row;

    double input_absmin; /* |min value| in source matrix */
    int input_absmin_col;
    int input_absmin_row;

    double input_absmax; /* |max value| in source matrix */
    int input_absmax_col;
    int input_absmax_row;

    /* X stuff */
    int mapped;  /* has the current window been mapped to the display? */
    int bitmap; /* are there only 2 values in the input */
    int pixalloc; /* has a pixmap been allocated? */
    Pixmap pix;        /* output pixmap */
    int startcolor; /* starting color */
    int colormap; /* colormap chosen */
    int numcolors; /* colors are in the range 
		      [startcolor,startcolor+numcolors-1] */
    int force_black; /* if > 0, force scaling 0 values to be black */
    int what_is_black; /* "black" color if force_black chosen */
    int screen_width; /* cols */
    int screen_height; /* rows */
    Window window; /* the window to draw to */
    Display *display; /* display that holds pixmap */
    GC *gcs; /* array of GCs used for drawing output matrix, one per color */
    WidStruct wids;    /* widgets needed */
    int *screen_array; /* screen array used for printing */
} RCASParameters;

/* zoom stuff */
typedef struct _ZoomParms
{
    /* row/col values of the submatrix to be zoomed */
    int input_startrow;
    int input_endrow;
    int input_startcol;
    int input_endcol;
    
    /* screen width and height of output zoom screen */
    int screen_width;
    int screen_height;

    /* pointer to "parent" RCAS */
    int zoomcount;
} ZoomParms;

#define MAIN_CANVAS_WIDTH 700
#define MAIN_CANVAS_HEIGHT 700
#define ZOOM_CANVAS_WIDTH 500
#define ZOOM_CANVAS_HEIGHT 500

/* end zoom stuff */
/* types of color maps to get */
#define GRAYSCALE 0
#define SPECTRAL 1
#define PETE 2
#define HOT 3
#define COOL 4
#define COPPER 5
#define SUNRISE 6
#define SUNSET 7
#define WEATHER 8
/* #define SPECTRAL2 9 */
#define BLUE_YELLOW 9

/* scaling types and data */
#define SCALE_LINEAR 0
#define SCALE_ABS_LINEAR 1
#define SCALE_LN 2
#define SCALE_LOG_10 3
#define ONE_OVER_LN_10 0.43429448190325182765
#define MAX_SCALING_EXPONENT 300
#define MAX_SCALING_VALUE 1e300
#define MIN_SCALING_VALUE 1e-300

typedef struct _s_color
{
    RGB value;
    int portion;

} s_color;

/*
 * This data structure is used to hold information for each icon.
 */

typedef struct _icon {
	struct _icon *suc;
    Widget frameWidget;				/* Id of Frame widget */
    Widget labelWidget;				/* Id of actual Label widget */
    Pixel normalColor;				/* Color (pixel value) for outset look */
    Pixmap normalPixmap;			/* Pixmap for outset look */
    Pixel selectColor;				/* Color (pixel value) for inset look */
    Pixmap selectPixmap;			/* Pixmap for inset look */
	void (*function)(int flag);		/* Function to activate/deactivate mode */
} Icon;


/* GLOBAL EXTERNS (in globals.c) */
/* data for GetColors  */
extern s_color spectral_values[];
extern s_color gray_values[];
extern s_color pete_values[];
extern s_color hot_values[];
extern s_color cool_values[];
extern s_color copper_values[];
extern s_color sunrise_values[];
extern s_color sunset_values[];
extern s_color weather_values[];
/* extern s_color spectral_values2[]; */
extern s_color blue_yellow_values[];

/* RCAS stuff */
/* extern int zoomcount; */
extern RCASParameters RCAS[];
extern GC foregroundGC;

/* Widgets and other global X objects */

extern Display *canvas_display;
extern Colormap Default_cmap;
extern int canvas_depth;
extern Visual *canvas_visual;

/* old zoom.xbm stuff -- crazy names are to avoid most confilicts */
#define ZoOm_WiDtH 48
#define ZoOm_HeIgHt 48
extern char ZoOm_BiTs[];
/* end of globals */
#define COLOR_BAR_WIDTH 25

/* codes for DrawMat */
#define DRAW_MAT_USE_FILENAME 0
#define DRAW_MAT_USE_MATRIX 1

/* zoom stuff */
#define MAXZOOM 10
/* offset for humans to specify matrix points (i.e. internally emily
 * uses 0 based addressing for matrix coords, but for people it'll start
 * at STARTING_OUT_MAT_ADDR (7/27/94)
 */
#define STARTING_OUT_MAT_ADDR 1

/* PROTOTYPES START HERE -- MUST BE AT THE END OF THE FILE */
/* defined in xdraw.c */
int    DrawMat(int argc, char *argv[], int code, DaMatrix *input_mat, 
	       int nptns, int *ptn);
void	CanvasClearWindow(int zoomcount);
void	CanvasDrawPoint(int x, int y, int col);
void	CanvasErasePoint(int x, int y, int col);
void	CanvasDrawLine(int x1, int y1, int x2, int y2, int col);
void	CanvasXOR(int flag);
void	InitializeModes(int zoomcount);

void	MakeCanvas(Widget parent, int zoomcount);
void	MakeCursorPos(Widget parent, int zoomcount);
void	MakeColorBar(Widget parent, int zoomcount);
void	MakeMenus(Widget parent, int zoomcount);
void	MakeMsgWin(Widget parent, int zoomcount);

/* icons: the one exception to the Make*(widget, zoomcount) template */
void	MakeIcons(Widget parent, int zoomcount);
void	MakeIcon(void (*f)(int), char *file, int zoomcount);

void	SetMsgWinString(char *s, int zoomcount);
void	FillColorBar(Widget colorbar, int zoomcount);
void    SetColorMinMax(double min, double max, double inmin, double inmax, int minrow, int mincol, int maxrow, int maxcol, int zoomcount);
void    LoadColorTable(RGB colors[], Colormap c);
void getGC(Display *display, Window win, GC *gc_ptr, unsigned long c);
void GetColors(int type, RGB *colors);
void CopyPixToCanvas( Display *display, Pixmap pix, Window w, Dimension height, Dimension width, GC gc );
void GetMapping(int type, Colormap c);

/* callbacks */
/* void load_pixmap(Widget dialog, XtPointer client_data, XmFileSelectionBoxCallbackStruct *cbs); */
void load_pixmap(Widget dialog, XtPointer client_data, XtPointer call_data);
void CWexpose(Widget w, XtPointer client_data, XtPointer call_data); 
void fileMenuCallback(Widget, XtPointer, XtPointer);
void ColorMapCallback(Widget, XtPointer, XtPointer);
void ScaleMenuCallback(Widget, XtPointer, XtPointer);
void kts_cb(Widget w, XtPointer client_data, XtPointer call_data);
void BackgroundColorCallback(Widget, XtPointer, XtPointer);
void CWresize(Widget w, XtPointer client_data, XtPointer call_data); 
void input(Widget, XtPointer, XtPointer);

extern void (*CanvasInputHandler)(Event);
void ShowRCASResults( RCASParameters *RCAS );
void		click(Widget, XEvent *, String *, Cardinal *);
/* Icon *	makeIcon(Widget parent, int i, void (*f)()); */
void BuildChildrenWidgets( int zoomcount );
void SetupFormWidget(Widget row_col, Widget menu, Widget msg, Widget canvas);

/* defined in rcas.c */
int ReadCompressAndScaleMat( RCASParameters *rcas );
void InitRCASParameters( RCASParameters *rcas );
int BigDrawMat(int argc, char **argv, int code, DaMatrix *input_mat, 
	       int nptns, int *ptn);
void BigDrawMat2(char *s, int code, DaMatrix *input_mat, int nptns, int *ptn);
int AllocRCAS( RCASParameters rcas[], int max );
void DeallocRCAS( RCASParameters rcas[], int member, RCASNuke nuke );
void FreeRCAS( RCASParameters *rcas );

/* defined in xhelpers.c */
void HBZoomHandler(Event e);
void HBZoomHookup(int flag);
void DefaultInputHandler(Event e);
int InWidStruct( Widget w, WidStruct *wids);
int GetZoomcount( Widget w );
/* Widget GetTopShell(Widget w); */
void CloseZoomWindow(int zc );
void CloseAllZoomWindows();
void SetCursor(int shape, Cursor *cursor);
void ResetCursor(Cursor cursor);
void XErrorMsg(char *s);
double GetMatrixValue(int zc, int irow, int icol);
void PointInfo(int zc, int col,int row,double value, double matval, int color);
void FindDecentVisual( Display *d, int screen, int vis_class, int *cmap_size,
		       Visual **v);

/* defined in zoom.c */
int ZoomMat( ZoomParms *z );

/* defined in print_dialog.c */
void PrintDialog(Widget parent, int zc);
void print_cancel_callback(Widget widget, XtPointer client_data, XtPointer call_data);
void print_callback(Widget widget, XtPointer client_data, XtPointer call_data);
void print_toggle_callback(Widget widget, XtPointer client_data, XtPointer call_data);

/* defined in print_ps.c */
void WritePostScriptHeader(FILE *outfp, char *name, 
			   int bb_left, int bb_bot, int bb_right, int bb_top);
void WritePostScriptCode(FILE *outfp, int bb_left, int bb_bot, int desired_h,
			 int desired_w, int matrix_size);
/* void WritePostScriptBlock(FILE *outfp,int xdim,int ydim,int *array); */
 void WritePostScriptBlock(FILE *outfp,int xdim,int ydim,int *array, int dup);
void WritePostScriptFooter(FILE *outfp);
void PrintArray( FILE *outfp, int *array, int cols, int rows, char *name, int colormap);

/* defined in scale_array.c */
int ScaleArray( DenseMatrix *comp, int *array, ScaleRequest *s, int colormap);

/* generic interface code */
void run_emily( int nrows, int ncols, int nnz, 
		double *vals, int *rowptrs, int *colinds,
		int nptns, int *ptn);

/* ARGC_PTR for X stuff */
#if XtSpecificationRelease > 4
#define ARGC_PTR (int *)
#else
#define ARGC_PTR (unsigned int *)
#endif

#endif
