/*
 * File: hbin.inter.c -- interface file to H-B input routine for matlab
 *
 * References:
 *  Users' Guide for the Harwell-Boeing Sparse Matrix Collection (Release I),
 *     Duff, Grimes, and Lewis, publisher?, Oct. 1992
 *
 *  Matlab External Interface Guide, The Math Works, Natick, Mass, Aug. 1992
 *
 * Includes: mexFunction for hbin()
 *
 * Started: 2/10/94
 *
 * Modified: 2/12/94 -- clean up comments
 *           3/22/94 -- added support for RHS, G, and X
 *           ...
 *           6/29/94 -- make RHS guess and RHSs col vectors.
 *
 */

#define V4_COMPAT
#include <hb.h>


/*
 * Function: mexFunction
 *
 * Purpose: to provide an interface between HB input routines and matlab
 *
 * Usage: [M,RHS,G,X]=hbin('matrix_file')
 *        where M = the matrix
 *              RHS = the right hand sides
 *              G = the guess vector
 *              X = the exact solution 
 *
 * Parameters: STRING_IN - name of the input file
 *
 * Return Values: MATRIX_OUT - a sparse matrix
 * 
 * Calls: mexErrMsgTxt mxGetM mxGetN mxIsString malloc mxGetPr ReadHBFile
 *        mxCreateSparse mxSetPr mxSetPi mxSetJc mxSetIr
 *
 * Notes: this is unintelligible code without the Matlab External
 *        Interface Guide.  Also, this code does very little actual
 *        work -- it's all done by either matlab or ReadHBFile
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started:   2/10/94
 *
 * Modified:  2/12/94 -- add comments
 *           3/22/94 -- added support for RHS, G, and X
 *
 */

/* determine matlab MexFunction type depending on -DMATLAB_42 */
#ifdef MATLAB_42
void 
#else
int 
#endif
#ifdef __STDC__
mexFunction(
	int		nlhs,
	Matrix	*plhs[],
	int		nrhs,
	Matrix	*prhs[]
	)
#else
void mexFunction(nlhs, plhs, nrhs, prhs)
int nlhs, nrhs;
Matrix *plhs[], *prhs[];
#endif
{
    double	*data;
    int	rows,width,i;
    int sparse;
    char *s;
    int d;
    int nzmax = 0;
    int *col_indices = (int *) NULL;
    int *row_indices = (int *) NULL;
    struct res r;
    char unread[100];
    HBMatrix hb;

    /* Check for proper number of arguments */

    if (((nrhs < 1) || (nrhs > 4)) && (nlhs != 1))
    {
	mexErrMsgTxt("usage: [M,R,G,S] = hbin(s), where s is a string");
    }

    /* get row and column info */
    rows = mxGetM(STRING_IN);
    width = mxGetN(STRING_IN);

    /* check for input string */
    if (!mxIsString(STRING_IN) && (rows != 1))
    {
	mexErrMsgTxt("usage: [M,R,G,S] = hbin(s), where s is a string");
    }
    
    data = mxGetPr(STRING_IN);

    /* transform input data to a string */
    /* i'm pretty sure there's a matlab function for this, but oh, well */

    s = malloc( width + 1 ); /* OK for malloc here */
    for (i = 0; i < width; i++)
    {
	d = (int) data[i];
	s[i] = (char) d;
    }
    s[width] = '\0';
    
    /* read the file */
    if (ReadHBFile(s, &hb, 1) == OK)
    {
	/* all was well */
	/* set real part to values */
	MATRIX_OUT = mxCreateSparse(hb.head.NumRows,     /* M */
				    hb.head.NumCols,     /* N */
				    hb.head.NumRowIndex, /* NZ */
				    0);                  /* complex? */

	/* set up real values */
	mxSetPr(MATRIX_OUT, hb.values); 
	mxSetPi(MATRIX_OUT, (double *) NULL); /* no complex vals */
	/* jc = column indices */
	mxSetJc(MATRIX_OUT, hb.col_inds);
	/* ir = row indices */
	mxSetIr(MATRIX_OUT, hb.row_inds);

	if (nlhs == 1)
	    /* return(0);*/ 
	    return;

	/* mxCreateFull(hb.head.NumRows, hb.head.NumCols, complex_flag) */
	/* now for the RHS */
        if ((hb.head.NumRHS <= 0) || (hb.RHS == (double *) NULL))
	{
	    RHS_OUT = mxCreateFull(0,0,0);
	    mxSetPr(RHS_OUT, (double *) NULL);
	    mxSetPi(RHS_OUT, (double *) NULL); /* no complex vals */
	    if (nlhs >= 3)
	    {
		GUESS_OUT = mxCreateFull(0,0,0);
		mxSetPr(GUESS_OUT, (double *) NULL);
		mxSetPi(GUESS_OUT, (double *) NULL); 
		if (nlhs == 4)
		{
		    SOLUTION_OUT = mxCreateFull(0,0,0);
		    mxSetPr(SOLUTION_OUT, (double *) NULL);
		    mxSetPi(SOLUTION_OUT, (double *) NULL); 
		    strcpy(unread,"RHS, guess, and solution");
		}
		else
		    strcpy(unread,"RHS and guess");
	    }
	    else
		strcpy(unread,"RHS");
	    fprintf(stderr,"No RHS data was read in.\n  The %s vector%s were assigned null values.\n",unread, ((nlhs == 2)?" ":"s "));
	    /* return(0); */
	    return;
	}
	/* set up RHS (Nrows x Num_rhs) column vector */
	/* RHS_OUT = mxCreateFull(1,hb.head.NumCols,0); */

	/* RHS_OUT = mxCreateFull(hb.head.NumRHS,hb.head.NumCols,0);*/
	RHS_OUT = mxCreateFull(hb.head.NumRows,hb.head.NumRHS,0);

	/* set real data to the RHS data read in */
	mxSetPr(RHS_OUT, hb.RHS);
	mxSetPi(RHS_OUT, (double *) NULL); /* no complex values */
	
	/* if only 2 input vectors, return OK */
	if (nlhs == 2)
	    /* return(0); */ 
	    return;

	/* if there's no input guess vector, set guess vector to null */
	if (hb.guess == (double *) NULL)
	{
	    GUESS_OUT = mxCreateFull(0,0,0);
	    mxSetPr(GUESS_OUT, (double *) NULL);
	    mxSetPi(GUESS_OUT, (double *) NULL); /* no complex vals */
	    fprintf(stderr,"No guess vector was read in.\n Guess vector assigned a null value\n");
	} 
	else /* set up the guess (Nrows x Num_rhs) column vector */
	{
	    /* set up Guess */
	    /* GUESS_OUT = mxCreateFull(1,hb.head.NumCols,0); */
	    /* GUESS_OUT = mxCreateFull(hb.head.NumRHS,hb.head.NumCols,0); */
	    GUESS_OUT = mxCreateFull(hb.head.NumRows,hb.head.NumRHS,0); 
	    mxSetPr(GUESS_OUT, hb.guess);
	    mxSetPi(GUESS_OUT, (double *) NULL); /* no complex values */
	}

	/* if only 3 LHS's, return OK */
	if (nlhs == 3)
	    return;
	    /* return(0); */

	/* else if the exact data isn't present */
	if (hb.exact == (double *) NULL)
	{
	    /* return a null vector */
	    SOLUTION_OUT = mxCreateFull(0,0,0);
	    mxSetPr(SOLUTION_OUT, (double *) NULL);
	    mxSetPi(SOLUTION_OUT, (double *) NULL); 
	    fprintf(stderr,"No solution vector was read in. The solution vector is assigned a null value\n");
	}
	else  /* return the exact soln. column (NRows x NRhs)  vector */
	{
	    /* SOLUTION_OUT = mxCreateFull(1,hb.head.NumCols,0); */
            /* SOLUTION_OUT=mxCreateFull(hb.head.NumRHS,hb.head.NumCols,0); */
	    SOLUTION_OUT = mxCreateFull(hb.head.NumRows,hb.head.NumRHS,0); 
	    mxSetPr(SOLUTION_OUT, hb.exact);
	    mxSetPi(SOLUTION_OUT, (double *) NULL); /* no complex values */
	}
	/* return(0); */
	return;
    }
}


