/*
 * hbmat.h -- header file for HB sparse matrix input/output with matlab
 * 
 * HB (Harwell Boeing) format from "User's Guide for the Harwell-Boeing
 *    Sparse Matrix Collection (Release 1)", Duff, Grimes, and Lewis,
 *    October 1992.
 * 
 * Matlab info from Matlab External Interface Guide, The Math Works, August
 *    1992
 *
 * Note: Use this file for new code.
 * 
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: 2/10/94 
 *
 * Modified: 2/12/94 -- clean up comments
 *           2/13/94 -- add support for hbdraw.h
 *           ...
 *           3/18/94 -- added debugging level/trace macros
 */

#ifndef HBMAT_H
#define HBMAT_H

#ifdef STANDALONE
#define RETURN_PROMPT 1
#else 
/* matlab code */
#include "mex.h"
#define RETURN_PROMPT 0
/* Input Arguments */
#define	STRING_IN	prhs[0]
#define	BSTART_IN	prhs[1]
/* #define	Y_IN	prhs[1] */


/* Output Arguments */
#define	MATRIX_OUT      plhs[0] 
#define	RHS_OUT         plhs[1] 
#define	GUESS_OUT       plhs[2] 
#define	SOLUTION_OUT    plhs[3] 

/* #define	max(A, B)	((A) > (B) ? (A) : (B)) */
/* #define	min(A, B)	((A) < (B) ? (A) : (B)) */

#define NON_ZERO_TOL 1e-6
#define MAT_FULL 0
#define MAT_SPARSE 1

/* this is from matstat.  unused in hb, but maybe will be... */
struct res 
{
    double max;
    double min;
    double absmin;
    double absmax;
    double sum;
    double abssum;
    int nonzero;
    
};
#endif  /* of undef STANDALONE */

/* stdio  definitions */
#define STDIN_FD 0
#define STDOUT_FD 1
#define STDERR_FD 2
#define STDIN_ARG "-"

/* debugging macros (DEBUG_LEVEL and TRACE_ON are defined in hb.h) */
/* higher number is more specialized */
#define DEBUG(lvl,s) if (DEBUG_LEVEL >= lvl) fprintf(stderr,s)
#define DEBUG1(s) DEBUG(1,s)
#define DEBUG2(s) DEBUG(2,s)
#define DEBUG3(s) DEBUG(3,s)
#define TRACE(s) if (TRACE_ON) fprintf(stderr,s);

/* utility macros */
#define myabs(a) (a>=0?a:-a)
#define mymax(a,b) (a>b?a:b)
#define mymin(a,b) (a>b?b:a)
#define myfloor(a) (a>=0?(int) a:(int) a-1)
#define myceil(a) (a>=0?(int) a+1:(int) a)
/* double myfp(int/float/char/double a) == fractional part of a */
#define myfp(a) ((double) (myabs(a) - (int) myabs(a)))
/* IDXRANGE(a,b) returns the number of different array indices between a 
   and b */
#define IDXRANGE(a,b) myabs( (b-a) )+1
/* RANGE1(a,b) returns the range of values covered between a and b */
#define RANGE1(a,b) myabs( (b-a) )
/* RELPOS(cur, min, max) returns the relative position of cur between min
   and max */
/*#define RELPOS(cur, min, max) ((double) RANGE1(min,cur) / (double) RANGE1(min,max)) */
#define RELPOS(cur, min, max) ((double) (cur-min) / (double) RANGE1(min,max))

/* SCALE(cur,min,max,scaleamt) scales cur between min and max  and multiplies it by scaleamt */
#define SCALE(cur, min, max, scaleamt) (RELPOS(cur,min,max)*scaleamt)
/* SWAP(a,b,t) swaps a and b using t as temporary storage */
#define SWAP(a,b,t) t=a;a=b;b=t

/* ENFORCE_RANGE(in,min,max) forces in to be between min and max */
#define ENFORCE_RANGE(in,min,max) ((in>min)?((in<max)?in:max):min)
/* PARTITION(val,ptns,max,answer) returns in answer which partition val is in */
#define PARTITION(val,ptn,max,answer) answer = 0; for (i = 0; i < max; i++) { if ((ptn[i] <= val) && (ptn[i] > val)) { answer = i; break; } }
/* readhb defines */
/* size of the static big input buffer: 8294400 = 100 * 1024 * 81 char/line */
#ifdef SMALL_STACK
#define BIG_BUF_SIZE 2048000
#else
#define BIG_BUF_SIZE 8294400
#endif
/* value to be used in a pattern matrix */
#define DEFAULT_PATTERN_VALUE 1.0

/* typedef double FPType; // default floating point type */

/* HB matrix constants */
/* min. number of lines in a HB header */
#define MIN_HB_HEADER_LINES 4
/* max. number of lines in a HB header */
#define MAX_HB_HEADER_LINES 5
/* assumed maximum size of an input line */
#define MAX_LINE_SIZE 101

/* HB header constants */

/* how long a title is */
#define TITLELEN 72
/* how long a key is */
#define KEYLEN 8

/* how many integers are in line 2 of an HB matrix */
#define LINE2_COUNT 5

/* how long an integer is in an HB matrix */
#define HEADER_INT_LEN 14

/* how many spaces to blow off in line 3 */
#define LINE3_SPACE_COUNT 11

/* format lengths for row 4 */
#define LINE4_POINTER_FORMAT_LEN 16
#define LINE4_ROW_FORMAT_LEN 16
#define LINE4_VALUE_FORMAT_LEN 20
#define LINE4_RHS_FORMAT_LEN 20

/* where the counts start in line 5 */
#define LINE5_RHS_COUNT_START 14

/* output format statements */
/* line 1: A72,A8 */
/* line 2: 5I14 */
/* line 3: A3 (%c%c%c), 11X, 4I4 */
/* line 4: 2A16,2A20 ("(%d%c%d)/(%d%c%4.1f)(formats) */
/*         assume the formats are assembled elsewhere into strings */
/* line 5: A3 (%c%c%c), 11X, 2I14 */
extern char *HeaderFormats[MAX_HB_HEADER_LINES];
 
#define FORMAT_HEAD1 "%-72s%-8s\n"
#define FORMAT
/* error codes */
#define NOERR 0
#define IOERR -1
#define ARGERR -2 
#define MEMERR -3
#define BUGERR -98
#define UNSUPERR -99

/* enums for matrix type */

/* first character of line 3 -- type of data to be found */
enum OHMY
{
    IsReal = 0,     /* R */
    IsComplex = 1,  /* C */
    IsPattern = 2   /* P */
};

typedef enum OHMY NumericalDataType;


/* second character of line 3 -- type of symmetry in the matrix */
enum two
{
    IsSymmetric = 0,      /* S */
    IsUnsymmetric = 1,    /* U */
    IsHermetian = 2,      /* H */
    IsSkewSymmetric = 3,  /* Z */
    IsRectangular = 4     /* R */
};
typedef enum two SymmetryType;

/* third character of line 3 -- assembled or elemental */

enum three
{
    IsAssembled = 0,  /* A */
    IsElemental = 1   /* E */
};
typedef enum three AssemblyType;

/* first character of line 5 -- RHS Vector Format Type */
enum four
{
    IsFull = 0, /* F */ 
    IsSame = 1 /* M */
};
typedef enum four RHSVectorFormatType;

/* 
 * Fortran format parsing data
 */
typedef struct
{
    int count;
    char format;
    int width;
    int decimal_places;
} Format;

/* 
 * NOTE: The values in this definition are IN THE ORDER of the disk
 *       data.  If alterations need to be made, please maintain this order.
 */
typedef struct HBHeaderStruct
{
    /* line 1 (A72, A8) */
    char Title[75];
    char Key[10];

    /* line 2 (5I14) */
    int TotalLines;  /* TOTCRD */
    int PointerLines; /* col pointers  (PTRCRD) */
    int IndexLines; /* row indices (INDCRD) */
    int ValueLines; /* VALCRD */
    int RHSLines; /* RHSCRD */

    /* line 3 (A3, 11X, AI14) */
    NumericalDataType NumericalData; /* MXTYPE */
    SymmetryType Symmetry;           /*  ...   */
    AssemblyType Assembly;           /*  ...   */

    int NumRows; /* NROW */
    int NumCols; /* NCOL */
    int NumRowIndex; /* NNZERO */
    int NumElemMatEntries;  /* NELTVL */

    /* line 4 (2A16, 2A20) */
    Format PointerFormat; /* from PTRFMT */
    Format RowIndexFormat; /* from INDFMT */
    Format ValueFormat; /* from VALFMT */
    Format RHSFormat; /* from RHSFMT */

    /* line 5 (A3,11X,2I14) */
    RHSVectorFormatType RHSVectorFormat;
    char GuessSupplied; /* G if a guess is supplied (RHSTYP) */
    char ExactSupplied; /* X if an exact solution is supplied */
    int NumRHS; /* number of right hand sides (NRHS) */
    int NumRHSRowIndex; /* number of row indices (NRHSIX) */
    
    /* calculated values */
    double MaxValue;
    double AbsMaxValue;
    double MinValue;
    double AbsMinValue;
} HBHeader;

/*
 * File section numbers 
 */
#define POINTER_SECTION 0
#define ROW_SECTION 1
#define VALUE_SECTION 2
#define RHS_SECTION 3
#define GUESS_SECTION 4
#define SOLUTION_SECTION 5
/* MAX_SECTIONS must be one too big! (body.c) */
#define MAX_SECTIONS 6

#ifdef HANDLE_PTNS
#define MAX_PTNS 5000 /* should be enough */
#endif

/*
 * The HB matrix definition 
 */
typedef struct
{
    /* header */
    HBHeader head;

    /* data */
    int *col_inds;
    int *row_inds;
    double *values;
    double *RHS; /* RHS's */
    double *guess; /* starting guesses */
    double *exact; /* exact solutions */
}  HBMatrix;

/*
 * Dense matrix definition
 */
/* 
 * Indexing of a 1D array:
 *   if row_major (C format)
 *      idx = row * ncols + col;
 *   if col_major (FORTRAN format)
 *      idx = col * nrows + row;
 *   assuming 0 based addressing 
 */
#define NORMAL_1D 1
#define LT_1D 2
#define UT_1D 3
typedef struct
{
    int NumRows;   /* number of rows in the matrix */
    int NumCols;   /* number of cols in the matrix */
    int one_d;     /* is matrix stored in one-d format */
    /* Note: if one_d == 2, this implies a lower triangular system */
    /*          one_d == 3, this is an upper triangular system */
    int row_major; /* is matrix stored in row major format? */
    union
    {
	double *d1;
	double **d2;
    } data; 
} DenseMatrix;


/*
 * MatrixType definition 
 */
enum five
{
    HB = 0,
    Dense = 1
    /* SegArray = 2 */
};

typedef enum five MatrixType;

/*
 * DaMatrix -- the generic matrix structure 
 */

typedef struct
{
    MatrixType type; /* type of matrix: sparse(HB) or dense */
    union { /* the actual matrix */
	HBMatrix *hb;
	DenseMatrix *d;
    } data;
} DaMatrix;

/* some shorthands */
#define hb_head data.hb->head
#define hb_values data.hb->values
#define hb_row_inds data.hb->row_inds
#define hb_col_inds data.hb->col_inds
#define hb_RHS data.hb->RHS
#define hb_guess data.hb->guess
#define hb_exact data.hb->exact
#define hb_NumCols data.hb->head.NumCols
#define hb_NumRows data.hb->head.NumRows

#define d_data data.d->data
#define d_NumCols data.d->NumCols
#define d_NumRows data.d->NumRows
#define d_one_d data.d->one_d
#define d_row_major data.d->row_major

#define sa_count data.sa->count
#define sa_min data.sa->min
#define sa_max data.sa->max
#define sa_segstart data.sa->segstart
#define sa_segend data.sa->segend

/* DFA defines */
/* fortran format is of the form:
 * 
 * ( [n1 f1] [n2] f2 n3 )
 *
 * where n1, n2, and n3 are (potentially) floating point numbers
 *     of the form xx.yy  where xx = field width and yy = digits behind 
 *     the decimal point (d.p.)
 * and f1, f2 are format characters
 *
 */

typedef enum
{
    ST_DEAD = -1,     /* the dead state */
    ST_INIT = 0,
    ST_LPAREN = 1,    /* left paren state */
    ST_WIDTH = 2,     /* for a number, the part before the d.p. */
    ST_PRECISION = 3, /* the part of a number after the d.p. */
    ST_FORMAT = 4,    /* state of reading in format chars */
    ST_DONE = 5       /* finished OK state */
} FormatStates;

#define MAX_FORMATS 2
#define MAX_NUMBERS 3
/* valid uppercase FORTRAN format chars (for this program, anyway) */
/* char *ValidFormatChars = "IEPFDQ"; */


/* generic return values */
#define OK 0
#define BAD (-1)

/* generic name of program */
#define DEFAULT_NAME "emily"

/*
 * Function prototypes  -- THESE HAVE TO BE AFTER THE OTHER HEADER DEFN'S
 */
/* read_main.c */
/* int main(int argc, char **argv); */

/* util.c STANDALONE stuff*/
#ifdef STANDALONE
void error(char *string, int rc);
#define MALLOC(count,size) malloc(count*size)
#define FREE free
#else /* matlab case */
#ifdef MATLAB_42
void error(char *string, int rc);
#else
#define error(st,rc) printf("%s\n",st); return(rc) 
#endif
#define MALLOC(count,size) mxCalloc(count,size)
#define FREE mxFree
#endif

/* util.c */
#ifndef _NO_PROTO
void HBCleanup( HBMatrix *hb );
int FormatInt( char *inbuff, int size);
double FormatFloat( char *inbuff, int size );
void FormatNumDataType( char c, NumericalDataType *n);
void FormatSymType( char c, SymmetryType *n);
void FormatAssemType( char c, AssemblyType *n);
int GetInputLine(int fd, char *buff);
int Make2DArray( double ***ptr, int rows, int cols, int row_major);
void Free2DArray( double **ptr, int rows);
void HBInit( HBMatrix *hb );
int GetBuffLine(char *inbuff, char **outbuff, int max, int *nl_seen);
char *Get1DArray(int count, int size, char *name, HBMatrix *hb);
DenseMatrix *GetDenseMatrix( int nrows, int ncols, int row_major, int one_d, double initval );
void FreeDenseMatrix( DenseMatrix *d );
void FreeDaMatrix( DaMatrix *d );


/* header.c */
int ReadHBHeader( int fd, HBHeader *hb, int *count );
void DumpHBHeader( HBHeader *hb);

/* parser.c */
void FormatToString( Format *f, char *s );
extern int ParseFormat(char *inbuff, int size, Format *f);

/* body.c */
int ReadHBBody( int fd, HBMatrix *hb, int use_c );
int ProcInts(Format *format, char *buffer, int *array, int *lines, int bufsize);
int ProcDoubles(Format *format, char *buffer, double *array, int *lines,
		double *max, double *min, double *absmax, double *absmin, 
		int bufsize);
int ReadHBFile(char *fn, HBMatrix *mat, int use_c);
void DumpHBMatrixColumnRange( int lo_col, int hi_col, HBMatrix *hb, int use_c );
#ifdef HANDLE_PTNS
/* int ReadPtnVector(int ptnfd, int *ptn_vect, int *num_ptns); */
int ReadPtnFile(char *fn, int *ptn_vect, int *num_ptns);
int GetPtnCount(char *fn, int *num_ptns, int *chaco_input); 
int GetPtnVect(char *fn, int *ptn_vect, int num_ptns, int chaco_input); 
#endif
 /* cbody.c */
int CProcInts(Format *format, char *buffer, int *array, int *lines, int bufsize);
/* mytime.c */
double mytime();

/* end of ifdef HBMAT_H */
#endif 
#endif 


