/*
 * File: header.c
 *
 * References:
 *  Users' Guide for the Harwell-Boeing Sparse Matrix Collection (Release I),
 *     Duff, Grimes, and Lewis, publisher?, Oct. 1992
 * 
 *  NOTE: THIS REFERENCE IS INVALUABLE FOR UNDERSTANDING THIS FILE!
 *
 * Includes: ReadHBHeader and DumpHBHeader
 *
 * Started: 2/4/94
 *
 * Modified: 2/12/94 -- clean up comments
 *           3/17/94 -- add comments about each line from ref. above 
 *           3/18/94 -- add TRACE/DEBUG code
 */
#include <hb.h>


/*
 * Function: ReadHBHeader 
 * 
 * Purpose: Read in the header of a Harwell-Boeing matrix
 * 
 * Usage: int fd;
 *        HBMatrix hb;
 *        int count, err, use_c;
 *        
 *        fd = open("matrix_file",O_RDONLY);
 *        use_c = 1;
 *        if (fd > 0)
 *        {
 *            // read in header 
 *            if ((err = ReadHBHeader( fd, &(hb.head), &count)) == OK)
 *                err = ReadHBBody( fd, &hb, use_c);
 *        }
 *            
 * Parameters: fd = a file descriptor pointing to the input file
 *             hb = the hb header structure (see hb.h)
 *             count = number of characters processed from inbuff
 *
 * Return Values:  OK - if header read in OK
 *                !OK - if header not read in OK (error() will be called)
 * 
 * Calls: GetInputLine strncpy error FormatInt FormatDouble atof 
 *        FormatNumDataType FormatAssemType FormatSymType ParseFormat
 *        toupper
 *
 * Notes: this code is almost unintelligible without reading the 
 *        HB header format info in the references 
 * 
 * Programmer: Tom Loos, LH310, 812 855-3609, tloos@cs.indiana.edu
 *
 * Started: 2/4/94
 * 
 * Modified: 2/12/94 -- add/clean up  comments
 *           3/1 add old header check
 *           ...
 *           6/28/94 -- fix line 5 bugs.
 */
int ReadHBHeader( int fd, HBHeader *hb, int *count )
{
    int lines_to_parse = MIN_HB_HEADER_LINES;
    int curcount, linecounter;
    char inbuff[MAX_LINE_SIZE];
    char tempbuffer[MAX_LINE_SIZE];
    char formatstr[20];
    char debugmsg[100];
    char is_old_format;

    TRACE("In ReadHBHeader\n");
    /* line 1 (A72, A8) */
    /* Col  1-72 Title of matrix */
    /*     73-80 Key */

    if ((curcount = GetInputLine(fd,inbuff)) > 0)
    {
	*count = curcount;
	sprintf(debugmsg,"line 1 = %s, len = %d\n",inbuff, strlen(inbuff));
	DEBUG2(debugmsg);
	if (curcount < TITLELEN+KEYLEN)
	{
	    error("Couldn't get line 1 of the header\n",IOERR);
	}

    
	/* do title */
	strncpy(hb->Title, inbuff, TITLELEN);

	/* and the key */
	strncpy(hb->Key, inbuff+TITLELEN, KEYLEN);
	/* HACK */
	hb->Key[KEYLEN] = '\0';
	
	
	sprintf(debugmsg,"Line 1\nTitle = %s\n",hb->Title);
	DEBUG2(debugmsg);
	sprintf(debugmsg,"Key = %s, TL = %d, KL = %d\n",hb->Key, TITLELEN, 
		KEYLEN);
	DEBUG2(debugmsg);
    }
    else
    {
	error("Couldn't read in line 1 of the header\n",IOERR);
    }

    /* line 2: (5I14) */
    /* Col  1-14 Total number of lines (in the file) excluding header */
    /*     15-28 Number of lines for pointers */
    /*     29-42 Number of lines for row (or variable) indices */
    /*     43-56 Number of lines for numerical values */
    /*     57-70 Number of lines for right hand sides (including starting */
    /*             guesses and soln. vectors if present -- 0 if no RHS data */

    if ((curcount = GetInputLine(fd,inbuff)) > 0)
    {
	*count += curcount;
	linecounter = 0;
	if (curcount < (5 * LINE2_COUNT))
	{
	    error("Couldn't get line 1 of the header\n",IOERR);
	}
	*count += curcount;

	sprintf(debugmsg,"line 2 = %s\n",inbuff); 
	DEBUG2(debugmsg);

	hb->TotalLines = FormatInt(inbuff, HEADER_INT_LEN);
	linecounter += HEADER_INT_LEN;

	hb->PointerLines = FormatInt(inbuff+linecounter, HEADER_INT_LEN);
	linecounter += HEADER_INT_LEN;

	hb->IndexLines = FormatInt(inbuff+linecounter, HEADER_INT_LEN);
	linecounter += HEADER_INT_LEN;

	hb->ValueLines = FormatInt(inbuff+linecounter, HEADER_INT_LEN);
	linecounter += HEADER_INT_LEN;

	hb->RHSLines = FormatInt(inbuff+linecounter, HEADER_INT_LEN);

	if (hb->RHSLines > 0)
	    lines_to_parse = MAX_HB_HEADER_LINES;

	sprintf(debugmsg,"Line 2\nTotal = %d, Max Bytes = %d\n",hb->TotalLines,
	                    hb->TotalLines * MAX_LINE_SIZE);
	DEBUG2(debugmsg);

	sprintf(debugmsg,"Pointer = %d\n",hb->PointerLines);
	DEBUG2(debugmsg);

	sprintf(debugmsg,"Index = %d\n",hb->IndexLines);
	DEBUG2(debugmsg);

	sprintf(debugmsg,"Value = %d\n",hb->ValueLines);
	DEBUG2(debugmsg);

	sprintf(debugmsg,"RHS = %d ==> Header Lines To Parse = %d\n",
		hb->RHSLines, lines_to_parse);
	DEBUG2(debugmsg);

    }
    else
    {
	error("Couldn't read in line 2 of the header\n",IOERR);
    }

    /* line 3 (A3, 11X, 4I14) */
    /* Col  1- 3 Matrix type  */
    /*      1: Data type: real (R), complex (C), or pattern only (P) */
    /*      2: Symmetry: symmetric (S), unsymmetric (U), Hermetian (H),  */
    /*                   skew symmetric (Z) or rectangular (R) */
    /*      3: Assembly type: assembled (A), or elemental/unassembled (E) */
    /*     15-28 Number of rows (or variables)  */
    /*     29-42 Number of columns (or elements)  */
    /*     43-56 Number of row (or variables) indices (equals number of  */
    /*             entries for assembled matrices) */
    /*     57-70 Number of elemental mat. entries (0 for assembled matrices) */

    if ((curcount = GetInputLine(fd,inbuff)) > 0)
    {
	*count += curcount;
	linecounter = 0;
	/* parse the 3 character type field */
	FormatNumDataType( inbuff[linecounter++], &(hb->NumericalData));
	
	FormatSymType( inbuff[linecounter++], &(hb->Symmetry));
	FormatAssemType( inbuff[linecounter++], &(hb->Assembly));
	
	linecounter += LINE3_SPACE_COUNT;

	hb->NumRows = FormatInt(inbuff+linecounter, HEADER_INT_LEN);
	linecounter += HEADER_INT_LEN;

	hb->NumCols = FormatInt(inbuff+linecounter, HEADER_INT_LEN);
	linecounter += HEADER_INT_LEN;

	hb->NumRowIndex = FormatInt(inbuff+linecounter, HEADER_INT_LEN);
	linecounter += HEADER_INT_LEN;

	hb->NumElemMatEntries = FormatInt(inbuff+linecounter, HEADER_INT_LEN);
	
	/* debugging code */
	sprintf(debugmsg,
		"Line 3\nMatrix Type = %d, Symmetry = %d, Assembly = %d\n",
		hb->NumericalData, hb->Symmetry, hb->Assembly);
	DEBUG2(debugmsg);
	sprintf(debugmsg,"Row Count = %d\n",hb->NumRows);
	DEBUG2(debugmsg);
	sprintf(debugmsg,"Column Count = %d\n",hb->NumCols);
	DEBUG2(debugmsg);
	sprintf(debugmsg,"Number of Row Indices = %d\n",hb->NumRowIndex);
	DEBUG2(debugmsg);
	if (hb->Assembly == IsAssembled)
	{
	   sprintf(debugmsg,
		   "This is also the number of data entries for this matrix\n");
	   DEBUG2(debugmsg);
	 }
	sprintf(debugmsg,
		"Number of elemental matrix entries = %d\n",
		hb->NumElemMatEntries);
	DEBUG2(debugmsg);

    }
    else
    {
	error("Couldn't read in line 3 of the header\n",IOERR);
    }

    /* line 4 (2A16, 2A20) */
    /* Col  1-16 Format for pointers */
    /*     17-32 Format for row (or variable) indices */
    /*     33-52 Format for numerical values of coefficient matrix */
    /*     53-72 Format for numerical values of RHS's */

    if ((curcount = GetInputLine(fd,inbuff)) > 0)
    {
	*count += curcount;
	linecounter = 0;
	/* get pointer format */
	ParseFormat(inbuff,
		     LINE4_POINTER_FORMAT_LEN,
		     &(hb->PointerFormat));
	
	FormatToString(&(hb->PointerFormat), formatstr);
	sprintf(debugmsg,"Returned Pointer format is: %s\n",formatstr);
	DEBUG2(debugmsg);
	
	/* get row index format */
	linecounter += LINE4_POINTER_FORMAT_LEN;
	ParseFormat(&(inbuff[linecounter]),
		       LINE4_ROW_FORMAT_LEN,
		       &(hb->RowIndexFormat));
	
	FormatToString(&(hb->RowIndexFormat), formatstr);
	sprintf(debugmsg,"Returned Row Index format is: %s\n",formatstr);
	DEBUG2(debugmsg);

	/* get value format */
	linecounter += LINE4_ROW_FORMAT_LEN;
	ParseFormat(&(inbuff[linecounter]),
		       LINE4_VALUE_FORMAT_LEN,
		       &(hb->ValueFormat));

	FormatToString(&(hb->ValueFormat), formatstr);
	sprintf(debugmsg,"Returned Value format is: %s\n",formatstr);
	DEBUG2(debugmsg);

	/* if we need to get RHS format, do so */
	if (lines_to_parse == MAX_HB_HEADER_LINES)
	{
	    linecounter += LINE4_VALUE_FORMAT_LEN;
	    ParseFormat(&(inbuff[linecounter]),
			LINE4_RHS_FORMAT_LEN,
			&(hb->RHSFormat));

	    FormatToString(&(hb->RHSFormat), formatstr);
	    sprintf(debugmsg,"Returned RHS format is: %s\n",formatstr);
	    DEBUG2(debugmsg);

	}

    } /* end of if we got a good input line */
    else
    {
	error("Couldn't read in line 4 of the header\n",IOERR);
    }
     
    /* line 5 (A3,11X,2I14) */
    /* Col 1 RHS type - F for full storage, M for same as mat */
    /*     2  G if a starting vector is supplied (Guess) */
    /*     3  X if an exact soln. vector(s) are given */
    /*     15-28 Number of RHS's */
    /*     29-42 Number of row indices (ignored in case of unass. mats) */
	   
    if (lines_to_parse == MAX_HB_HEADER_LINES)
    {
	/* check for old format (3/1 -- tjl) */
	read(fd, &is_old_format, 1);
	/* hmmm... convert to upper case 5/9/94 */
	is_old_format = toupper(is_old_format);
	lseek(fd, -1, SEEK_CUR);
	/* if in old format, is_old_format should be something other than
	   an F or M */
	if ((is_old_format != 'F') && (is_old_format != 'M'))
	{
	    hb->RHSVectorFormat = IsFull;
	    hb->GuessSupplied =  ' ';
	    hb->ExactSupplied =  ' ';
	    hb->NumRHS = 1;
	    hb->NumRHSRowIndex = 0;
	}
	    
	else if ((curcount = GetInputLine(fd,inbuff)) > 0)
	{
	    *count += curcount;
	    if (inbuff[0] == 'F')
		hb->RHSVectorFormat = IsFull;
	    else
		hb->RHSVectorFormat = IsSame;
	    /* commented out toupper calls on 5/9 since GetInputLine was 
	       changed */
	    /* 6/28/94 fix 5th line */

	    /* 6/28 -- GetInputLine does conversion to upper case */
	    /*  fix that i'd commented out the next 2 lines by accident */
	    hb->GuessSupplied = inbuff[1];
	    hb->ExactSupplied = inbuff[2]; 
	    linecounter = LINE5_RHS_COUNT_START;
	    hb->NumRHS = FormatInt( inbuff + linecounter, HEADER_INT_LEN);
	    /* 6/28/94:
	     *
	     * From p. 11 of the standard ref:
	     *  "Col 29-42: Number of row indices (NRHSIX)
	     *   (ignored in the case of unassembled matrices)
	     *
	     * So, if the matrix is unassembled (elemental), ignore 
	     * the rest of the line.
	     */
	    if (hb->Assembly == IsElemental) 
	    {
		linecounter += HEADER_INT_LEN;
		hb->NumRHSRowIndex = FormatInt( inbuff + linecounter, HEADER_INT_LEN);
	    }
	}
	else
	{
	    error("Couldn't read in line 5 of the header\n",IOERR);
	}
    }
    /* END!!! */
    TRACE("leaving ReadHBHeader\n");
    return(OK);
}

/*
 * Function: DumpHBHeader
 *
 * Purpose: to make dump the header of a HBMatrix in human readable format
 *
 * Parameters: hb - a pointer to a HB matrix
 *
 * Return Values: none
 * 
 * Calls: printf FormatToString
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started:  about 2/1/94
 *
 * Modified:  2/12/94 - add comments
 *
 */
void DumpHBHeader( HBHeader *hb )
{

    char formatstr[100];

    printf("Harwell-Boeing Format Matrix File Header Dump\n\n");

    /* skip the format control char */
    printf("Title : %s\n",((hb->Title) + 1));
    printf("Key: %s\n",hb->Key);
    printf("Total Lines Excluding Header (TOTCRD) : %d\n",hb->TotalLines);
    printf("      Lines for Pointers (PTRCRD)     : %d\n",hb->PointerLines);
    printf("      Lines for Row Indices (INDCRD)  : %d\n",hb->IndexLines);
    printf("      Lines for Num. Values (VALCRD)  : %d\n",hb->ValueLines);
    printf("      Lines for RHS's                 : %d\n\n",hb->RHSLines);

    /* formatting of matrix type */
    printf("Matrix Type (MXTYPE): The matrix is ");

    /* numerical data type format */
    if (hb->NumericalData == IsReal)
	printf("real, ");
    else if (hb->NumericalData == IsComplex)
	printf("complex, ");
    else if (hb->NumericalData == IsPattern)
	printf("a pattern only, ");
    else
	printf("of erroneous type, ");
    
    /* symmetry format */
    if (hb->Symmetry == IsSymmetric)
	printf("symmetric, ");
    else if (hb->Symmetry == IsUnsymmetric)
	printf("unsymmetric, ");
    else if (hb->Symmetry == IsHermetian)
	printf("Hermetian, ");
    else if (hb->Symmetry == IsSkewSymmetric)
	printf("skew symmetric, ");
    else if (hb->Symmetry == IsRectangular)
	printf("rectangular, ");
    else
	printf("of erroneous type, ");

    /* assembly format */
    if (hb->Assembly == IsAssembled)
	printf("and assembled\n");
    else if (hb->Assembly == IsElemental)
	printf("and elemental\n");
    else 
	printf("and of erroneous type\n");

    printf("Number of rows in matrix (NROWS)      : %d\n",hb->NumRows);
    printf("Number of columns in matrix (NCOLS)   : %d\n",hb->NumCols);
    if (hb->Assembly == IsAssembled)
    {
	printf("Number of entries in matrix (NNZERO)  : %d\n",hb->NumRowIndex);
	if (hb->NumElemMatEntries != 0)
	{
	    printf("Number of elemental matrix entries wrong!!! = %d\n",hb->NumElemMatEntries);
	}
    }
    else
    {
	printf("Number of row indices (NNZERO)        : %d\n",hb->NumRowIndex);
	printf("Number of elemental entries (NELTVL)  : %d\n",hb->NumElemMatEntries);
    }
    FormatToString(&(hb->PointerFormat), formatstr);    
    printf("Pointer format (PTRFMT)               : %s\n", formatstr);

    FormatToString(&(hb->RowIndexFormat), formatstr);
    printf("Row Index format (INDFMT)             : %s\n", formatstr);
    
    FormatToString(&(hb->ValueFormat), formatstr);
    printf("Numerical Values Format (VALFMT)      : %s\n", formatstr);

    if (hb->RHSLines > 0) /* do we have a line 5 of the header? */
    {
	FormatToString(&(hb->RHSFormat), formatstr);
	printf("RHS format (RHSFMT)                   : %s\n", formatstr);


	if (hb->RHSVectorFormat == IsFull)
	    printf("Vector Format                     : full storage\n");
	else if (hb->RHSVectorFormat == IsSame)
	    printf("Vector Format                     : same storage as matrix\n");
	else
	    printf("Vector Format                     : is erroneous\n");

	if (hb->GuessSupplied == 'G')
	    printf("A guess is supplied\n");
	else
	    printf("A guess is not supplied\n");
    
	if (hb->ExactSupplied == 'X')
	    printf("A exact solution vector is supplied\n");
	else
	    printf("A exact solution vector is not supplied\n");

	printf("Number of right hand sides (NRHS)     : %d\n",hb->NumRHS);
	if (hb->Assembly == IsAssembled)
	    printf("Number of row indices (NRHSIX)        : %d\n",hb->NumRHSRowIndex);
	else 
	    printf("Number of row indices is ignored      : %d\n",hb->NumRHSRowIndex);
    }
    printf("Maximum value of the matrix = %e, max of fabs(matrix values) = %e\n",hb->MaxValue, hb->AbsMaxValue);
    printf("Minimum value of the matrix = %e, min of fabs(matrix values) = %e\n",hb->MinValue, hb->AbsMinValue);
	   
}


    
    
    
    


