#include <hb.h>
/*
 * File: parser.c -- routines for dealing with FORTRAN format specs.
 *
 * Includes: FormatToString ParseFormat
 *
 * Started: 2/1/94
 *
 * Modified: 2/12/94 -- added/cleaned up comments
 *           3/18/94 -- added TRACE calls
 *
 */


/* valid uppercase FORTRAN format chars (for this program, anyway) */
/* LEAVE THIS HERE FOR Compilation */

/* global */
char *ValidFormatChars = "IEPFDQ";


/*
 * Function: FormatToString
 *
 * Purpose: converts an internal format struct to a string representing the
 *          corresponding FORTRAN format spec.
 *
 * Parameters:  f - a pointer to the format struct
 *              s - a char buffer for the output string
 *
 * Return Values: none
 * 
 * Calls: sprintf
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started:   about 2/5/94
 *
 * Modified:  2/12/94 -- added comments
 *
 */
void FormatToString( Format *f, char *s )
{
    TRACE("entering FormatToString\n");

    /* if need to use width.decimal_places format do so */
    if (f->decimal_places > 0)
	sprintf(s," (%d%c%d.%d)",f->count,
		f->format,f->width,f->decimal_places);
    else /* only width format */
	sprintf(s," (%d%c%d)",f->count,f->format,f->width);

    TRACE("leaving FormatToString\n");
}

/*
 * Function: ParseFormat
 *
 * Purpose: to parse FORTRAN format specs.
 *
 * Note:
 * A fortran format specification is of the form:
 * 
 * ( [n1 f1] [n2] f2 n3 )
 *
 * where n1, n2, and n3 are (potentially) floating point numbers
 *     of the form xx.yy  where xx = field width and yy = digits behind d.p.
 * and f1, f2 are format characters.
 *
 * Algorithm used: DFA 
 *
 * Parameters: inbuff - the input buffer containing the FORTRAN format data.
 *             size - the size of the input data field
 *             f - the format structure to be returned
 *
 * Calls: toupper strchr
 * 
 * Called by: ParseHBHeader
 *
 * Programmer: Tom Loos (tloos@cs.indiana.edu), LH 310, 812 855-3609
 *
 * Started: 6:50 PM   2/1/94
 *
 * Modified: 2/12/94 - clean up comments
 *
 */ 

int ParseFormat( char *inbuff, int size, Format *f)
{
    char curchar;
    FormatStates state = ST_INIT;
    int count = (-1); /* one behind beginning */
    int digs_seen = (-1);
    int digs_more;
    int formats_seen = (-1);
    int width[MAX_NUMBERS];
    int prec[MAX_NUMBERS];
    char debugmsg[100];
    int isformat;
    int curdigit;

    char char_formats[MAX_FORMATS+1];
    int i;


    TRACE("entering ParseFormat\n");
    /* init the number and format arrays */
    for (i = 0; i < MAX_NUMBERS; i++)
	width[i] = prec[i] = 0;

    for (i = 0; i <= MAX_FORMATS; i++)
	char_formats[i] = ' ';

    /* recognize the format with a DFA */
    /* while not done and not dead and still have input */
    while ( ( state != ST_DONE) && (state != ST_DEAD) && ( count < size))
    {
	/* update count */
	count++;

	/* ignore all spaces */
	if (inbuff[count] == ' ')
	{
	    while (inbuff[count] == ' ') 
	    {
		count++;
		if (count >= size)
		    break;
	    }
	}

	/* convert all chars to uppercase for easier processing */
	/* as of 5/9 this may be wasteful... */
	curchar = toupper(inbuff[count]);

	/* calculate if character is a format char */
	isformat = (strchr(ValidFormatChars, curchar) != (char *) NULL);

	/* calculate if current character is a digit, and if so, it's value */
	curdigit = ((curchar >= '0') && (curchar <= '9')) ? curchar - '0':-1;

	sprintf(debugmsg,
		"state = %d, curchar = %c, curdigit = %d, isformat = %d\n", 
		state,curchar,curdigit, isformat);
	DEBUG2(debugmsg);
	
	switch (state)
	{
	case ST_INIT:
	/*
	 * processing for this state:
	 *
	 *	if we don't see a left paren
	 *		goto dead state
	 *	else
	 *		goto left paren state
	 *	endif
	 */		

	    if (curchar != '(')
		state = ST_DEAD;
	    else
		state = ST_LPAREN;
	    break;

	case ST_LPAREN:
	/*
	 * processing for this state:
	 *
	 * if we see a digit
	 *    if we've seen too many digits
	 *        goto dead state
	 *    else
	 *        goto collect width state
	 * else if we see a valid format char
	 *    goto format state
	 * else 
	 *    goto dead state
	 */

	    if (curdigit >= 0)
	    {
		digs_seen++;
		if (digs_seen == MAX_NUMBERS)
		    state = ST_DEAD;
		else
		{
		    state = ST_WIDTH;
		    width[digs_seen] = curdigit;
		}
	    }
	    else if (isformat)
	    {
		formats_seen++;
		if (formats_seen == MAX_FORMATS)
			state = ST_DEAD;
		else
		{
			state = ST_FORMAT;
			char_formats[formats_seen] = curchar;
		}
	    }
	    else 
		state = ST_DEAD;
	    break;

	case ST_WIDTH:
	/*
	 * processing for this state:
	 *
	 * if curchar is a format char
	 *   goto format state
	 * else if curchar is a '.'
	 *   advance counter
	 *   goto precision state
	 * else if curchar is a ')'
	 *   goto done state
	 * else if curchar is a digit
	 *   add on the digit
	 * OTHERWISE
	 *   goto dead state
	 */

	    if (isformat)
	    {
		formats_seen++;
		if (formats_seen == MAX_FORMATS)
			state = ST_DEAD;
		else
		{
			state = ST_FORMAT;
			char_formats[formats_seen] = curchar;
		}
	    }
	    else if (curchar == '.')
	    {
		/* count++; advance character pointer */
		state = ST_PRECISION;
	    }
	    else if (curchar == ')')
	    {
		state = ST_DONE;
	    }
	    else if (curdigit >= 0)
	    {
		width[digs_seen] *= 10;
		width[digs_seen] += curdigit;
	    }
	    else
		state = ST_DEAD;
	    break;

	case ST_PRECISION:
	    /*
	     * Processing for this state:
	     *
	     * if curchar is a digit
	     *    update precision value
	     * else if curchar is a format character
	     *    goto format state
	     * else if curchar is the right paren
	     *    goto done state
	     * else
	     *    goto dead state
	     */
	    if (curdigit >= 0)
	    {
		prec[digs_seen] *= 10;
		prec[digs_seen] += curdigit;
	    }
	    else if (isformat)
	    {
		formats_seen++;
		if (formats_seen == MAX_FORMATS)
			state = ST_DEAD;
		else
		{
			state = ST_FORMAT;
			char_formats[formats_seen] = curchar;
		}
	    }
	    else if (curchar == ')')
	    {
		state = ST_DONE;
	    }
	    else
	    {
		state = ST_DEAD;
	    }
	    break;

	case ST_FORMAT:
	    /*
	     * processing for this state:
	     *
	     * if (curchar is a valid digit)
	     *     goto width state
	     * else // can't see anything else from here
	     *     goto dead state
	     */
	    if (curdigit >= 0)
	    {
		digs_seen++;
		if (digs_seen == MAX_NUMBERS)
		    state = ST_DEAD;
		else
		{
		    state = ST_WIDTH;
		    width[digs_seen] = curdigit;
		}
	    }
	    else
		state = ST_DEAD;
	    break;
	default:
	    sprintf(debugmsg,"DEFAULT STATE REACHED: state = %d\n",state);
	    DEBUG1(debugmsg);
	    state = ST_DEAD;
	}/* end of switch state */
    } /* end of while loop */
    /* return proper return value */

    if (state == ST_DONE)
    {
	/* convert digs/formats seen from indices to "intuitive" values */
	digs_seen++;
	formats_seen++;

#ifdef DEBUG_LEVEL
	sprintf(debugmsg,"digs_seen = %d,formats_seen = %d\n",digs_seen,formats_seen);
	DEBUG2(debugmsg);
	for (i = 0; i < digs_seen; i++)
	{
	    if (i < formats_seen)
	    {
		sprintf(debugmsg,"format[%d] = %c\n",i,char_formats[i]);
		DEBUG2(debugmsg);
	    }
	    sprintf(debugmsg,"width[%d] = %d, prec[%d] = %d\n",i,width[i],
		    i,prec[i]);
	    DEBUG2(debugmsg);
	 }
#endif

	/* figure out how many more digits we've seen */
	/* should be either 0 or 1 */
	digs_more = digs_seen - formats_seen;
	if ((digs_more == 0) && (formats_seen > 1))
	{
	    TRACE("ParseFormat: leaving BAD\n");
	    return(BAD);
	}
	
	switch (digs_seen)
	{
	case 1:
	    /* format of the form:  i5 */
	    f->count = 1;
	    f->format = char_formats[0];
	    f->width = width[0];
	    f->decimal_places = prec[0];
	    break;

	case 2:
	    /* format of the form: 4i5 */
	    f->count = width[0];
	    f->format = char_formats[0];
	    f->width = width[1];
	    f->decimal_places = prec[1];
	    break;
	case 3:
	    /* format of the form 1p4f20.13 */
	    /* ignore 1p */

	    f->count = width[1];
	    f->format = char_formats[1];
	    f->width = width[2];
	    f->decimal_places = prec[2];
	    break;
	default:
	    printf("bad format 1\n");
	    return(BAD);
	    
	} /* end of switch digs seen */
	TRACE("leaving ParseFormat OK\n");
	return(OK);
    }
    /* if not done, died or ran out of space */
    DEBUG1("ParseFormat: reached bad state!\n");
    TRACE("leaving ParseFormat BAD\n");
    return (BAD);
}    

/*
 *main()         //DEBUGGING MAIN ROUTINE
 *{
 *   char buff[1000];
 *   Format f;
 *
 *   fprintf(stderr,"enter format: ");
 *   while (fgets(buff,100,stdin) != (char *) NULL)
 *   {
 *	if (ParseFormat(buff, 30, &f) == OK)
 *	{
 *	    printf("parsed OK\n");
 *	}
 *	else
 *	{
 *	    printf("failed parse\n");
 *	}
 *	fprintf(stderr,"enter format: ");
 *   }
 *}
 */
