#include <hb.h>
/*
 * File: rcas.c
 *
 * References:
 *  Users' Guide for the Harwell-Boeing Sparse Matrix Collection (Release I),
 *     Duff, Grimes, and Lewis, publisher?, Oct. 1992
 *
 *  Matlab External Interface Guide, The Math Works, Natick, Mass, Aug. 1992
 *
 *  The O'Reilly Series of X Books
 *
 * Includes: ReadCompressAndScaleMat() InitRCASParameters() BigDrawMat()
 *           BigDrawMat2() AllocRCAS() DeallocRCAS() FreeRCAS()
 *
 * Started: 2/22/94
 *
 * Modified: 3/18/94 -- add Alloc/Dealloc/FreeRCAS
 *           3/22/94 -- call init from AllocRCAS()
 *
 */


/*
 * Function: InitRCASParameters()
 *
 * Purpose: Sets the parameters structure up with a default set of values
 *
 * Parameters: rcas - a pointer to a RCASParameters structure:
 *             those parameters are:
 *
 *             RCASAction action: what to do
 *               Any higher level action implies the lower level actions
 *               will be done after it.
 *               So, RCASread will do a read, compress, and scale (RCAS!)
 *               RCAScompress will do a compress and scale
 *               RCASscale will just do a scale
 *             file: filename of input matrix (only needed for Read)
 *             input: the input matrix (needed for Read and Compress)
 *             comp: the compression matrix 
 *             pix: the output drawn matrix as a pixmap
 *             cstrat: the compression strategy (see compress_strat.c for 
 *                     some examples
 *             sstrat: the scaling strategy (see scale_strat.c for examples)
 *             screen_height: the height/max_y/max_rows of the output screen 
 *                            in pixels
 *             screen_width: the width/max_x/max_cols of the output screen 
 *                            in pixels
 *             conformal: do we conform to screen or matrix?
 *             maxcolor: the maximum color number allowed
 *             gcs: an array of graphics contexts (GCs) -- there should be
 *                  at least (maxcolor+1) of them
 *
 * Return Values:  none
 * 
 * Calls: none
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started:   2/22/94
 *
 * Modified:  
 *
 */

void InitRCASParameters(RCASParameters *rcas)
{
    rcas->used = 0; /* I'm not used */
    rcas->action = RCASread; /* do it all, baby */

    /* handle the R,C, and A pointers */
    rcas->file = (char *) NULL;
    rcas->input = (DaMatrix *) NULL;
    rcas->comp = (DenseMatrix *) NULL;

    /* set up default strategy to be ln scaling (7/6/94) */
    rcas->cstrat = CSsum; /* sum up all entries */
    rcas->sstrat = SClog; /* scale them linearly */
    
    /* conform to matrix by default */
    rcas->conformal = CONFORMAL_TO_MATRIX;

    /* submatrix values */
    rcas->input_startcol = (-1);
    rcas->input_endcol = (-1);
    rcas->input_startrow = (-1);
    rcas->input_endrow = (-1);


    /* init maxima and minima */
    rcas->max= rcas->absmax = rcas->input_max = rcas->input_absmax = -HUGE_VAL;
    rcas->min= rcas->absmin = rcas->input_min = rcas->input_absmin = HUGE_VAL;
    rcas->zero_in_input = 0; /* no zeros in this input */
    rcas->log10=0; /* use ln scaling */

    /* set up bad values for X stuff */
    rcas->mapped = 0; /* window hasn't been mapped yet */
    rcas->bitmap = 0; /* by default, see many colors */
    rcas->pixalloc = 0; /* virtually set pixmap to null */
    rcas->numcolors = (-1);
    rcas->startcolor = (-1);
    rcas->force_black = 1; /* force scaling to black */
    rcas->colormap = SPECTRAL; /* set default colormap to the spectral map */
    rcas->what_is_black = BLACK_INDEX;
    rcas->screen_width = (-1);
    rcas->screen_height = (-1);
    rcas->window = (Window) NULL;
    rcas->display = (Display *) NULL;
    rcas->gcs = (GC *) NULL;

    /* init widgets */
    rcas->wids.toplevel = (Widget) NULL;
    rcas->wids.form = (Widget) NULL;
    rcas->wids.menu = (Widget) NULL;
    rcas->wids.row_col = (Widget) NULL;
    rcas->wids.canvas = (Widget) NULL;
    rcas->wids.msg_win = (Widget) NULL;
    rcas->wids.cur_pos = (Widget) NULL;
    rcas->wids.color_bar = (Widget) NULL;
    rcas->wids.color_min = (Widget) NULL;
    rcas->wids.color_max = (Widget) NULL;
    rcas->wids.msg_label = (Widget) NULL;
    /* ptn stuff */
    rcas->vectfile = (char *) NULL;

    rcas->NumRowPtns = 0;
    rcas->RowPtn = (int *) NULL;

    rcas->NumColPtns = 0;
    rcas->ColPtn = (int *) NULL;

    rcas->UsedColPtns = 0;
    rcas->NewColPtn = (int *) NULL;

    rcas->UsedRowPtns = 0;
    rcas->NewRowPtn = (int *) NULL;
    /* 1/14/94 -- add screen array */
    rcas->screen_array = (int *) NULL;

}
/*
 * Function: ReadCompressAndScaleMat()
 *
 * Purpose: Driver routine that calls the HB input, compression and scaling
 *          routines as needed.
 *
 * Parameters: rcas - a pointer to a RCASParameters structure:
 *             See the header comment to InitRCASParameters() for a 
 *             description of the RCASParameters structure
 *
 * Return Values:  OK if OK, BAD otherwise
 * 
 * Calls: ReadHBMat, CompressMat, ScaleMat, etc.
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started:   2/22/94
 *
 * Modified:  2/28 -- added flip_rows
 *            2/28 -- added randy_order
 *            3/18 -- changed randy_order to sp_col_major (sparse input 
 *                    is column major vs. being row major from HB) and
 *                    nuked flip_rows
 *            3/20 -- added setting up of a submat
 *            4/6  -- handle all matrix elements the same 
 *                    (i.e. vumat(ones(100)); case) in RCASscale
 *            4/14 -- call CloseAllZoomWindows() on read case
 *            1/14/95 -- add RCASscalearray support.
 *            5/20/95 -- use new ReadPtnFile (from AET IO)
 *
 */

int ReadCompressAndScaleMat( RCASParameters *rcas )
{
    CompressionRequest compreq;
    ScaleRequest scalereq;
    int rc;
    int use_c;
    int temp, i;
    char debugstr[100];
    struct stat statbuf;
    Cursor watch_cursor;
    int display_times = DISPLAY_TIMING;
    double t1, t2;
#ifdef HANDLE_PTNS
    int *NewRowPtn;
    int *NewColPtn;
    int screen_height, screen_width;
    int new_screen_height, new_screen_width;
    int sel_cols, sel_rows;
    int maxptn, minptn, absmaxptn, absminptn;
    int useful_row_ptns, useful_col_ptns;
#endif

    sprintf(debugstr,"RCAS entered: rcas action = %d\n",rcas->action);
    TRACE(debugstr);
/*  fprintf(stderr,"%s",debugstr);  */
    /* ALWAYS USE C INDEXING IF I CAN GET AWAY WITH IT */
    use_c = 1;

    /* set the cursor to the watch cursor for this window */
    SetCursor(XC_watch, &watch_cursor);
    /* make sure we can see the watch cursor */
    XFlush(RCAS[0].display);
    /* basic idea -- switch on action, break only on failure or at bottom */
    switch (rcas->action)
    {
    case RCASread: /*  read, compress and scale  */
	/* check out the zoom file */
t1 = mytime();
	sprintf(debugstr,"RCAS: reading %s\n",rcas->file);
	DEBUG1(debugstr);
/* fprintf(stderr,"%s",debugstr);   */
	
	/* 9/8/94 make emily the back end of a pipe */
	/* pass in one argument: a "-" , to read in stdin */
	/* and if we can't find the file, call it a day */

	if ((strcmp(rcas->file,STDIN_ARG) != 0) &&
	    (stat(rcas->file, &statbuf ) == (-1)))
	{
	    rc = BAD;
	    sprintf(debugstr,"Error in accessing file %s",rcas->file);
	    XErrorMsg(debugstr);
	    break;
	}
	/* now we know we can find the file, so clean up old windows */
	CloseAllZoomWindows();
	/* and clear the canvas */
	CanvasClearWindow(rcas-RCAS);
	/* assume input is an HB format file */
	rcas->input->type = HB;
	if ((rc = ReadHBFile( rcas->file, rcas->input->data.hb, 1)) != OK)
 	{
	    sprintf(debugstr,"RCAS: ReadHBFile failed\n");
	    DEBUG1(debugstr);
	    break;
	}
	sprintf(debugstr,"RCAS: OK reading %s\n",rcas->file);
	DEBUG1(debugstr);

#ifdef HANDLE_PTNS
	/* if we have a partition file */
	if (rcas->vectfile != (char *) NULL)
	{
	    int ptnfd; 
/*  fprintf(stderr,"we have a vector file = %s\n",rcas->vectfile); */
	    /* try to open the file */
	    if ((ptnfd = open(rcas->vectfile, O_RDONLY)) < 0)
	    {
		sprintf(debugstr,"RCAS: failed to open partition file %s.\n",
			rcas->vectfile);
		DEBUG1(debugstr);
		fprintf(stderr,"%s Continuing without partitions.\n",rcas->vectfile);
		rcas->NumRowPtns = rcas->NumColPtns = 0;
	    }
	    /* 5/20/95 new ReadPtnFile uses file name, not fd, so close 
	       file */
	    close(ptnfd);
	    /* if file opens, try to read in the ptn vector */
	    if ((rc = ReadPtnFile( rcas->vectfile, rcas->RowPtn, 
					  &(rcas->NumRowPtns))) != OK)
	    {
		sprintf(debugstr,"RCAS: error reading partition file %s.\n",
			rcas->vectfile);
		DEBUG1(debugstr);
		fprintf(stderr,"%s Continuing without partitions.\n",rcas->vectfile);
		rcas->NumRowPtns = rcas->NumColPtns = 0;

	    }
	    /* if file read ok, copy vector (for now == 9/22/94) */
	    else /* for now, just one ptn vector.  Copy row ptn to cols */
	    {
		rcas->NumColPtns = rcas->NumRowPtns;
		for (i = 0; i <= rcas->NumColPtns; i++)
		{
		    rcas->ColPtn[i] = rcas->RowPtn[i];
		}
	    }
	} /* end of if we have a ptn file */

	sprintf(debugstr,"RCAS: OK reading %s\n",rcas->vectfile);
	DEBUG1(debugstr);
#endif 

	/* show whole new matrix */
	rcas->input_startcol = 0;
	rcas->input_startrow = 0;
	/* subtract 1 since doing (implicit) C addressing */
	rcas->input_endrow = rcas->input->hb_NumRows-1; /* 4/13/94 for -1 */
	rcas->input_endcol = rcas->input->hb_NumCols-1; /* 4/13/94 for -1 */

	/* fall through expected if AOK */
t2 = mytime();
if (display_times)
    printf("\nRead time = %e sec.\n",t2-t1);
    
    case RCAScompress:  /* compress and scale */
t1 = mytime();
	sprintf(debugstr,"RCAS: compressing\n");
	DEBUG1(debugstr);
/*  fprintf(stderr,"%s",debugstr);   */

	if (rcas->input->type == HB)
	   /*  compreq.use_c = (rcas->input->hb_row_inds[0] == 0); */
	    compreq.use_c = (rcas->input->hb_col_inds[0] == 0);
	else
	    compreq.use_c = 1;
	compreq.sp_col_major = 0; /* flip rows and cols  */
	compreq.strat = rcas->cstrat;
	compreq.conformal = rcas->conformal;
	/* reorder submatrix rectangle, if necessary */
	if (rcas->input_startrow > rcas->input_endrow)
	{
	    temp = rcas->input_startrow;
	    rcas->input_startrow = rcas->input_endrow;
	    rcas->input_endrow = temp;
	}
	if (rcas->input_startcol > rcas->input_endcol)
	{
	    temp = rcas->input_startcol;
	    rcas->input_startcol = rcas->input_endcol;
	    rcas->input_endcol = temp;
	}
	compreq.input_startrow = rcas->input_startrow;
	compreq.input_endrow = rcas->input_endrow;
	compreq.input_startcol = rcas->input_startcol;
	compreq.input_endcol = rcas->input_endcol;
#ifdef HANDLE_PTNS
	
	/*
	 * get local copies of data for potential changes without 
	 *  mucking up the "official" copy in RCAS
	 */
	rcas->UsedRowPtns = rcas->NumRowPtns;
	rcas->UsedColPtns = rcas->NumColPtns;
	NewRowPtn = rcas->NewRowPtn;
	NewColPtn = rcas->NewColPtn;
	screen_height = rcas->screen_height;
	screen_width = rcas->screen_width;

	/* calculate the number of selected input rows/cols */
	sel_cols = IDXRANGE(rcas->input_startcol, rcas->input_endcol);
	sel_rows = IDXRANGE(rcas->input_startrow, rcas->input_endrow);

	/* init max_x and y */
	compreq.max_x = rcas->screen_width;
	compreq.max_y = rcas->screen_height;

	/* figure out actual screen size */
	CalculateScreenSize(sel_rows,sel_cols,&screen_height, &screen_width,
			    rcas->conformal);

	/* handle row partitions if need be (more than 2 globally is need) */
/*
 * fprintf(stderr,"Pre CSP: UsedRowPtns = %d, rcas height = %d, screen_height = %d, max_y = %d\n",rcas->UsedRowPtns, rcas->screen_height,screen_height,compreq.max_y);
 */
	if (rcas->UsedRowPtns > 1)
	{
	    /* figure out how many row ptns we're REALLY using */
	    useful_row_ptns = CalcScreenPtn(rcas->RowPtn,NewRowPtn, 
			      &(rcas->UsedRowPtns),rcas->input_startrow, 
			        rcas->input_endrow, screen_height);
	    /* if still using 2 or more, update max_y */
	    /* if (rcas->UsedRowPtns > 1) */
	    /* 	compreq.max_y = rcas->screen_height - rcas->UsedRowPtns + 1; */
	    if (useful_row_ptns > 1) 
	    {
	         compreq.max_y = rcas->screen_height - useful_row_ptns + 1; 
	    }
	}
/*
 * fprintf(stderr,"Post CSP:UsedRowPtns = %d, rcas height = %d, screen_height = %d, max_y = %d\n",rcas->UsedRowPtns, rcas->screen_height,screen_height,compreq.max_y);
 * fprintf(stderr,"UsedColPtns = %d, rcas width = %d, screen_width = %d, max_x = %d\n",rcas->UsedColPtns, rcas->screen_width,screen_width,compreq.max_x);
 */

	/* handle col partitions if need be (more than 2 globally is need) */
	if (rcas->UsedColPtns > 1)
	{
	    /* figure out how many col ptns we're REALLY using */
	    useful_col_ptns = CalcScreenPtn(rcas->ColPtn,NewColPtn, 
				  &(rcas->UsedColPtns), rcas->input_startcol, 
				    rcas->input_endcol, screen_width);
	    /* if still using 2 or more, update max_x */
	    /* if (rcas->UsedColPtns > 1)
	     *   compreq.max_x = rcas->screen_width - rcas->UsedColPtns + 1;
	     */
	    if (useful_col_ptns > 1)
	    {
	      compreq.max_x = rcas->screen_width - useful_col_ptns + 1;
	    }
	}
/*
 * fprintf(stderr,"after CSP: UsedColPtns = %d, rcas width = %d, screen_width = %d, max_x = %d, useful cols = %d\n",rcas->UsedColPtns, rcas->screen_width,screen_width,compreq.max_x,useful_col_ptns);
 */

	/*
	 * now that we've figured out what max_x and max_y should be, may
	 * have to recalculate the ptn vectors, since the newly calculated
	 * screen size could be different (leading to nasty off-by-1 errors)
	 * 
	 * do that here.
	 */
	if ((compreq.max_x != rcas->screen_width) || 
	    (compreq.max_y != rcas->screen_height))
	{
/*    fprintf(stderr,"max has changed, recalculating screen height and width\n");  */

	    /* set up for call to CalculateScreenSize */
	    new_screen_width = compreq.max_x;
	    new_screen_height = compreq.max_y;
	    CalculateScreenSize(sel_rows,sel_cols,&new_screen_height, 
				&new_screen_width, rcas->conformal);
	    /* see if the actual screen_width changed */
	    /*if ((rcas->UsedColPtns > 1) && 
	     * ((new_screen_width+rcas->UsedColPtns - 1) != screen_width))
	     */
	    if ((useful_col_ptns > 1) && 
	       ((new_screen_width+useful_col_ptns - 1) != screen_width))
	    {
/*
 *		fprintf(stderr,"Screen width changed from %d to %d\n",
 *		screen_width, new_screen_width);
 */
		/* figure out new partitioning */
	      useful_col_ptns = CalcScreenPtn(rcas->ColPtn,NewColPtn, &(rcas->UsedColPtns),
					      rcas->input_startcol, rcas->input_endcol, 
					      new_screen_width+useful_col_ptns-1);
	    }

	    /* see if the screen height changed */
	    if ((useful_row_ptns > 1) && 
		((new_screen_height+useful_row_ptns - 1) != screen_height))
	    {
	/*  
	 * fprintf(stderr,"Screen height changed from %d to %d\n",
	 *	screen_height, new_screen_height);
	 */

		/* figure out new partitioning */
		useful_row_ptns = CalcScreenPtn(rcas->RowPtn,NewRowPtn, &(rcas->UsedRowPtns),
						rcas->input_startrow, rcas->input_endrow, 
						new_screen_height+useful_row_ptns-1);
	    }
	} /* end of check for new screen sizes */

	/* fprintf(stderr,"calling CompressMat with max_x = %d, max_y = %d\n",
	 *	compreq.max_x, compreq.max_y);
	 */
#else
	compreq.max_x = rcas->screen_width;
	compreq.max_y = rcas->screen_height;
#endif

	if ((rc = CompressMat(rcas->input, &(rcas->comp), &compreq)) != OK)
	{
	    sprintf(debugstr,"RCAS: CompressMat failed\n");
	    DEBUG1(debugstr);
	    break;
	}
	sprintf(debugstr,"RCAS: done compressing\n"); 
/* fprintf(stderr,"%s",debugstr); */
	DEBUG1(debugstr);
	/* fill out rcas struct for max, min, etc. */
	rcas->max = compreq.mat_max;
	rcas->min = compreq.mat_min;
	rcas->absmax = compreq.mat_absmax;
	rcas->absmin = compreq.mat_absmin;
	rcas->zero_in_input = compreq.zero_in_input;

	rcas->input_min = compreq.input_min;
	rcas->input_max = compreq.input_max;

	rcas->input_absmin = compreq.input_absmin;
	rcas->input_absmax = compreq.input_absmax;
#ifdef HANDLE_PTNS
	if (rcas->UsedRowPtns > 1)
	{
	    PARTITION(compreq.input_max_row, NewRowPtn, rcas->UsedRowPtns, 
		      maxptn);
	    PARTITION(compreq.input_min_row, NewRowPtn, rcas->UsedRowPtns, 
		      minptn);
	    PARTITION(compreq.input_absmax_row,NewRowPtn, 
		      rcas->UsedRowPtns,absmaxptn);
	    PARTITION(compreq.input_absmin_row,NewRowPtn, 
		      rcas->UsedRowPtns,absminptn);
	}
	else
	{
	    maxptn = minptn = absminptn = absmaxptn = 0;
	}
/*
 * fprintf(stderr,
 *	"For rows:minptn = %d,maxptn = %d,absminptn = %d,absmaxptn = %d\n",
 *	 minptn,maxptn,absminptn,absmaxptn);
 */

	rcas->input_min_row = compreq.input_min_row + minptn;
	rcas->input_max_row = compreq.input_max_row + maxptn;

	rcas->input_absmax_row = compreq.input_absmax_row + absmaxptn ;
	rcas->input_absmin_row = compreq.input_absmin_row + absminptn ;

/*
 *  fprintf(stderr,"For rows: min = %d, max = %d, absmin = %d, absmax = %d\n",
 *	rcas->input_min_row,
 *	rcas->input_max_row,
 *	rcas->input_absmin_row,
 *	rcas->input_absmax_row);
 */

	if (rcas->UsedColPtns > 1)
	{
	    PARTITION(compreq.input_min_col, NewColPtn, rcas->UsedColPtns, 
		      minptn);
	    PARTITION(compreq.input_max_col, NewColPtn, rcas->UsedColPtns, 
		      maxptn);
	    PARTITION(compreq.input_absmin_col,NewColPtn, 
		      rcas->UsedColPtns,absminptn);
	    PARTITION(compreq.input_absmax_col,NewColPtn, 
		      rcas->UsedColPtns,absmaxptn);
	}
	else
	{
	    minptn = maxptn  = absminptn = absmaxptn = 0;
	}

/*
 * fprintf(stderr,
 *	"For cols:minptn = %d,maxptn = %d,absminptn = %d,absmaxptn = %d\n",
 *	minptn,maxptn,absminptn,absmaxptn);
 */

	rcas->input_min_col = compreq.input_min_col + minptn;
	rcas->input_max_col = compreq.input_max_col + maxptn;

	rcas->input_absmin_col = compreq.input_absmin_col + absminptn;
	rcas->input_absmax_col = compreq.input_absmax_col + absmaxptn;
/*
 *fprintf(stderr,"For cols: min = %d, max = %d, absmin = %d, absmax = %d\n",
 *	rcas->input_min_col,
 *	rcas->input_max_col,
 *	rcas->input_absmin_col,
 *	rcas->input_absmax_col);
 */

#else
	rcas->input_min_row = compreq.input_min_row;
	rcas->input_max_row = compreq.input_max_row;

	rcas->input_min_col = compreq.input_min_col;
	rcas->input_max_col = compreq.input_max_col;

	rcas->input_absmin_row = compreq.input_absmin_row;
	rcas->input_absmax_row = compreq.input_absmax_row;

	rcas->input_absmin_col = compreq.input_absmin_col;
	rcas->input_absmax_col = compreq.input_absmax_col;

#endif

	rcas->bitmap = compreq.bitmap;
t2 = mytime();
if (display_times)
    printf("Compression time = %e sec.\n",t2-t1);

	/* fall through expected */

    case RCASscale:
    case RCASscalearray: /* both treated here */
t1 = mytime();
	sprintf(debugstr,"RCAS: scaling\n");
	DEBUG1(debugstr);
/* fprintf(stderr,"Scaling\n");  */

	scalereq.bitmap = rcas->bitmap;
	scalereq.min = rcas->min;
	scalereq.max = rcas->max;
	
	/* if all the numbers are the same, just paint the background black */
	/* 4/6/94 */
	/*
	 * if (myabs(scalereq.max - scalereq.min) < EPSILON)
	 * {
	 *
	 * 4/19/94 -- use CanvasClearWindow
	 *   CanvasClearWindow(rcas-RCAS); 
	 *
	 * XFillRectangle(rcas->display,
	 *	   rcas->pix, 
	 *		   rcas->gcs[rcas->what_is_black],
	 *		   0,0,
	 *		   rcas->screen_width,
	 *		   rcas->screen_height);
	 *
	 *    break;
	 *}
	 */
	scalereq.absmin = rcas->absmin;
	scalereq.absmax = rcas->absmax;
	scalereq.zero_in_input = rcas->zero_in_input; 
	scalereq.log10 = rcas->log10;
	scalereq.gcs = rcas->gcs;
	scalereq.force_black = rcas->force_black;
	scalereq.what_is_black = rcas->what_is_black;
	scalereq.startcolor = rcas->startcolor;
	scalereq.numcolors = rcas->numcolors;
	scalereq.display = rcas->display;
	scalereq.strat = rcas->sstrat;
/* #ifdef HANDLE_PTNS */
	scalereq.num_row_ptns = rcas->UsedRowPtns;
	scalereq.num_col_ptns = rcas->UsedColPtns;
	scalereq.row_ptn = rcas->NewRowPtn;
	scalereq.col_ptn = rcas->NewColPtn;
	scalereq.start_row = rcas->input_startrow;
	scalereq.end_row = rcas->input_endrow;
	scalereq.start_col = rcas->input_startcol;
	scalereq.end_col = rcas->input_endcol;
/* #endif */
	if (rcas->action == RCASscalearray)
	{
	    if ((rc = ScaleArray(rcas->comp, rcas->screen_array, 
				 &scalereq, rcas->colormap)) != OK)
	    {
		sprintf(debugstr,"RCAS: ScaleArray failed\n");
		DEBUG1(debugstr);
	    }
	}
	else
	{
	    if ((rc = ScaleMat(rcas->comp, rcas->pix, &scalereq)) != OK)
	    {
		sprintf(debugstr,"RCAS: ScaleMat failed\n");
		DEBUG1(debugstr);
	    }
	}
	sprintf(debugstr,"RCAS: done scaling\n");
	DEBUG1(debugstr);
t2 = mytime();
if (display_times)
    printf("Scaling time = %e sec.\n",t2-t1);
	break;

    case RCASexpose:
	sprintf(debugstr, "RCAS called with RCASexpose! BUG!!!\n");
	DEBUG1(debugstr);
	break;
    default:
	error("ReadCompressAndScaleMat: invalid action",ARGERR);
	/* shouldn't get here */
	return(BAD);
    }

    /* reset cursor */
    /* XtVaSetValues(motif_screen,
     *		  XmNmenuCursor, old_cursor,
     *		  NULL);
     */
    ResetCursor(watch_cursor);
    sprintf(debugstr,"RCAS: rc = %d\n",rc);
    TRACE(debugstr);
    return(rc);
}
void
BigDrawMat2(char *s, int code, DaMatrix *input_mat, int nptns, int *ptn)
{
    char debugstr[100];
    static char *argv[3];
    int argc,i;

    sprintf(debugstr,"in BDM2: s = %s\n",s);
    TRACE(debugstr);

    argc = 2;

    argv[0] = MALLOC(1,sizeof(DEFAULT_NAME)+1);
    strcpy(argv[0],DEFAULT_NAME);
    argv[1] = s;

    /* yeah, I know, how strange, but this is totally CYA */
    argv[2] = MALLOC(1,1);
    argv[2][0] = (char) NULL;

    BigDrawMat(argc,argv,code,input_mat, nptns, ptn);
}
/*
 * Function: BigDrawMat()
 *
 * Parameters: argc, argv: just like for main
 *
 * Purpose: provide a basic framework for spawning off separate windows
 *          (I suppose I could fire up all the kids at once, or some
 *           variations on that theme or priority sort them or ...)
 * 
 * Note: This is a rough looking piece of code, but it's the right general 
 *       idea.
 *
 * Arguments: argc == argument count
 *            argv == 2D character array of arguments in the standard form
 *            code == flag that's set to DRAW_MAT_USE_FILENAME (if doing I/O)
 *                    or DRAW_MAT_USE_MATRIX (if already read in)
 *            input_mat == input matrix
 *            nptns == number of partitions
 *            ptn   == partition data
 * 
 * Algorithm:  for every argument to be processed,
 *                 fork off a kiddo
 *                 process that argument
 *                 quit when necessary 
 *
 * Modified: 5/1/95 -- only handle two kinds of cmd line inputs 
 *                     emily.foo hb_file
 *                     emily.foo hb_file ptn_vector
 *                     (i.e. lose support for multiple hb file cmd lines)
 *           1/13/96 -- rationalize fork code
 *           6/18/96 -- pass in ptn data
 */
int BigDrawMat(int argc, char *argv[], int code, DaMatrix *input_mat,
	       int nptns, int *ptn)
{
    char debugstr[100];
    int pid,i,flag,itercount, myargc;
    int main_pid,return_prompt = RETURN_PROMPT;
    int fork_kids = 1;  /* ALWAYS FORK OFF A KID! */
    sprintf(debugstr,"BigDrawMat: argv = %x, argv[0] = %x\n",argv,argv[0]);
    TRACE(debugstr);

    myargc = argc;
    /* 12/1/95 -- handle fork_kids */
    if (!fork_kids)
    {
	/* printf("BDM -- calling DrawMat\n"); */
	DrawMat(myargc,argv,code,input_mat, nptns, ptn);
	printf("BDM -- return from DrawMat\n"); 
	return(0);
    }

    /* fork off a controlling process */
    if ((main_pid = fork()) == 0)
    {
	/* I am the kid -- do the real work */
	myargc = argc;
	sprintf(debugstr,"BigDrawMat: argv = %x, argv[0] = %x\n",argv,argv[0]);
	DEBUG1(debugstr);
	DrawMat(myargc,argv,code,input_mat, nptns, ptn);
	/* shouldn't get here ... */
	return(0);
    }
    else /* I am the parent -- */
    {
	if (return_prompt) /* if returning, abandon my child to the wilds of the OS */
	{
	    return(0);
	}
	else /* wait for the kiddo */
	{
	    wait(NULL); 
	}
    }
    return(0); /* parent should get here */
}
/*
 * Function: AllocRCAS
 *
 * Purpose: to get a free RCAS out of an array of max RCASParameters
 *
 * Parameters: r - an array of RCASParameters structures
 *             max - the number of struct's in the array
 *
 * Return Values: -1 no free entries in the table
 *                any pos. number - the RCAS entry to use
 * 
 * Calls: InitRCASParameters()
 *
 * Notes: 
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started:  3/18/94
 *
 * Modified:  3/22/94 -- call InitRCASParameters()
 *
 */

int AllocRCAS( RCASParameters r[], int max )
{
    int i;

    for (i = 0; i < max; i++)
    {
	if (r[i].used == RCAS_FREE)
	{
	    if (r[i].pixalloc)
		XFreePixmap(canvas_display,r[i].pix);
	    InitRCASParameters( &(r[i]));
	    r[i].used = RCAS_USED;
	    return(i);
	}
    }
    return(-1);
}

void DeallocRCAS( RCASParameters r[], int member, RCASNuke nuke )
{
    r[member].used = RCAS_FREE;
    r[member].mapped = 0; /* 6/2/94 */
    XtUnrealizeWidget( r[member].wids.toplevel); /* 4/10/94 */
    FreeDenseMatrix( r[member].comp );
    if ( r[member].pixalloc )
    {
	XFreePixmap( r[member].display, r[member].pix );
	r[member].pixalloc = 0;
    }
    
    if (nuke == RCAS_NUKE_IT)
    {
	FreeRCAS( &(r[member]) );
    }
}

void FreeRCAS( RCASParameters *r )
{
    TRACE("FreeRCAS entered\n");
    FreeDaMatrix( r->input );
    TRACE("FreeRCAS exited\n");
}
