#ifndef HANDLE_PTNS
#include <hb.h>

/* 5/31/94 -- FOR DEBUGGING ONLY */
/* #define FILE_DUMP  */

/*
 * File: scale.c
 *
 * References:
 *  Users' Guide for the Harwell-Boeing Sparse Matrix Collection (Release I),
 *     Duff, Grimes, and Lewis, publisher?, Oct. 1992
 *
 *  Matlab External Interface Guide, The Math Works, Natick, Mass, Aug. 1992
 * 
 *  Volume I: Xlib Programming Manual, Adrian Nye, O'Reilly and Associates,
 *    7th printing, 12/91
 *
 * Includes: ScaleMat
 *
 * Started: 2/17/94
 *
 * Modified: 2/26/94 -- add startcolor
 *           2/28/94 -- did force_black right
 *           3/19/94 -- change from XDrawPoint to XDrawLine calls
 *           ...
 *           7/5/94 -- add support for log10
 *
 */
int ScaleMat( DenseMatrix *comp, Pixmap p, ScaleRequest *s )
{
    int scanline, col, idx, curcolor;
    int color, ncols, nrows, numcolors, startcolor, black, startx;
    char debugstr[100];
    ScaleStrategyParameters parms;

#ifdef FILE_DUMP    
    FILE *debugfp;

    if ((debugfp = fopen("scale_dump","w")) != NULL)
	fprintf(debugfp,"Scale dump started: min, max = %f to %f: absmin, absmax = %f to %f, numcolors = %d\n",s->min, s->max,s->absmin,s->absmax,s->numcolors);
#endif

    TRACE("ScaleMat: entered!\n");
    /* check input args -- need check for pixmap */
    if (!(comp->one_d) || !(comp->row_major))
    {
	error("ScaleMat: comp is not a 1D row major matrix -- BUG!!",ARGERR);
	return(BAD);
    }

    /* fill out parms */
    parms.min = s->min;
    parms.max = s->max;
    parms.absmin = s->absmin;
    parms.absmax = s->absmax;
    /* the == is right -- only if it's 0 */
    if (s->absmin == 0.0)
	parms.lmin = -MAX_SCALING_EXPONENT*log(10); /* say it's 10**-300 */
    else
	parms.lmin = log(s->absmin);

    if (s->absmax == 0.0)
	parms.lmax = -MAX_SCALING_EXPONENT*log(10); /* say it's 10**-300 */
    else
	parms.lmax = log(s->absmax);

    /* if using log10 scaling */
    if (s->log10 != 0)
    {
	/* set the parms to log10 scaling */
	parms.log10 = 1;
	/* and downsize the min/max log values */
	parms.lmin = parms.lmin * ONE_OVER_LN_10;
	parms.lmax = parms.lmax * ONE_OVER_LN_10;
    }
    else
	parms.log10 = 0;
/*
 *fprintf(stderr,"lmin = %f, lmax = %f, log10 = %d\n",parms.lmin,parms.lmax,
 *	parms.log10);
 */
    /* reset to true absmin for linear absolute scaling if zero seen in input*/
    if (s->zero_in_input)
	parms.absmin = 0.0; 
    parms.bitmap = s->bitmap;
    parms.numcolors = numcolors = s->numcolors;
    
    ncols = comp->NumCols;
    nrows = comp->NumRows;
    startcolor = s->startcolor;
    black = s->what_is_black;

    idx  = 0;
    sprintf(debugstr,"ScaleMat: before main loop! start, num = %d, %d\n",startcolor,numcolors);
    DEBUG2(debugstr);
    sprintf(debugstr,"nrows = %d, ncols = %d, bitmap = %d\n",nrows,ncols,s->bitmap);
    /* printf(debugstr) */ /* UNCOMMENT FOR BITMAP */
    DEBUG2(debugstr);
    for (scanline = 0; scanline < nrows; scanline++)
    {
#ifdef FILE_DUMP
fprintf(debugfp,"\nOn Scanline %d:\n%4d: ",scanline,0);
#endif
	/* init current line to start at left and unknown color */
	startx = 0; curcolor = -1;
	for (col = 0; col < ncols; col++)
	{
	    parms.val = comp->data.d1[ idx ];
	    
	    /* handle zero --- see hb.h for definition of SCALE_EPSILON */
	    if (fabs(parms.val) < SCALE_EPSILON)  
	    {
		if (s->force_black)
		    color = black - startcolor;
		else
		    color = 0; /* hmm...*/
	    }
	    /* call strategy and clamp colors */
	    else 
	    {
		color = s->strat(&parms);
		if (color < 0)
		{
		    fprintf(stderr,"BUG: adjusting color from %d to 0\n",color);
		    fprintf(stderr,"min = %e, max = %e, val = %e, numcolors = %d\n",
			   parms.min, parms.max, parms.val,parms.numcolors);
		    color = 0;
		}
		else if (color > numcolors)
		{
		    fprintf(stderr,"BUG: adjusting color from %d to %d\n",color, numcolors);
		    fprintf(stderr,"min = %e, max = %e, val = %e\n",
			   parms.min, parms.max, parms.val);
		    color = numcolors;
		}
	    }
#ifdef FILE_DUMP
fprintf(debugfp,"%9.1e %3d ",parms.val, color);
if ((col % 8) == 0 && (col != 0))
    fprintf(debugfp,"\n%4d: ",col+1);
#endif	
	    /* draw it */
	    /* XDrawPoint( s->display, p, s->gcs[color+startcolor], 
	     *		 col, scanline ); 
	     */
	    if (curcolor == (-1))
	    {
		curcolor = color+startcolor;
	    }
	    else if (curcolor != color+startcolor)
	    {
		XDrawLine(s->display, p, s->gcs[curcolor], startx, scanline,
			  col-1,scanline);
		startx = col;
		curcolor = color+startcolor;
	    }
	    /* update index */
	    idx++;
	} /* end of col loop */

	/* finish off line */
	XDrawLine(s->display, p, s->gcs[curcolor], startx, scanline,
		  col-1,scanline);

    } /* end of scanline loop */
    TRACE("ScaleMat: exit!\n");
#ifdef FILE_DUMP
fprintf(debugfp,"\n");
fclose(debugfp);
#endif
    return(OK);
} /* end of scale mat */
	    
#endif 
