/* #define FILE_DUMP */

#include <hb.h>
/*
 * This handles calculation of a screen partition from an input partition
 *
 * An input partition is based on the size of the input matrix.  A screen
 * partition is based on the number of cols/rows in the screen (the 
 * screen_dimension parameter) and the window of the input we're interested
 * in.  Num_ptns is initially the number of input partitions, but will be 
 * updated to the number of actual partitions in viewing window screen
 *
 * Assumptions: THE INPUT PARTITION ARRAY IS SORTED!!!
 *
 * Return Value: the actual number of partitions visibile on the screen.
 *   new_ptn[] holds the screen indices of the partitions and num_ptns
 *   will be updated.
 *
 * Started: 9/22/94 by tjl
 *
 * Modified:
 */

/*
 * File: scale_array.c -- handles creating an array of the form:
 *   r1 g1 b1
 *   r2 g2 b2
 *   ...
 * 
 * where r_i,g_i,b_i = the red, green, and blue components, respectively, of 
 *                     the i^{th} pixel.
 * 
 * This is done so that the array can be printed, etc.
 *
 * References:
 *  Users' Guide for the Harwell-Boeing Sparse Matrix Collection (Release I),
 *     Duff, Grimes, and Lewis, publisher?, Oct. 1992
 *
 *  Matlab External Interface Guide, The Math Works, Natick, Mass, Aug. 1992
 * 
 *  Volume I: Xlib Programming Manual, Adrian Nye, O'Reilly and Associates,
 *    7th printing, 12/91
 *
 * Includes: ScaleArray 
 *
 * Started: 1/14/94 -- from scale2.c to output to an array vs. a pixmap
 *
 * Modified: 
 *
 */
int ScaleArray( DenseMatrix *comp, int *array, ScaleRequest *s, int colormap)
{
    int scanline, col, idx;
    int color, ncols, nrows, numcolors, startcolor, black, startx;
    int overwrite_row, overwrite_col;
    int num_row_ptns; /* number of row partitions */
    int num_col_ptns; /* number of column partitions */
    int *NewRowPtn; /* the row partition vector */
    int *NewColPtn; /* the col partition vector */
    int start_row; /* starting row of input */
    int end_row; /* ending row of input */
    int start_col; /* starting col of input */
    int end_col; /* ending col of input */

    char debugstr[100];
    ScaleStrategyParameters parms;
    int CurColPtn, CurRowPtn;
    double FudgeFactor = 0.1;
    /* 1/14/94 -- RGB array for GetColors output */
    RGB colors[MAXIMUM_COLORS];
    int out_idx = 0;

#ifdef FILE_DUMP    
    FILE *debugfp;

    if ((debugfp = fopen("scale_dump","w")) != NULL)
	fprintf(debugfp,"Scale dump started: min, max = %f to %f: absmin, absmax = %f to %f, numcolors = %d\n",s->min, s->max,s->absmin,s->absmax,s->numcolors);
#endif

    /* pull the data from the scalereq */
    num_row_ptns = s->num_row_ptns;
    num_col_ptns = s->num_col_ptns;
    NewRowPtn = s->row_ptn;
    NewColPtn = s->col_ptn;
    start_row = s->start_row; 
    end_row = s->end_row; 
    start_col = s->start_col;
    end_col = s->end_col; 
/* 
 *  fprintf(stderr,"IN scale2.c, num_row_ptns = %d, num_col_ptns = %d\n",
 *	    num_row_ptns, num_col_ptns);
 */
    TRACE("ScaleArray: entered!\n");
    
    /* check input args -- need check for pixmap */
    if (!(comp->one_d) || !(comp->row_major))
    {
	error("ScaleMat: comp is not a 1D row major matrix -- BUG!!",ARGERR);
	return(BAD);
    }

    /* fill out parms */
    parms.min = s->min;
    parms.max = s->max;
    parms.absmin = s->absmin;
    parms.absmax = s->absmax;
    /* up the min/max values if needed for overwrite vals */
    if ((num_row_ptns > 1) || (num_col_ptns > 1))
    {
	parms.max = parms.max + FudgeFactor* (parms.max - parms.min);
	parms.absmax = parms.absmax + FudgeFactor* (parms.absmax - parms.absmin);
    }
    /* the == is right -- only if it's 0 */
    if (parms.absmin == 0.0)
	parms.lmin = -MAX_SCALING_EXPONENT*log(10); /* say it's 10**-300 */
    else
	parms.lmin = log(parms.absmin);

    if (parms.absmax == 0.0)
	parms.lmax = -MAX_SCALING_EXPONENT*log(10); /* say it's 10**-300 */
    else
	parms.lmax = log(parms.absmax);


    /* if using log10 scaling */
    if (s->log10 != 0)
    {
	/* set the parms to log10 scaling */
	parms.log10 = 1;
	/* and downsize the min/max log values */
	parms.lmin = parms.lmin * ONE_OVER_LN_10;
	parms.lmax = parms.lmax * ONE_OVER_LN_10;
    }
    else
	parms.log10 = 0;
/*
 *fprintf(stderr,"lmin = %f, lmax = %f, log10 = %d\n",parms.lmin,parms.lmax,
 *	parms.log10);
 */
    /* reset to true absmin for linear absolute scaling if zero seen in input*/
    if (s->zero_in_input)
	parms.absmin = 0.0; 
    parms.bitmap = s->bitmap;
    parms.numcolors = numcolors = s->numcolors;
    
    if (num_col_ptns > 1)
	ncols = comp->NumCols + num_col_ptns - 1 ;
    else
	ncols = comp->NumCols;
	
    if (num_row_ptns > 1)
	nrows = comp->NumRows + num_row_ptns - 1;
    else
	nrows = comp->NumRows;

/*
 *   fprintf(stderr,"Scale2: compNumRows = %d, compNumCols = %d,nrows = %d, ncols = %d\n",comp->NumRows, comp->NumCols,nrows,ncols);
 */
    startcolor = s->startcolor;
    black = s->what_is_black;

    idx = 0;
    sprintf(debugstr,"ScaleMat: before main loop! start, num = %d, %d\n",startcolor,numcolors);
    DEBUG2(debugstr);
    sprintf(debugstr,"nrows = %d, ncols = %d, bitmap = %d\n",nrows,ncols,s->bitmap);
    DEBUG2(debugstr);

    /* see if we need to do any partitioning */
/*
 * fprintf(stderr,"NCP = %d, NRP = %d\n",num_row_ptns, num_col_ptns);
 * fprintf(stderr,"start row = %d, end row = %d... start col = %d, end col = %d\n",start_row, end_row,start_col,end_col);
 */

    /* 1/14/95 -- call GetColors */
    GetColors(colormap, colors);

    /* 9/22 - duplicate code for speed */
    if ((num_col_ptns <= 1) && (num_row_ptns <= 1))
    {
/* fprintf(stderr,"No Partitions Please!\n"); */
	for (scanline = 0; scanline < nrows; scanline++)
	{
#ifdef FILE_DUMP
	    fprintf(debugfp,"\nOn Scanline %d:\n%4d: ",scanline,0);
#endif
/* init current line to start at left and unknown color */
	    startx = 0; 
	    for (col = 0; col < ncols; col++)
	    {
		parms.val = comp->data.d1[ idx ];
	    
		/* handle zero --- see hb.h for definition of SCALE_EPSILON */
		if (fabs(parms.val) < SCALE_EPSILON)  
		{
		    if (s->force_black)
			color = black - startcolor;
		    else
			color = 0; /* hmm...*/
		}
		/* call strategy and clamp colors */
		else 
		{
		    color = s->strat(&parms);
		    if (color < 0)
		    {
			fprintf(stderr,"BUG: adjusting color from %d to 0\n",color);
			fprintf(stderr,"min = %e, max = %e, val = %e, numcolors = %d\n",
				parms.min, parms.max, parms.val,parms.numcolors);
			color = 0;
		    }
		    else if (color > numcolors)
		    {
			fprintf(stderr,"BUG: adjusting color from %d to %d\n",color, numcolors);
			fprintf(stderr,"min = %e, max = %e, val = %e\n",
				parms.min, parms.max, parms.val);
			color = numcolors;
		    }
		}
#ifdef FILE_DUMP
		fprintf(debugfp,"%9.1e %3d ",parms.val, color);
		if ((col % 8) == 0 && (col != 0))
		    fprintf(debugfp,"\n%4d: ",col+1);
#endif	
		/* draw it */
		/* XDrawPoint( s->display, p, s->gcs[color+startcolor], 
		 *		 col, scanline ); 
		 */


		/* output to array */
		array[out_idx] = colors[color+startcolor].Red;
		out_idx++;
		array[out_idx] = colors[color+startcolor].Green;
		out_idx++;
		array[out_idx] = colors[color+startcolor].Blue;
		out_idx++;

		/* update index */
		idx++;
	    } /* end of col loop */
	} /* end of scanline loop */
	TRACE("ScaleArray: exit!\n");

#ifdef FILE_DUMP
	fprintf(debugfp,"\n");
	fclose(debugfp);
#endif
	return(OK);
    } /* end of if no partitions */

    /* WE HAVE PARTITIONS */
    CurRowPtn = CurColPtn = 0;
    if (num_row_ptns <= 1) /* only have col ptns, I guess */
    {
	NewRowPtn[0] = 0;
	NewRowPtn[1] = nrows;
    }
    if (num_col_ptns <= 1) /* only have row ptns, I guess */
    {
	NewColPtn[0] = 0;
	NewColPtn[1] = ncols;
    }
    
    /* for all rows (scanlines) on the screen */
    for (scanline = 0; scanline < nrows; scanline++)
    {
#ifdef FILE_DUMP
	fprintf(debugfp,"\nOn Scanline %d:\n%4d: ",scanline,0);
#endif
        /* init current line to start at left and unknown color */
	startx = 0;

	/* partition stuff -- current column partition is 0 */
	CurColPtn = 0;

	/* if not on a new partition line, leave row alone */
	if (scanline != NewRowPtn[CurRowPtn+1])
	{
	    overwrite_row = 0;
	}
	else /* changing row partitions */
	{
	    /* update the row ptn */
	    overwrite_row = 1;
 	    CurRowPtn++;
	}
	/* now go through the cols */
	for (col = 0; col < ncols; col++)
	{
	    overwrite_col = 0;
	    /* first,see if on new col ptn */
	    if (col == NewColPtn[CurColPtn+1])
	    {
		CurColPtn++;
		if (!overwrite_row)
		{
		    overwrite_col = 1;
		}
	    }
	    /* handle overwrite of a row/col */
	    if ((overwrite_row) || (overwrite_col))
	    {
		/* parms.val = parms.absmax; */
		/* force it to be white */
		/* 7/12/95 -- change ptn lines to be black if BG is the ptn line color */
		if (black == PTN_LINE_COLOR)
		{
		    color = BLACK_INDEX - startcolor;
		}
		else
		{
		    color = PTN_LINE_COLOR - startcolor;
		}
	    }
	    else /* do the normal scaling */
	    {
		/* get the value */
		parms.val = comp->data.d1[ idx ];
		/* update the index into the compression matrix */
		idx++;
	    
		/* handle zero --- see hb.h for definition of SCALE_EPSILON */
		if (fabs(parms.val) < SCALE_EPSILON)  
		{
		    if (s->force_black)
			color = black - startcolor;
		    else
			color = 0; /* hmm...*/
		}
		/* call strategy and clamp colors */
		else 
		{
		    color = s->strat(&parms);
		    if (color < 0)
		    {
			fprintf(stderr,"BUG: adjusting color from %d to 0\n",color);
			fprintf(stderr,"min = %e, max = %e, val = %e, numcolors = %d\n",
				parms.min, parms.max, parms.val,parms.numcolors);
			color = 0;
		    }
		    else if (color > numcolors)
		    {
			fprintf(stderr,"BUG: adjusting color from %d to %d\n",color, numcolors);
			fprintf(stderr,"min = %e, max = %e, val = %e\n",
			    parms.min, parms.max, parms.val);
			color = numcolors;
		    }
		}
	    }
#ifdef FILE_DUMP
	    fprintf(debugfp,"%9.1e %3d ",parms.val, color);
	    if ((col % 8) == 0 && (col != 0))
		fprintf(debugfp,"\n%4d: ",col+1);
#endif	
	    /* output to array */
	    array[out_idx] = colors[color+startcolor].Red;
	    out_idx++;
	    array[out_idx] = colors[color+startcolor].Green;
	    out_idx++;
	    array[out_idx] = colors[color+startcolor].Blue;
	    out_idx++;

	} /* end of col loop */

    } /* end of scanline loop */
    TRACE("ScaleMat: exit!\n");

#ifdef FILE_DUMP
    fprintf(debugfp,"\n");
    fclose(debugfp);
#endif
    return(OK);

} /* end of scale_array */
