/*
 * File: util.c 
 *
 * References: 
 *  Users' Guide for the Harwell-Boeing Sparse Matrix Collection (Release I),
 *     Duff, Grimes, and Lewis, publisher?, Oct. 1992
 *
 * Includes: utility routines for readhb3: formatting, IO, and error handling
 *
 * Started: 2/4/94
 *
 * Modified: 2/11/94 to add MALLOC and FREE macro invocations
 *           2/12/94 to clean up comments, etc.
 *           3/18/94 replace some (commented out) fprintfs with DEBUGx calls
 *           
 * Tom Loos, LH310, 855-3609, tloos@cs.indiana.edu
 *
 */

#include <hb.h>

/*
 * Function: HBCleanup()
 *
 * Purpose: to do generic cleanup on an HB Matrix (right now, just unmalloc
 *          everything 
 *
 * Usage: HBMatrix hb;
 *        if (some_ugly_error)
 *        {
 *            HBCleanup(&hb);  
 *            printf("you trashed the matrix, dude -- see ya later!\n");
 *            exit(BAD);
 *
 * Parameter: hb - a pointer to an HB matrix
 *
 * Return Value: none
 * 
 * Calls: FREE
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: 2/7/94 
 * 
 * Modified: 2/11/94 use FREE instead of free
 *           2/13/94 added check for NULL input matrix
 */
void HBCleanup( HBMatrix *hb)
{
    TRACE("entered HBCleanup\n");    
    /* do work only if we have a non-null matrix */
    if (hb != (HBMatrix *) NULL)
    {
	if (hb->exact != (double *) NULL)
	    FREE(hb->exact);

	if (hb->guess != (double *) NULL)
	    FREE(hb->guess);

	if (hb->RHS != (double *) NULL)
	    FREE(hb->RHS);

	if (hb->values != (double *) NULL)
	    FREE(hb->values);

	if (hb->row_inds != (int *) NULL)
	    FREE(hb->row_inds);

	if (hb->col_inds != (int *) NULL)
	    FREE(hb->col_inds);
    }
    TRACE("left HBCleanup\n");    
}
/*
 * Function Name: error
 *
 * Purpose: to print out an error message and terminate
 *
 * Parameters: string - error message to be printed
 *             rc - return code
 *
 * Return Values: none
 * 
 * Calls: fprintf exit
 *
 * Notes: only used in STANDALONE mode (there's an error macro for 
 *        matlab (non-STANDALONE) mode )
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started:   about 2/4/94
 *
 * Modified:  2/11/94 - add ifdef STANDALONE
 *            2/12/94 - added this big ol' comment
 *
 */
#ifdef STANDALONE
void error(char *string, int rc)
{
    fprintf(stderr, "%s (error code = %d)\n", string, rc);
    exit( rc );
}
#endif

/*
 * Function: FormatInt
 *
 * Purpose: To get an integer (with potential leading 0's) from
 *          a character array of size size (that is i.e. atoi() 
 *
 * Parameters: inbuff - the input array 
 *             size - how big the integer should be
 *
 * Return Value: the integer that the char array represents
 *
 * Calls Made: none
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: 1/24/94
 *
 * Modified: 1/31/94: cleaned up and added comments
 *           2/12/94: added negative number support
 *           1/13/96: added check to see if at EOS.
 */
int FormatInt( char *inbuff, int size)
{
    int i;
    int answer = 0;
    int sign = 1;
    TRACE("entered FormatInt\n");    
    for (i = 0; i < size; i++)
    {
	if (inbuff[i] == '\0')
	    break;
	if ((inbuff[i] >= '0') && (inbuff[i] <= '9'))
	    answer = (inbuff[i]-'0') + 10 * answer;
	else if (inbuff[i] == '-')
	    sign = -1;
    }

    TRACE("leaving FormatInt\n");    
    return answer*sign;
}
/*
 * Function: FormatFloat
 *
 * Purpose: To get a floating point number from a string, converting
 *          from FORTRAN format to C format.
 *
 * Usage:  char floatstr[50];
 *         double f;
 *         strcpy(floatstr,"1D-30"); // atof won't handle this
 *         f = FormatFloat(floatstr,strlen(floatstr));
 *
 * Parameters: inbuff - the input string
 *             size - the maximum size of the string
 * 
 * Return Value: the float the string represents
 *
 * Calls Made: atof
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: about 2/4/94
 *
 * Modified:  2/12/94 - add comments
 */
double FormatFloat( char *inbuff, int size )
{
    char *s,c;
    double f;


    /* turn inbuff into a string, but save the character */
    TRACE("entered FormatFloat\n");
    c = inbuff[size];
    inbuff[size] = '\0';

    /* get a pointer to the array (leave inbuff alone) */
    s = inbuff;

    /* turn d/Ds and q/Qs into an e, to follow the C convention */

    while (*s++ != '\0')
    {
	if ((*s == 'd') || (*s == 'D') || (*s == 'Q') || (*s == 'q'))
	{
	    *s = 'e';
	    break;
	}
    }

    /* convert string to float */
    f = atof(inbuff);
    
    /* restore original contents of inbuff */
    inbuff[size] = c;
    
    TRACE("leaving FormatFloat\n");
    /* return f */
    return(f);
}

/* 
 * Function: FormatNumDataType
 * 
 * Purpose: Return the enum value for a numeric data type given the
 * input character from the file 
 * 
 * Parameters: c - the input character
 *             n - a pointer to the NumericalDataType to be set
 *
 * Return Values: none
 *
 * Calls: toupper sprintf error
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: about 2/1/94
 *
 * Modified: 2/12/94 - add comments 
 *
 */
void FormatNumDataType( char c, NumericalDataType *n)
{
    char errmsg[100];
    c = toupper(c);
    TRACE("entered FormatNumDataType\n");
    switch (c)
    {
    case 'R': 
	*n = IsReal;
	break;
    case 'C': 
	*n = IsComplex;
	break;
    case 'P': 
	*n = IsPattern;
	break;
    default :
#ifdef STANDALONE	
	sprintf(errmsg,"error in FormatNumDataType, c = %c\n",c);
	error(errmsg, ARGERR);
#endif
	break;
    }
    TRACE("leaving FormatNumDataType\n");
}

/* 
 * Function: FormatSymType
 * 
 * Purpose: Return the enum value for a symmetry type given the
 * input character from the file 
 * 
 * Parameters: c - the input character
 *             n - a pointer to the SymmetryType to be set
 *
 * Return Values: none
 *
 * Calls: toupper sprintf error
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: about 2/1/94
 *
 * Modified: 2/12/94 - add comments 
 *
 */
void FormatSymType( char c, SymmetryType *n)
{
    char errmsg[100];
    c = toupper(c);
    TRACE("entered FormatSymType\n");

    switch (c)
    {
    case 'S': 
	*n = IsSymmetric;
	break;
    case 'U': 
	*n = IsUnsymmetric;
	break;
    case 'H': 
	*n = IsHermetian;
	break;
    case 'Z': 
	*n = IsSkewSymmetric;
	break;
    case 'R': 
	*n = IsRectangular;
	break;
    default :
#ifdef STANDALONE
	sprintf(errmsg,"error in FormatSymType, c = %c\n",c);
	error(errmsg, ARGERR);
#endif
	break;
    }
    TRACE("leaving FormatSymType\n");
}

/* 
 * Function: FormatAssemType
 * 
 * Purpose: Return the enum value for an assembly type given the
 * input character from the file 
 * 
 * Parameters: c - the input character
 *             n - a pointer to the AssemblyType to be set
 *
 * Return Values: none
 *
 * Calls: toupper sprintf error
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: about 2/1/94
 *
 * Modified: 2/12/94 - add comments 
 *
 */
void FormatAssemType( char c, AssemblyType *n)
{
    char errmsg[100];

    c = toupper(c);
    TRACE("entered FormatAssemType\n");
    switch (c)
    {
    case 'A': 
	*n = IsAssembled;
	break;
    case 'E': 
	*n = IsElemental;
	break;
    default :
#ifdef STANDALONE
	sprintf(errmsg,"error in FormatAssemType, c = %c\n",c);
	error(errmsg, ARGERR);
#endif
	break;
    }
    TRACE("leaving FormatAssemType\n");
}

/*
 * Function: GetInputLine
 *
 * Purpose: get one line (terminated by a \n) from the input  and return 
 * that line as a string. 
 *
 * Note: this does 1 char at a time I/O (yuck)
 *
 * Parameters: fd = the input file descriptor
 *             buff = a buffer for the input line
 *
 * Return Value: the number of characters actually taken from the input
 *
 * Calls: read
 *
 * Programmer: Tom Loos, tloos@cs.indiana.edu, LH 310, 812 855-3609
 *
 * Started: 1/27/94
 *
 * Modified: 1/30/94 - actually return a string not a newline terminated
 *                     array! -- tjl
 *           2/12/94 - add more comments -- tjl
 *           5/9/94  - convert input to upper case -- tjl
 */
int GetInputLine(int fd, char *buff)
{
    int i,in;

    /* init char count */
    TRACE("entered GetInputLine -- exit not reported\n");
    i = 0;

    /* while not newline, accumulate chars */
    /* convert input to upper case 5/9/94 */
    while ((in=read(fd,&(buff[i]),1)) && 
	   ((buff[i] = (toupper(buff[i]))) != '\n'))
    {
	i++;
    }

    /* if we have a newline, turn line into a string */
    if (buff[i] == '\n')
    {
	buff[i] = '\0'; 
	return ++i;
    }
    /* else IO problem */
    return IOERR;
}

/*
 * Function: Make2DArray
 *
 * Purpose: This creates a 2d array out of the void (or at
 *          least out of malloc()).
 *
 * Parameters: ptr  -- a pointer to a 2D array pointer (yes, a triple pointer)
 *             rows -- the number of rows to be created
 *             cols -- the number of columns to be created
 *             row_major -- 0 if a col major matrix to be created, anything
 *                          else for a row major matrix 
 *
 * Return Values: OK if all went well
 *                MEMERR if it didn't
 *
 * Calls: MALLOC fprintf
 *
 * Notes: never used outside of STANDALONE mode
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 * 
 * Started: 2/6/94 
 * 
 * Modified: 2/12/94 -- clean up comments and change malloc to MALLOC so
 *             can use this with MATLAB (tjl)
 *           3/18/94 -- TRACE and added row_major parameter and it's support
 *
 */
int Make2DArray( double ***ptr, int rows, int cols, int row_major)
{
    double **local;
    int i, major_size, minor_size;
    
    TRACE("entered Make2DArray\n");
    /* allocate space for the row pointers */
    
    /* if ((local = (double **) malloc(sizeof(double *) * rows)) == NULL) */

    if ((local = (double **) MALLOC(rows, sizeof(double *))) == NULL)
    {
	return(MEMERR);
    }
    
    if (row_major)
    {
	major_size = rows; /* major size is rows for row major */
	minor_size = cols;
    }
    else
    {
	major_size = cols;
	minor_size = rows;
    }
    /* malloc the columns (1 col. per row) */
    for (i = 0; i < major_size; i++)
    {
	if ((local[i] = (double *) MALLOC(minor_size,sizeof(double))) == NULL) 
	{
	    fprintf(stderr,"MakeArray: error allocating local[%d]\n",i);
	    return(MEMERR);
	}
    }
    
    *ptr = local;
    TRACE("leaving Make2DArray\n");
    return(OK);
}

/*
 * Function: Free2DArray
 *
 * Purpose: This destroys a malloc'd row major 2 d array of doubles
 *
 * Parameters: ptr  -- the 2D array of doubles 
 *             rows -- the number of rows to be destroyed 
 *
 * Return Values: none
 *
 * Calls: FREE
 * 
 * Note: never tested outside of STANDALONE mode
 *
 * Programmer:  Tom Loos, tloos@cs.indiana.edu, LH 310, 812 855-3609
 * 
 * Started: 2/6/94 
 * 
 * Modified: 2/12/94 use FREE for matlab and clean up comments
 *
 */
void Free2DArray( double **ptr, int rows )
{
    int i,j;
    TRACE("entering Free2DArray\n");
    /* free up the columns */
    for (i = 0; i < rows; i++)
    {
	FREE(ptr[i]);
    }
    /* finally free up the row index pointer */
    FREE(ptr);
    TRACE("leaving Free2DArray\n");
}

/*
 * Function: HBInit()
 *
 * Purpose: set up default values for an HB matrix 
 *
 * Algorithm:  a whole bunch of assignment statements 
 *
 * Return Values: none
 * 
 * Calls: none
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: 2/6/94
 * 
 * Modified: 2/12/94 -- clean up comments
 *
 */
void HBInit(HBMatrix *hb)
{
    /* header init */
    TRACE("entering HBInit\n");
    strcpy(hb->head.Title,"default title");
    strcpy(hb->head.Key,"key");
    hb->head.TotalLines = 0;
    hb->head.PointerLines = 0;
    hb->head.IndexLines = 0;
    hb->head.ValueLines = 0;
    hb->head.RHSLines = 0;
    
    hb->head.NumericalData = IsReal;
    hb->head.Symmetry = IsSymmetric;
    hb->head.Assembly = IsAssembled;
    hb->head.NumRows = 0;
    hb->head.NumCols = 0;
    hb->head.NumRowIndex = 0;
    hb->head.NumElemMatEntries = 0;

    /* init formats */
    hb->head.PointerFormat.count = 0;
    hb->head.PointerFormat.format = ' ';
    hb->head.PointerFormat.width = 0;
    hb->head.PointerFormat.decimal_places = 0;

    hb->head.RowIndexFormat.count = 0;
    hb->head.RowIndexFormat.format = ' ';
    hb->head.RowIndexFormat.width = 0;
    hb->head.RowIndexFormat.decimal_places = 0;

    hb->head.ValueFormat.count = 0;
    hb->head.ValueFormat.format = ' ';
    hb->head.ValueFormat.width = 0;
    hb->head.ValueFormat.decimal_places = 0;

    hb->head.RHSFormat.count = 0;
    hb->head.RHSFormat.format = ' ';
    hb->head.RHSFormat.width = 0;
    hb->head.RHSFormat.decimal_places = 0;

    /* done with format init */

    hb->head.RHSVectorFormat = IsFull;
    hb->head.GuessSupplied = ' ';
    hb->head.ExactSupplied = ' ';
    hb->head.NumRHS = 0;
    hb->head.NumRHSRowIndex = 0;
    
    hb->head.MaxValue = -HUGE_VAL; /* this is smaller than anything */
    hb->head.MinValue = HUGE_VAL; /* this is larger */
    hb->head.AbsMinValue = HUGE_VAL; /* this is way too big */
    hb->head.AbsMaxValue = 0.0; /* this is the smallest abs. number */
    
    /* done with header init */

    hb->col_inds = (int *) NULL;
    hb->row_inds = (int *) NULL;
    hb->values = (double *) NULL;
    hb->RHS = (double *) NULL;
    hb->guess = (double *) NULL;
    hb->exact = (double *) NULL;
    TRACE("leaving HBInit\n");
}
    
/*
 * Function: GetBuffLine
 *
 * Purpose: get one line (terminated by a \n) from the big buffer
 *          and return that line.  The line can have at most max characters
 *
 * Parameters: inbuff = the (big) buffer that holds the raw input data 
 *             outbuff = a buffer for the desired line
 *             max = the maximum number of characters to be extracted from
 *                   the buffer
 *             nl_seen = return 1 if a new line actually seen or 0 if 
 *                       not seen (max reached)
 *
 * Return Value: the number of characters actually taken from the buffer
 *
 * Calls: none
 *
 * Note: If no newline is seen, one byte WILL BE DESTROYED to accomodate
 *       the NULL character.  newlines will be converted to '\0' in
 *       in buff
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: 1/27/94
 *
 * Modified:  2/ 4/94 -- add parameter nl_seen
 *            2/ 9/94 -- try to get rid of \n's in the input to avoid data copy
 *            2/12/94 -- clean up comments slightly
 * 
 */

int GetBuffLine(char *inbuff, char **outbuff, int max, int *nl_seen)
{
    int i;
    i = 0;

    /* while we don't have a newline and we haven't tried to get too many
       chars */

    TRACE("entering GetBuffLine\n");
    /* change to look for string with out copy */
    *outbuff = inbuff;
    while ((inbuff[i] != '\n') && (inbuff[i] != '\0') && (i < max)) 
    {
	i++;
    }
    /* set newline seen (nl_seen) */
    if (inbuff[i] == '\n')
	*nl_seen = 1;
    else
	*nl_seen = 0;
	
    /* return a string */
    inbuff[i] = '\0';

    /* and the right count */
    TRACE("leaving GetBuffLine\n");
    return i+1;
}
/*
 * Function: Get1DArray
 *
 * Purpose: malloc out a 1 dimensional array and return it with error
 *          handling
 *
 * Parameters: count = the number of items to be malloc'd
 *             size = the size of one item
 *             name = the name of the array (for error printout)
 *             hb = the matrix to be cleaned up in case of an error
 *
 * Calls: MALLOC
 *
 * Return Value: a character array of the total size requested 
 *               total size = count * size
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: 2/7/94
 *
 * Modified: 2/11/94 - added MALLOC reference for matlab and deleted var 
 *                     bytes_needed (tjl)
 *           2/12/94 - clean up comments slightly 
 *           3/18/94 - add DEBUG calls
 */
char *Get1DArray(int count, int size, char *name, HBMatrix *hb)
{

    static char *ptr; /* static to assure it will pass back ok */

    char errmsg[100];

    /* if malloc fails, clean up and print error */
    /* start 2/11 */
    /* if ((ptr = malloc(bytes_needed)) == (char *) NULL) */

    TRACE("entering Get1DArray\n");
    sprintf(errmsg, "Get1DArray: getting %d by %d bytes\n",count, size); 
    DEBUG1(errmsg);

    if ((ptr = (char *) MALLOC(count,size)) == (char *) NULL) 
    {
	sprintf(errmsg,
		"Get1DArray: couldn't allocate %d elements for the %s array",
		count,name);
	HBCleanup(hb);
	error(errmsg,MEMERR);
    }
    /* else return pointer */
    TRACE("leaving Get1DArray\n");
    return ptr;
}
/*
 * Function: GetDenseMatrix
 *
 * Purpose: malloc out a dense matrix, initialize it to the desired value,
 *  and return it with error handling
 *
 * Parameters: nrows     = number of rows in matrix
 *             ncols     = number of columns in matrix
 *             row_major = is matrix row major? 
 *                         (ignored if ONE_D_STORAGE > 0)
 *             one_d     = should storage be 1 dimensional or 2 dimensional
 *             initval   = initial value of the entries of the matrix
 *
 * Calls: MALLOC Get1DArray Make2DArray
 *
 * Return Value: the a pointer to the dense matrix 
 *
 * Note: row_major is ignored if one_d is 0 (implying a 2 D array)
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: 2/13/94
 *
 * Modified: 2/16/94 -- set up one_d parameter
 *           3/18/94 -- replace fprintf with DEBUG calls & added surport
 *                      for 2D col major dense matrices
 */
DenseMatrix *GetDenseMatrix( int nrows, int ncols, int row_major, int one_d, double initval)
{
    static DenseMatrix *d;
    char errmsg[100];
    int badnews, i, j,max_i, max_j;

    TRACE("entering GetDenseMatrix\n");
    if ((d=(DenseMatrix *) MALLOC(1,sizeof(DenseMatrix))) == 
	(DenseMatrix *) NULL)
    {
	error("failed to allocate a Dense Matrix",MEMERR);
	/* probably won't get here... */
	return( d ); 
    }

    /* init dense matrix */
    d->NumRows = nrows;
    d->NumCols = ncols;
    /* d->one_d = ONE_D_STORAGE; */

    d->one_d = one_d;
    d->row_major = row_major;

    /* if one_d storage, handle that */
    if (d->one_d)
    {
	d->data.d1 = (double *) Get1DArray( nrows * ncols, 
					 sizeof(double), 
					 "DenseMatrix",
					 (HBMatrix *) NULL);

	badnews = (d->data.d1 == ((double *) NULL));

	/* initialize the data */
	if (!badnews)
	{
	    for (i = 0; i < nrows * ncols; i++)
		d->data.d1[i] = initval;
	}
	sprintf(errmsg,"handled 1D case\n");
	DEBUG2(errmsg);
    }
    else
    {
	sprintf(errmsg,"here 1.8 (2D Dense Matrix)\n");
	DEBUG3(errmsg);
	(void) Make2DArray( &(d->data.d2), nrows, ncols, row_major );
	sprintf(errmsg,"here 1.79\n");
	DEBUG3(errmsg);
	badnews = (d->data.d2 == (double **) NULL);
	
	/* initialize the data */
	if (!badnews)
	{
	    
	    if (row_major) /* do row_major access */
	    {
		max_i = nrows;
		max_j = ncols;
	    }
	    else /* do column major access */
	    {
		max_i = ncols;
		max_j = nrows;
	    }

	    /* init the matrix */
	    for (i = 0; i < max_i; i++)
		for (j = 0; j < max_j; j++)
		    d->data.d2[i][j] = initval;
	}
	else
	{
	    error("failed to allocate the element storage for a dense matrix",
		  MEMERR);
	    return(d);
	}
    }
    TRACE("leaving GetDenseMatrix\n");
    return(d);
}

/*
 * Function: FreeDenseMatrix
 *
 * Purpose: free all storage created by GetDenseMatrix()
 *
 * Parameters: d = a pointer to a dense matrix
 *
 * Calls: FREE Free2DArray
 *
 * Return Value: none
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: 2/13/94
 *
 * Modified: 3/20/94 remove #ONE_D_STORAGE and check for NULL input
 */
void FreeDenseMatrix( DenseMatrix *d )
{

    /* free up the data depending on the type of storage used */

    TRACE("entering FreeDenseMatrix\n");

    if (d != (DenseMatrix *) NULL)
    {
	if (d->one_d)
	    FREE( d->data.d1 );
	else
	{
	    if (d->row_major)
		Free2DArray( d->data.d2, d->NumRows );
	    else
		Free2DArray( d->data.d2, d->NumCols );
	}

	/* free up the pointer */
	FREE( d );
    }
    TRACE("leaving FreeDenseMatrix\n");
}
/*
 * Function: FreeDaMatrix
 *
 * Purpose: free all storage malloc'd for a DaMatrix
 *
 * Parameters: d = a pointer to Da (the) Matrix (now you know how Da Matrices 
 *                 were named -- I couldn't think of anything better)
 *
 * Calls: HBCleanup FreeDenseMatrix
 *
 * Return Value: none
 *
 * Programmer: Tom Loos, Indiana Univ, LH310, +01 (812) 855-3609,
 *             tloos@cs.indiana.edu
 *
 * Started: 3/20/94
 *
 * Modified: 
 */
void FreeDaMatrix( DaMatrix *d )
{

    TRACE("entering FreeDaMatrix\n");
    /*
     * assume the matrix itself was not malloc'd before, so don't nuke the 
     *  pointer, just the child matrix storage
     */
    if (d != (DaMatrix *) NULL)
    {
	if (d->type == HB)
	    HBCleanup(d->data.hb);
	else
	    FreeDenseMatrix(d->data.d);
    }
    TRACE("leaving FreeDaMatrix\n");
}

